/*
 * InfoSNES - Super Nintendo (TM) emulator
 * Copyrigh(c) 2003 Jay's Factory (jays_factory@excite.co.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*-------------------------------------------------------------------*/
/*  Include Files                                                    */
/*-------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "snes.h"
#include "65816.h"
#include "ppu.h"
#include "io.h"
#include "type.h"

/*-------------------------------------------------------------------*/
/*  65816 Resources                                                  */
/*-------------------------------------------------------------------*/

sreg a,x,y,s;
u16 pc,d;
u8 pbr,dbr;
int cycles, oldcycles;
u32 oldpc, oldoldpc;

struct
{
        int c,z,i,d,b,x,m,v,n,e;
} p;

void (*opcoderout[4][256])();

/* Memory pointers for opcode fetching */
/* Here I assume we always run in ROM or RAM */
u8 *mempointers[256][2]; 

int smode;
u8 opcode;
s8 offset;

u8 temp, temp2, tempb, al,ah;
u16 tempa, tempa2, tempw, tempw2;
u32 templ,templ2;
int tempi,tempv, hc;

/*-------------------------------------------------------------------*/
/*  Other Resources                                                  */
/*-------------------------------------------------------------------*/

int loggedins=0;
int caeecount=0;
int inint=0;
int output=0;
int instructions=0;

/*-------------------------------------------------------------------*/
/*  Functions                                                        */
/*-------------------------------------------------------------------*/

void dumpregs()
{
  printf("A=%04X X=%04X Y=%04X S=%04X PC=%02X:%04X DBR=%02X D=%04X\n",a.w,x.w,y.w,s.w,pbr,pc,dbr,d);
  printf("%c%c%c%c%c%c%c%c%c\n",(p.c)?'C':' ',(p.z)?'Z':' ',(p.i)?'I':' ',(p.d)?'D':' ',(p.x)?'X':' ',(p.m)?'M':' ',(p.v)?'V':' ',(p.n)?'N':' ',(p.e)?'E':' ');
  printf("mode %i  instruction %i\n",smode,instructions);
  printf("%06X %06X\n",oldpc,oldoldpc);
}

inline u8 readmem(u8 bank, u16 addr)
{
  bank&=0x7F;

  if (bank<0x40)
  {
    switch (addr&0xF000)
    {
    case 0:
    case 0x1000:
      return ram[addr];
    case 0x2000:
      return readppu(addr);
    case 0x3000:
      return 0xFF;
    case 0x4000:
      return readio(addr);
    case 0x5000:
      return 0xFF;
    case 0x6000:
      return 0xFF;
    case 0x7000:
      return 0xFF;
    case 0x8000: case 0x9000: case 0xA000: case 0xB000:
    case 0xC000: case 0xD000: case 0xE000: case 0xF000:
      return rom[bank][addr&0x7FFF];
    }
  }
  if (bank==0x70)
    return sram[addr];
  if (bank==0x7E)
    return ram[addr];
  if (bank==0x7F)
    return ram[addr+0x10000];
  if (bank==0x60)              /* This should work for Columns */
    return addr>>8;
  if (bank==0x79)              /* This should work for Humans */
    return 0xFF;
  printf("Bad read %02X:%04X   CAEE %i\n",bank,addr,caeecount);
  dumpregs();
  dumpram();
  exit(-1);
}

inline void writemem(u8 bank, u16 addr, u8 val)
{
  bank&=0x7F;

  if (bank<0x40)
  {
    switch (addr&0xF000)
    {
    case 0:
    case 0x1000:
      ram[addr]=val;
      return;
    case 0x2000:
      if ((addr>>8)==0x21) writeppu(addr,val);
      return;
    case 0x4000:
      writeio(addr,val);
      return;
    case 0x7000: /*Wario's Woods writes here during NMI - leftover debugging code?*/
      return;
    case 0x8000: /*First Drop writes to ROM - probably a bug in the demo*/
      return;
    }
  }
  if (bank==0x60)
    return;
  if (bank==0x70)
  {
    sram[addr]=val;
    return;
  }
  if (bank==0x7E)
  {
    ram[addr]=val;
    return;
  }
  if (bank==0x7F)
  {
    ram[addr+0x10000]=val;
    return;
  }
  printf("Bad write %02X:%04X %02X\n",bank,addr,val);
  dumpregs();
  dumpram();
  exit(-1);
}

void updatemode()
{
  if (p.e)
    smode=3;
  else
    smode=((p.x)?1:0)|((p.m)?2:0);
  if (p.x || p.e)
    x.b.h=y.b.h=0;
}

void badopcode()
{
  printf("Bad opcode %02X\n",opcode);
  pc--;
  dumpregs();
  dumpram();
  exit(-1);
}

#define ADCB8()  tempw=a.b.l+temp+((p.c)?1:0);                          \
                p.v=(!((a.b.l^temp)&0x80)&&((a.b.l^tempw)&0x80));       \
                setzn8(tempw);                                          \
                a.b.l=tempw&0xFF;                                       \
                p.c=tempw&0x100;

#define ADCB16() templ=a.w+tempw+((p.c)?1:0);                           \
                p.v=(!((a.w^tempw)&0x80)&&((a.w^templ)&0x80));          \
                setzn16(templ);                                         \
                a.w=templ&0xFFFF;                                       \
                p.c=templ&0x10000;

#define ADCBCD8()                ah=0;                          \
                tempb=a.b.l+temp+(p.c?1:0);                     \
                if (!tempb)                                     \
                   p.z=1;                                       \
                al=(a.b.l&0xF)+(temp&0xF)+(p.c?1:0);            \
                if (al>9)                                       \
                {                                               \
                        al-=10;                                 \
                        al&=0xF;                                \
                        ah=1;                                   \
                }                                               \
                ah+=((a.b.l>>4)+(temp>>4));                     \
                p.n=ah&8; \
                p.v=(((ah << 4) ^ a.b.l) & 128) && !((a.b.l ^ temp) & 128); \
                p.c=0;                                                      \
                if (ah>9)                                                   \
                {                                                           \
                        p.c=1;                                              \
                        ah-=10;                                             \
                        ah&=0xF;                                            \
                }                                                           \
                a.b.l=(al&0xF)|(ah<<4);

#define ADCBCD16()                ah=0;                          \
                tempw=a.w+tempw+(p.c?1:0);                     \
                if (!tempw)                                     \
                   p.z=1;                                       \
                al=(a.b.l&0xF)+(tempw&0xF)+(p.c?1:0);            \
                if (al>9)                                       \
                {                                               \
                        al-=10;                                 \
                        al&=0xF;                                \
                        ah=1;                                   \
                }                                               \
                ah+=((a.b.l>>4)+((tempw>>4)&0xF));                     \
                if (ah>9)                                               \
                {                                                        \
                        p.c=1;                                            \
                        ah-=10;                                            \
                        ah&=0xF;                                            \
                }                                                            \
                a.b.l=(al&0xF)|(ah<<4);                                       \
                ah=0;                                                          \
                al=(a.b.h&0xF)+((tempw>>8)&0xF)+(p.c?1:0);            \
                if (al>9)                                       \
                {                                               \
                        al-=10;                                 \
                        al&=0xF;                                \
                        ah=1;                                   \
                }                                               \
                ah+=((a.b.h>>4)+((temp>>12)&0xF));                     \
                if (ah&8) p.n=1;                                \
                p.v=(((ah << 12) ^ a.w) & 0x8000) && !((a.w ^ tempw) & 0x8000); \
                p.c=0;                                                      \
                if (ah>9)                                                   \
                {                                                           \
                        p.c=1;                                              \
                        ah-=10;                                             \
                        ah&=0xF;                                            \
                }                                                           \
                a.b.h=(al&0xF)|(ah<<4);

#define SBCB8()  tempw=temp+((p.c)?0:1); \
                templ=a.b.l-tempw;                           \
                tempv=(short)a.b.l-(short)tempw;            \
                p.v=((a.b.l^tempw)&(a.b.l^(u8)tempv)&0x80); \
                p.c=tempv>=0;                         \
                a.b.l=templ&0xFF;                           \
                setzn8(a.b.l);

#define SBCB16() templ2=tempw+((p.c)?0:1);                           \
                templ=a.w-templ2;                            \
                tempv=(long)a.w-(long)templ2;            \
                p.v=((a.w^templ2)&(a.w^(u16)tempv)&0x8000); \
                p.c=tempv>=0;                         \
                a.w=templ&0xFFFF;                           \
                setzn16(a.w);

#define SBCBCD8()  tempi=p.c; \
                tempw=a.b.l-(temp+(p.c)?0:1); \
                setzn8(tempw); \
                tempv=(signed)a.b.l-(signed)(temp+(p.c)?0:1);  \
                p.v=((a.b.l&128)>0)^((tempv&0x100)!=0); \
                hc=0;                               \
                al=(a.b.l&15)-(temp&15)-((p.c)?0:1);      \
                ah=(a.b.l>>4)-(temp>>4);                \
                if (al&16)                           \
                {                                   \
                         al-=6;                      \
                         al&=0xF;                    \
                         hc=1;                       \
                }                                   \
                if (hc) ah--;                       \
                p.c=1;  \
                if (ah&16)                           \
                {                                   \
                         p.c=0; \
                         ah-=6;                      \
                         ah&=0xF;                    \
                }                                   \
                a.b.l=(al&0xF)|((ah&0xF)<<4);


#define SBCBCD16()  tempi=p.c; \
                tempw2=a.w-(tempw+(p.c)?0:1); \
                setzn16(tempw2); \
                tempv=(signed)a.w-(signed)(tempw+(p.c)?0:1);  \
                p.v=((a.w&0x8000)>0)^((tempv&0x10000)!=0); \
                hc=0;                               \
                al=(a.b.l&15)-(tempw&15)-((p.c)?0:1);      \
                ah=(a.b.l>>4)-((tempw>>4)&0xF);                \
                if (al&16)                           \
                {                                   \
                         al-=6;                      \
                         al&=0xF;                    \
                         hc=1;                       \
                }                                   \
                if (hc) ah--;                       \
                hc=0;   \
                if (ah&16)                           \
                {                                   \
                         ah-=6;                      \
                         ah&=0xF;                    \
                         hc=1;                        \
                }                                   \
                a.b.l=(al&0xF)|((ah&0xF)<<4);         \
                al=(a.b.h&15)-((tempw>>8)&0xF)-hc;      \
                ah=(a.b.h>>4)-((tempw>>12)&0xF);                \
                hc=0; \
                if (al>9)                           \
                {                                   \
                         al-=6;                      \
                         al&=0xF;                    \
                         hc=1;                       \
                }                                   \
                if (hc) ah--;                       \
                p.c=1; \
                if (ah>9)                           \
                {                                   \
                         p.c=0; \
                         ah-=6;                      \
                         ah&=0xF;                    \
                }                                   \
                a.b.h=(al&0xF)|((ah&0xF)<<4);

#define setzn8(v) p.z=!(v); p.n=(v)&0x80
#define setzn16(v) p.z=!(v); p.n=(v)&0x8000

#define push(v) writemem(0,s.w--,v)
#define pull() readmem(0,++s.w)
#define imm() readfast(pbr,pc); pc++
#define immw() readfast(pbr,pc)|(readfast(pbr,pc+1)<<8); pc+=2
#define zp()   templ=readfast(pbr,pc)+d; pc++;
#define zpx()  templ=readfast(pbr,pc)+d+x.w; pc++;
#define zpy()  templ=readfast(pbr,pc)+d+y.w; pc++;
#define sr()   templ=readfast(pbr,pc)+s.w; pc++;
#define abs()  templ=readfast(pbr,pc)|(readfast(pbr,pc+1)<<8)|(dbr<<16); pc+=2;
#define absx() templ=(readfast(pbr,pc)|(readfast(pbr,pc+1)<<8)|(dbr<<16))+x.w; pc+=2;
#define absy() templ=(readfast(pbr,pc)|(readfast(pbr,pc+1)<<8)|(dbr<<16))+y.w; pc+=2;
#define along() templ=readfast(pbr,pc)|(readfast(pbr,pc+1)<<8)|(readfast(pbr,pc+2)<<16); pc+=3;
#define longx() templ=(readfast(pbr,pc)|(readfast(pbr,pc+1)<<8)|(readfast(pbr,pc+2)<<16))+x.w; pc+=3;

#define getb() readmem(templ>>16,templ)
#define getw() readmem(templ>>16,templ)|(readmem(templ>>16,templ+1)<<8)
#define setb(v) writemem(templ>>16,templ,v)
#define setw(v) writemem(templ>>16,templ,v); writemem(templ>>16,templ+1,v>>8)

inline void PHP()
{
  u8 temp=0;
  if (p.c) temp|=1;
  if (p.z) temp|=2;
  if (p.i) temp|=4;
  if (p.d) temp|=8;
  if (!p.e && p.x) temp|=0x10;
  if (!p.e && p.m) temp|=0x20;
  if (p.v) temp|=0x40;
  if (p.n) temp|=0x80;
  push(temp);
}

inline u8 compp()
{
  u8 temp=0;
  if (p.c) temp|=1;
  if (p.z) temp|=2;
  if (p.i) temp|=4;
  if (p.d) temp|=8;
  if (!p.e && p.x) temp|=0x10;
  if (!p.e && p.m) temp|=0x20;
  if (p.v) temp|=0x40;
  if (p.n) temp|=0x80;
  return temp;
}

inline void PLP()
{
  u8 temp=pull();
  p.c=temp&1;
  p.z=temp&2;
  p.i=temp&4;
  p.d=temp&8;
  if (!p.e) p.x=temp&0x10;
  if (!p.e) p.m=temp&0x20;
  p.v=temp&0x40;
  p.n=temp&0x80;
  updatemode();
}

inline void ind()
{
  tempw=readfast(pbr,pc)+d; pc++;
  templ=readfast(0,tempw)|(readfast(0,tempw+1)<<8)|(dbr<<16);
}

inline void indy()
{
  tempw=readfast(pbr,pc)+d; pc++;
  templ=(readfast(0,tempw)|(readfast(0,tempw+1)<<8)|(dbr<<16))+y.w;
}

inline void indsy()
{
  tempw=readfast(pbr,pc)+s.w; pc++;
  templ=(readfast(0,tempw)|(readfast(0,tempw+1)<<8)|(dbr<<16))+y.w;
}

inline void indl()
{
  tempw=readfast(pbr,pc)+d; pc++;
  templ=readfast(0,tempw)|(readfast(0,tempw+1)<<8)|(readfast(0,tempw+2)<<16);
}

inline void indly()
{
  tempw=readfast(pbr,pc)+d; pc++;
  templ=(readfast(0,tempw)|(readfast(0,tempw+1)<<8)|(readfast(0,tempw+2)<<16))+y.w;
}

inline void indx()
{
  tempw=readfast(pbr,pc)+(readfast(pbr,pc+1)<<8)+x.w; pc+=2;
  templ=readfast(pbr,tempw)|(readfast(pbr,tempw+1)<<8);
}

#define ADC8() if (p.d) { ADCBCD8(); } else { ADCB8(); }
#define ADC16() if (p.d) { ADCBCD16(); } else { ADCB16(); }
#define AND8() a.b.l&=temp; setzn8(a.b.l);
#define AND16() a.w&=tempw; setzn16(a.w);
#define ASL8() p.c=temp&0x80; temp<<=1; setzn8(temp);
#define ASL16() p.c=tempw&0x8000; tempw<<=1; setzn16(tempw);
#define BIT8() setzn8(a.b.l&temp); p.n=temp&0x80; p.v=temp&0x40;
#define BIT16() setzn16(a.w&tempw); p.n=tempw&0x8000; p.v=tempw&0x4000;
#define CMP8() setzn8(a.b.l-temp); p.c=(a.b.l>=temp);
#define CMP16() setzn16(a.w-tempw); p.c=(a.w>=tempw);
#define CPX8() setzn8(x.b.l-temp); p.c=(x.b.l>=temp);
#define CPX16() setzn16(x.w-tempw); p.c=(x.w>=tempw);
#define CPY8() setzn8(y.b.l-temp); p.c=(y.b.l>=temp);
#define CPY16() setzn16(y.w-tempw); p.c=(y.w>=tempw);
#define DEC8() temp--; setzn8(temp);
#define DEC16() tempw--; setzn16(tempw);
#define DEX8() x.b.l--; setzn8(x.b.l);
#define DEX16() x.w--; setzn16(x.w);
#define DEY8() y.b.l--; setzn8(y.b.l);
#define DEY16() y.w--; setzn16(y.w);
#define EOR8() a.b.l^=temp; setzn8(a.b.l);
#define EOR16() a.w^=tempw; setzn16(a.w);
#define INC8() temp++; setzn8(temp);
#define INC16() tempw++; setzn16(tempw);
#define INX8() x.b.l++; setzn8(x.b.l);
#define INX16() x.w++; setzn16(x.w);
#define INY8() y.b.l++; setzn8(y.b.l);
#define INY16() y.w++; setzn16(y.w);
#define LDA8() a.b.l=temp; setzn8(a.b.l);
#define LDA16() a.w=tempw; setzn16(a.w);
#define LDX8() x.b.l=temp; setzn8(x.b.l);
#define LDX16() x.w=tempw; setzn16(x.w);
#define LDY8() y.b.l=temp; setzn8(y.b.l);
#define LDY16() y.w=tempw; setzn16(y.w);
#define LSR8() p.c=temp&1; temp>>=1; setzn8(temp);
#define LSR16() p.c=tempw&1; tempw>>=1; setzn16(tempw);
#define ORA8() a.b.l|=temp; setzn8(a.b.l);
#define ORA16() a.w|=tempw; setzn16(a.w);
#define ROL8() tempi=temp&0x80; temp<<=1; if (p.c) temp|=1; setzn8(temp); p.c=tempi;
#define ROL16() tempi=tempw&0x8000; tempw<<=1; if (p.c) tempw|=1; setzn16(tempw); p.c=tempi;
#define ROR8() tempi=temp&1; temp>>=1; if (p.c) temp|=0x80; setzn8(temp); p.c=tempi;
#define ROR16() tempi=tempw&1; tempw>>=1; if (p.c) tempw|=0x8000; setzn16(tempw); p.c=tempi;
#define SBC8() if (p.d) { SBCBCD8(); } else { SBCB8(); }
#define SBC16() if (p.d) { SBCBCD16(); } else { SBCB16(); }
#define STA8() writemem(templ>>16,templ,a.b.l);
#define STA16() writemem(templ>>16,templ,a.b.l); writemem(templ>>16,templ+1,a.b.h);
#define STX8() writemem(templ>>16,templ,x.b.l);
#define STX16() writemem(templ>>16,templ,x.b.l); writemem(templ>>16,templ+1,x.b.h);
#define STY8() writemem(templ>>16,templ,y.b.l);
#define STY16() writemem(templ>>16,templ,y.b.l); writemem(templ>>16,templ+1,y.b.h);
#define STZ8() writemem(templ>>16,templ,0);
#define STZ16() writemem(templ>>16,templ,0); writemem(templ>>16,templ+1,0);
#define TRB8() p.z=!(a.b.l&temp); temp&=~a.b.l;
#define TRB16() p.z=!(a.w&tempw); tempw&=~a.w;
#define TSB8() p.z=!(a.b.l&temp); temp|=a.b.l;
#define TSB16() p.z=!(a.w&tempw); tempw|=a.w;

inline void JSR(u8 bank, u16 addr, int l)
{
  pc--;
  if (l) push(pbr);
  push(pc>>8);
  push(pc);
  pbr=bank;
  pc=addr;
}

/*ADC*/
void op69m0() { tempw=immw(); ADC16(); }
void op69m1() { temp=imm();   ADC8();  }
void op65m0() { zp(); tempw=getw(); ADC16(); }
void op65m1() { zp(); temp=getb();  ADC8();  }
void op75m0() { zpx(); tempw=getw(); ADC16(); }
void op75m1() { zpx(); temp=getb();  ADC8();  }
void op72m0() { ind(); tempw=getw(); ADC16(); }
void op72m1() { ind(); temp=getb();  ADC8();  }
void op61m0() { indx(); tempw=getw(); ADC16(); }
void op61m1() { indx(); temp=getb();  ADC8();  }
void op71m0() { indy(); tempw=getw(); ADC16(); }
void op71m1() { indy(); temp=getb();  ADC8();  }
void op67m0() { indl(); tempw=getw(); ADC16(); }
void op67m1() { indl(); temp=getb();  ADC8();  }
void op77m0() { indly(); tempw=getw(); ADC16(); }
void op77m1() { indly(); temp=getb();  ADC8();  }
void op6Dm0() { abs(); tempw=getw(); ADC16(); }
void op6Dm1() { abs(); temp=getb();  ADC8();  }
void op7Dm0() { absx(); tempw=getw(); ADC16(); }
void op7Dm1() { absx(); temp=getb();  ADC8();  }
void op79m0() { absy(); tempw=getw(); ADC16(); }
void op79m1() { absy(); temp=getb();  ADC8();  }
void op6Fm0() { along(); tempw=getw(); ADC16(); }
void op6Fm1() { along(); temp=getb();  ADC8();  }
void op7Fm0() { longx(); tempw=getw(); ADC16(); }
void op7Fm1() { longx(); temp=getb();  ADC8();  }
void op63m0() { sr(); tempw=getw(); ADC16(); }
void op63m1() { sr(); temp=getb();  ADC8();  }
void op73m0() { indsy(); tempw=getw(); ADC16(); }
void op73m1() { indsy(); temp=getb();  ADC8();  }

/*AND*/
void op29m0() { tempw=immw(); AND16(); }
void op29m1() { temp=imm();   AND8();  }
void op25m0() { zp(); tempw=getw(); AND16(); }
void op25m1() { zp(); temp=getb();  AND8();  }
void op35m0() { zpx(); tempw=getw(); AND16(); }
void op35m1() { zpx(); temp=getb();  AND8();  }
void op32m0() { ind(); tempw=getw(); AND16(); }
void op32m1() { ind(); temp=getb();  AND8();  }
void op21m0() { indx(); tempw=getw(); AND16(); }
void op21m1() { indx(); temp=getb();  AND8();  }
void op31m0() { indy(); tempw=getw(); AND16(); }
void op31m1() { indy(); temp=getb();  AND8();  }
void op27m0() { indl(); tempw=getw(); AND16(); }
void op27m1() { indl(); temp=getb();  AND8();  }
void op37m0() { indly(); tempw=getw(); AND16(); }
void op37m1() { indly(); temp=getb();  AND8();  }
void op2Dm0() { abs(); tempw=getw(); AND16(); }
void op2Dm1() { abs(); temp=getb();  AND8();  }
void op3Dm0() { absx(); tempw=getw(); AND16(); }
void op3Dm1() { absx(); temp=getb();  AND8();  }
void op39m0() { absy(); tempw=getw(); AND16(); }
void op39m1() { absy(); temp=getb();  AND8();  }
void op2Fm0() { along(); tempw=getw(); AND16(); }
void op2Fm1() { along(); temp=getb();  AND8();  }
void op3Fm0() { longx(); tempw=getw(); AND16(); }
void op3Fm1() { longx(); temp=getb();  AND8();  }
void op23m0() { sr(); tempw=getw(); AND16(); }
void op23m1() { sr(); temp=getb();  AND8();  }
void op33m0() { indsy(); tempw=getw(); AND16(); }
void op33m1() { indsy(); temp=getb();  AND8();  }

/*ASL*/
void op0Am0() { tempw=a.w;  ASL16(); a.w=tempw; }
void op0Am1() { temp=a.b.l; ASL8();  a.b.l=temp; }
void op06m0() { zp();   tempw=getw(); ASL16(); setw(tempw); }
void op06m1() { zp();   temp=getb();  ASL8();  setb(temp);  }
void op16m0() { zpx();  tempw=getw(); ASL16(); setw(tempw); }
void op16m1() { zpx();  temp=getb();  ASL8();  setb(temp);  }
void op0Em0() { abs();  tempw=getw(); ASL16(); setw(tempw); }
void op0Em1() { abs();  temp=getb();  ASL8();  setb(temp);  }
void op1Em0() { absx(); tempw=getw(); ASL16(); setw(tempw); }
void op1Em1() { absx(); temp=getb();  ASL8();  setb(temp);  }

/*BIT*/
void op89m0() { tempw=immw(); BIT16(); }
void op89m1() { temp=imm();   BIT8(); }
void op24m0() { zp();   tempw=getw(); BIT16(); }
void op24m1() { zp();   temp=getb();  BIT8();  }
void op34m0() { zpx();  tempw=getw(); BIT16(); }
void op34m1() { zpx();  temp=getb();  BIT8();  }
void op2Cm0() { abs();  tempw=getw(); BIT16(); }
void op2Cm1() { abs();  temp=getb();  BIT8();  }
void op3Cm0() { absx(); tempw=getw(); BIT16(); }
void op3Cm1() { absx(); temp=getb();  BIT8();  }

/*CMP*/
void opC9m0() { tempw=immw(); CMP16(); }
void opC9m1() { temp=imm();   CMP8();  }
void opC5m0() { zp(); tempw=getw(); CMP16(); }
void opC5m1() { zp(); temp=getb();  CMP8();  }
void opD5m0() { zpx(); tempw=getw(); CMP16(); }
void opD5m1() { zpx(); temp=getb();  CMP8();  }
void opD2m0() { ind(); tempw=getw(); CMP16(); }
void opD2m1() { ind(); temp=getb();  CMP8();  }
void opC1m0() { indx(); tempw=getw(); CMP16(); }
void opC1m1() { indx(); temp=getb();  CMP8();  }
void opD1m0() { indy(); tempw=getw(); CMP16(); }
void opD1m1() { indy(); temp=getb();  CMP8();  }
void opC7m0() { indl(); tempw=getw(); CMP16(); }
void opC7m1() { indl(); temp=getb();  CMP8();  }
void opD7m0() { indly(); tempw=getw(); CMP16(); }
void opD7m1() { indly(); temp=getb();  CMP8();  }
void opCDm0() { abs(); tempw=getw(); CMP16(); }
void opCDm1() { abs(); temp=getb();  CMP8();  }
void opDDm0() { absx(); tempw=getw(); CMP16(); }
void opDDm1() { absx(); temp=getb();  CMP8();  }
void opD9m0() { absy(); tempw=getw(); CMP16(); }
void opD9m1() { absy(); temp=getb();  CMP8();  }
void opCFm0() { along(); tempw=getw(); CMP16(); }
void opCFm1() { along(); temp=getb();  CMP8();  }
void opDFm0() { longx(); tempw=getw(); CMP16(); }
void opDFm1() { longx(); temp=getb();  CMP8();  }
void opC3m0() { sr(); tempw=getw(); CMP16(); }
void opC3m1() { sr(); temp=getb();  CMP8();  }
void opD3m0() { indsy(); tempw=getw(); CMP16(); }
void opD3m1() { indsy(); temp=getb();  CMP8();  }

/*CPX*/
void opE0x0() { tempw=immw(); CPX16(); }
void opE0x1() { temp=imm();   CPX8(); }
void opE4x0() { zp(); tempw=getw(); CPX16(); }
void opE4x1() { zp(); temp=getb();  CPX8(); }
void opECx0() { abs(); tempw=getw(); CPX16(); }
void opECx1() { abs(); temp=getb();  CPX8(); }

/*CPY*/
void opC0x0() { tempw=immw(); CPY16(); }
void opC0x1() { temp=imm();   CPY8(); }
void opC4x0() { zp(); tempw=getw(); CPY16(); }
void opC4x1() { zp(); temp=getb();  CPY8(); }
void opCCx0() { abs(); tempw=getw(); CPY16(); }
void opCCx1() { abs(); temp=getb();  CPY8(); }

/*DEC*/
void op3Am0() { tempw=a.w;  DEC16(); a.w=tempw; }
void op3Am1() { temp=a.b.l; DEC8();  a.b.l=temp; }
void opC6m0() { zp();   tempw=getw(); DEC16(); setw(tempw); }
void opC6m1() { zp();   temp=getb();  DEC8();  setb(temp);  }
void opD6m0() { zpx();  tempw=getw(); DEC16(); setw(tempw); }
void opD6m1() { zpx();  temp=getb();  DEC8();  setb(temp);  }
void opCEm0() { abs();  tempw=getw(); DEC16(); setw(tempw); }
void opCEm1() { abs();  temp=getb();  DEC8();  setb(temp);  }
void opDEm0() { absx(); tempw=getw(); DEC16(); setw(tempw); }
void opDEm1() { absx(); temp=getb();  DEC8();  setb(temp);  }

/*EOR*/
void op49m0() { tempw=immw(); EOR16(); }
void op49m1() { temp=imm();   EOR8();  }
void op45m0() { zp(); tempw=getw(); EOR16(); }
void op45m1() { zp(); temp=getb();  EOR8();  }
void op55m0() { zpx(); tempw=getw(); EOR16(); }
void op55m1() { zpx(); temp=getb();  EOR8();  }
void op52m0() { ind(); tempw=getw(); EOR16(); }
void op52m1() { ind(); temp=getb();  EOR8();  }
void op41m0() { indx(); tempw=getw(); EOR16(); }
void op41m1() { indx(); temp=getb();  EOR8();  }
void op51m0() { indy(); tempw=getw(); EOR16(); }
void op51m1() { indy(); temp=getb();  EOR8();  }
void op47m0() { indl(); tempw=getw(); EOR16(); }
void op47m1() { indl(); temp=getb();  EOR8();  }
void op57m0() { indly(); tempw=getw(); EOR16(); }
void op57m1() { indly(); temp=getb();  EOR8();  }
void op4Dm0() { abs(); tempw=getw(); EOR16(); }
void op4Dm1() { abs(); temp=getb();  EOR8();  }
void op5Dm0() { absx(); tempw=getw(); EOR16(); }
void op5Dm1() { absx(); temp=getb();  EOR8();  }
void op59m0() { absy(); tempw=getw(); EOR16(); }
void op59m1() { absy(); temp=getb();  EOR8();  }
void op4Fm0() { along(); tempw=getw(); EOR16(); }
void op4Fm1() { along(); temp=getb();  EOR8();  }
void op5Fm0() { longx(); tempw=getw(); EOR16(); }
void op5Fm1() { longx(); temp=getb();  EOR8();  }
void op43m0() { sr(); tempw=getw(); EOR16(); }
void op43m1() { sr(); temp=getb();  EOR8();  }
void op53m0() { indsy(); tempw=getw(); EOR16(); }
void op53m1() { indsy(); temp=getb();  EOR8();  }

/*INC*/
void op1Am0() { tempw=a.w;  INC16(); a.w=tempw; }
void op1Am1() { temp=a.b.l; INC8();  a.b.l=temp; }
void opE6m0() { zp();   tempw=getw(); INC16(); setw(tempw); }
void opE6m1() { zp();   temp=getb();  INC8();  setb(temp);  }
void opF6m0() { zpx();  tempw=getw(); INC16(); setw(tempw); }
void opF6m1() { zpx();  temp=getb();  INC8();  setb(temp);  }
void opEEm0() { abs();  tempw=getw(); INC16(); setw(tempw); }
void opEEm1() { abs();  temp=getb();  INC8();  setb(temp);  }
void opFEm0() { absx(); tempw=getw(); INC16(); setw(tempw); }
void opFEm1() { absx(); temp=getb();  INC8();  setb(temp);  }

/*LDA*/
void opA9m0() { tempw=immw(); LDA16(); }
void opA9m1() { temp=imm();   LDA8();  }
void opA5m0() { zp(); tempw=getw(); LDA16(); }
void opA5m1() { zp(); temp=getb();  LDA8();  }
void opB5m0() { zpx(); tempw=getw(); LDA16(); }
void opB5m1() { zpx(); temp=getb();  LDA8();  }
void opB2m0() { ind(); tempw=getw(); LDA16(); }
void opB2m1() { ind(); temp=getb();  LDA8();  }
void opA1m0() { indx(); tempw=getw(); LDA16(); }
void opA1m1() { indx(); temp=getb();  LDA8();  }
void opB1m0() { indy(); tempw=getw(); LDA16(); }
void opB1m1() { indy(); temp=getb();  LDA8();  }
void opA7m0() { indl(); tempw=getw(); LDA16(); }
void opA7m1() { indl(); temp=getb();  LDA8();  }
void opB7m0() { indly(); tempw=getw(); LDA16(); }
void opB7m1() { indly(); temp=getb();  LDA8();  }
void opADm0() { abs(); tempw=getw(); LDA16(); }
void opADm1() { abs(); temp=getb();  LDA8();  }
void opBDm0() { absx(); tempw=getw(); LDA16(); }
void opBDm1() { absx(); temp=getb();  LDA8();  }
void opB9m0() { absy(); tempw=getw(); LDA16(); }
void opB9m1() { absy(); temp=getb();  LDA8();  }
void opAFm0() { along(); tempw=getw(); LDA16(); }
void opAFm1() { along(); temp=getb();  LDA8();  }
void opBFm0() { longx(); tempw=getw(); LDA16(); }
void opBFm1() { longx(); temp=getb();  LDA8();  }
void opA3m0() { sr(); tempw=getw(); LDA16(); }
void opA3m1() { sr(); temp=getb();  LDA8();  }
void opB3m0() { indsy(); tempw=getw(); LDA16(); }
void opB3m1() { indsy(); temp=getb();  LDA8();  }

/*LDX*/
void opA2x0() { tempw=immw(); LDX16(); }
void opA2x1() { temp=imm();   LDX8(); }
void opA6x0() { zp();   tempw=getw(); LDX16(); }
void opA6x1() { zp();   temp=getb();  LDX8();  }
void opB6x0() { zpy();  tempw=getw(); LDX16(); }
void opB6x1() { zpy();  temp=getb();  LDX8();  }
void opAEx0() { abs();  tempw=getw(); LDX16(); }
void opAEx1() { abs();  temp=getb();  LDX8();  }
void opBEx0() { absy(); tempw=getw(); LDX16(); }
void opBEx1() { absy(); temp=getb();  LDX8();  }

/*LDY*/
void opA0x0() { tempw=immw(); LDY16(); }
void opA0x1() { temp=imm();   LDY8(); }
void opA4x0() { zp();   tempw=getw(); LDY16(); }
void opA4x1() { zp();   temp=getb();  LDY8();  }
void opB4x0() { zpx();  tempw=getw(); LDY16(); }
void opB4x1() { zpx();  temp=getb();  LDY8();  }
void opACx0() { abs();  tempw=getw(); LDY16(); }
void opACx1() { abs();  temp=getb();  LDY8();  }
void opBCx0() { absx(); tempw=getw(); LDY16(); }
void opBCx1() { absx(); temp=getb();  LDY8();  }

/*LSR*/
void op4Am0() { tempw=a.w;  LSR16(); a.w=tempw; }
void op4Am1() { temp=a.b.l; LSR8();  a.b.l=temp; }
void op46m0() { zp();   tempw=getw(); LSR16(); setw(tempw); }
void op46m1() { zp();   temp=getb();  LSR8();  setb(temp);  }
void op56m0() { zpx();  tempw=getw(); LSR16(); setw(tempw); }
void op56m1() { zpx();  temp=getb();  LSR8();  setb(temp);  }
void op4Em0() { abs();  tempw=getw(); LSR16(); setw(tempw); }
void op4Em1() { abs();  temp=getb();  LSR8();  setb(temp);  }
void op5Em0() { absx(); tempw=getw(); LSR16(); setw(tempw); }
void op5Em1() { absx(); temp=getb();  LSR8();  setb(temp);  }

/*ORA*/
void op09m0() { tempw=immw(); ORA16(); }
void op09m1() { temp=imm();   ORA8();  }
void op05m0() { zp(); tempw=getw(); ORA16(); }
void op05m1() { zp(); temp=getb();  ORA8();  }
void op15m0() { zpx(); tempw=getw(); ORA16(); }
void op15m1() { zpx(); temp=getb();  ORA8();  }
void op12m0() { ind(); tempw=getw(); ORA16(); }
void op12m1() { ind(); temp=getb();  ORA8();  }
void op01m0() { indx(); tempw=getw(); ORA16(); }
void op01m1() { indx(); temp=getb();  ORA8();  }
void op11m0() { indy(); tempw=getw(); ORA16(); }
void op11m1() { indy(); temp=getb();  ORA8();  }
void op07m0() { indl(); tempw=getw(); ORA16(); }
void op07m1() { indl(); temp=getb();  ORA8();  }
void op17m0() { indly(); tempw=getw(); ORA16(); }
void op17m1() { indly(); temp=getb();  ORA8();  }
void op0Dm0() { abs(); tempw=getw(); ORA16(); }
void op0Dm1() { abs(); temp=getb();  ORA8();  }
void op1Dm0() { absx(); tempw=getw(); ORA16(); }
void op1Dm1() { absx(); temp=getb();  ORA8();  }
void op19m0() { absy(); tempw=getw(); ORA16(); }
void op19m1() { absy(); temp=getb();  ORA8();  }
void op0Fm0() { along(); tempw=getw(); ORA16(); }
void op0Fm1() { along(); temp=getb();  ORA8();  }
void op1Fm0() { longx(); tempw=getw(); ORA16(); }
void op1Fm1() { longx(); temp=getb();  ORA8();  }
void op03m0() { sr(); tempw=getw(); ORA16(); }
void op03m1() { sr(); temp=getb();  ORA8();  }
void op13m0() { indsy(); tempw=getw(); ORA16(); }
void op13m1() { indsy(); temp=getb();  ORA8();  }

/*ROL*/
void op2Am0() { tempw=a.w;  ROL16(); a.w=tempw; }
void op2Am1() { temp=a.b.l; ROL8();  a.b.l=temp; }
void op26m0() { zp();   tempw=getw(); ROL16(); setw(tempw); }
void op26m1() { zp();   temp=getb();  ROL8();  setb(temp);  }
void op36m0() { zpx();  tempw=getw(); ROL16(); setw(tempw); }
void op36m1() { zpx();  temp=getb();  ROL8();  setb(temp);  }
void op2Em0() { abs();  tempw=getw(); ROL16(); setw(tempw); }
void op2Em1() { abs();  temp=getb();  ROL8();  setb(temp);  }
void op3Em0() { absx(); tempw=getw(); ROL16(); setw(tempw); }
void op3Em1() { absx(); temp=getb();  ROL8();  setb(temp);  }

/*ROR*/
void op6Am0() { tempw=a.w;  ROR16(); a.w=tempw; }
void op6Am1() { temp=a.b.l; ROR8();  a.b.l=temp; }
void op66m0() { zp();   tempw=getw(); ROR16(); setw(tempw); }
void op66m1() { zp();   temp=getb();  ROR8();  setb(temp);  }
void op76m0() { zpx();  tempw=getw(); ROR16(); setw(tempw); }
void op76m1() { zpx();  temp=getb();  ROR8();  setb(temp);  }
void op6Em0() { abs();  tempw=getw(); ROR16(); setw(tempw); }
void op6Em1() { abs();  temp=getb();  ROR8();  setb(temp);  }
void op7Em0() { absx(); tempw=getw(); ROR16(); setw(tempw); }
void op7Em1() { absx(); temp=getb();  ROR8();  setb(temp);  }

/*SBC*/
void opE9m0() { tempw=immw(); SBC16(); }
void opE9m1() { temp=imm();   SBC8();  }
void opE5m0() { zp(); tempw=getw(); SBC16(); }
void opE5m1() { zp(); temp=getb();  SBC8();  }
void opF5m0() { zpx(); tempw=getw(); SBC16(); }
void opF5m1() { zpx(); temp=getb();  SBC8();  }
void opF2m0() { ind(); tempw=getw(); SBC16(); }
void opF2m1() { ind(); temp=getb();  SBC8();  }
void opE1m0() { indx(); tempw=getw(); SBC16(); }
void opE1m1() { indx(); temp=getb();  SBC8();  }
void opF1m0() { indy(); tempw=getw(); SBC16(); }
void opF1m1() { indy(); temp=getb();  SBC8();  }
void opE7m0() { indl(); tempw=getw(); SBC16(); }
void opE7m1() { indl(); temp=getb();  SBC8();  }
void opF7m0() { indly(); tempw=getw(); SBC16(); }
void opF7m1() { indly(); temp=getb();  SBC8();  }
void opEDm0() { abs(); tempw=getw(); SBC16(); }
void opEDm1() { abs(); temp=getb();  SBC8();  }
void opFDm0() { absx(); tempw=getw(); SBC16(); }
void opFDm1() { absx(); temp=getb();  SBC8();  }
void opF9m0() { absy(); tempw=getw(); SBC16(); }
void opF9m1() { absy(); temp=getb();  SBC8();  }
void opEFm0() { along(); tempw=getw(); SBC16(); }
void opEFm1() { along(); temp=getb();  SBC8();  }
void opFFm0() { longx(); tempw=getw(); SBC16(); }
void opFFm1() { longx(); temp=getb();  SBC8();  }
void opE3m0() { sr(); tempw=getw(); SBC16(); }
void opE3m1() { sr(); temp=getb();  SBC8();  }
void opF3m0() { indsy(); tempw=getw(); SBC16(); }
void opF3m1() { indsy(); temp=getb();  SBC8();  }

/*STA*/
void op85m0() { zp(); STA16(); }
void op85m1() { zp(); STA8();  }
void op95m0() { zpx(); STA16(); }
void op95m1() { zpx(); STA8();  }
void op92m0() { ind(); STA16(); }
void op92m1() { ind(); STA8();  }
void op81m0() { indx(); STA16(); }
void op81m1() { indx(); STA8();  }
void op91m0() { indy(); STA16(); }
void op91m1() { indy(); STA8();  }
void op87m0() { indl(); STA16(); }
void op87m1() { indl(); STA8();  }
void op97m0() { indly(); STA16(); }
void op97m1() { indly(); STA8();  }
void op8Dm0() { abs(); STA16(); }
void op8Dm1() { abs(); STA8();  }
void op9Dm0() { absx(); STA16(); }
void op9Dm1() { absx(); STA8();  }
void op99m0() { absy(); STA16(); }
void op99m1() { absy(); STA8();  }
void op8Fm0() { along(); STA16(); }
void op8Fm1() { along(); STA8();  }
void op9Fm0() { longx(); STA16(); }
void op9Fm1() { longx(); STA8();  }
void op83m0() { sr(); STA16(); }
void op83m1() { sr(); STA8();  }
void op93m0() { indsy(); STA16(); }
void op93m1() { indsy(); STA8();  }

/*STX*/
void op86x0() { zp(); STX16(); }
void op86x1() { zp(); STX8();  }
void op96x0() { zpy(); STX16(); }
void op96x1() { zpy(); STX8();  }
void op8Ex0() { abs(); STX16(); }
void op8Ex1() { abs(); STX8();  }

/*STY*/
void op84x0() { zp(); STY16(); }
void op84x1() { zp(); STY8();  }
void op94x0() { zpx(); STY16(); }
void op94x1() { zpx(); STY8();  }
void op8Cx0() { abs(); STY16(); }
void op8Cx1() { abs(); STY8();  }

/*STZ*/
void op64m0() { zp(); STZ16(); }
void op64m1() { zp(); STZ8();  }
void op74m0() { zpx(); STZ16(); }
void op74m1() { zpx(); STZ8();  }
void op9Cm0() { abs(); STZ16(); }
void op9Cm1() { abs(); STZ8();  }
void op9Em0() { absx(); STZ16(); }
void op9Em1() { absx(); STZ8();  }

/*TRB*/
void op14m0() { zp(); TRB16(); }
void op14m1() { zp(); TRB8();  }
void op1Cm0() { abs(); TRB16(); }
void op1Cm1() { abs(); TRB8();  }

/*TSB*/
void op04m0() { zp(); TSB16(); }
void op04m1() { zp(); TSB8();  }
void op0Cm0() { abs(); TSB16(); }
void op0Cm1() { abs(); TSB8();  }

void op18() { p.c=0; }
void opD8() { p.d=0; }
void op58() { p.i=0; }
void opB8() { p.v=0; }
void op38() { p.c=1; }
void opF8() { p.d=1; }
void op78() { p.i=1; }

void opCAx0() { x.w--;   setzn16(x.w); }
void opCAx1() { x.b.l--; setzn8(x.b.l); }
void op88x0() { y.w--;   setzn16(y.w); }
void op88x1() { y.b.l--; setzn8(y.b.l); }
void opE8x0() { x.w++;   setzn16(x.w); }
void opE8x1() { x.b.l++; setzn8(x.b.l); }
void opC8x0() { y.w++;   setzn16(y.w); }
void opC8x1() { y.b.l++; setzn8(y.b.l); }
void opEA() { }

void op48m0() { push(a.b.h); push(a.b.l); }
void op48m1() { push(a.b.l); }
void op8B()   { push(dbr); }
void op0B()   { push(d>>8); push(d&0xFF); }
void op4B()   { push(pbr); }
void op08()   { PHP(); }
void opDAx0() { push(x.b.h); push(x.b.l); }
void opDAx1() { push(x.b.l); }
void op5Ax0() { push(y.b.h); push(y.b.l); }
void op5Ax1() { push(y.b.l); }

void op68m0() { a.b.l=pull(); a.b.h=pull(); setzn16(a.w); }
void op68m1() { a.b.l=pull(); setzn8(a.b.l); }
void opAB()   { dbr=pull(); }
void op2B()   { d=pull(); d|=(pull()<<8); }
void op28()   { PLP(); }
void opFAx0() { x.b.l=pull(); x.b.h=pull(); setzn16(x.w); }
void opFAx1() { x.b.l=pull(); setzn8(x.b.l); }
void op7Ax0() { y.b.l=pull(); y.b.h=pull(); setzn16(y.w); }
void op7Ax1() { y.b.l=pull(); setzn8(y.b.l); }

void opAAx0() { x.w=a.w; setzn16(x.w); }
void opAAx1() { x.b.l=a.b.l; setzn8(x.b.l); }
void opA8x0() { y.w=a.w; setzn16(y.w); }
void opA8x1() { y.b.l=a.b.l; setzn8(y.b.l); }
void op9Bx0() { y.w=x.w; setzn16(y.w); }
void op9Bx1() { y.b.l=x.b.l; setzn8(y.b.l); }
void opBBx0() { x.w=y.w; setzn16(x.w); }
void opBBx1() { x.b.l=y.b.l; setzn8(x.b.l); }
void opBA()   { x.w=s.w; }
void op5B()   { d=a.w; }
void op7B()   { a.w=d; }
void op1B()   { s.w=a.w; }
void op3B()   { a.w=s.w; }
void op8Am0() { a.w=x.w; setzn16(a.w); }
void op8Am1() { a.b.l=x.b.l; setzn8(a.b.l); }
void op98m0() { a.w=y.w; setzn16(a.w); }
void op98m1() { a.b.l=y.b.l; setzn8(a.b.l); }
void op9A()   { s.w=x.w; if (p.e) s.b.h=1; }

void opFB()   { tempi=p.e; p.e=p.c; p.c=tempi; updatemode(); }

void opDC()   { indl();  pc=templ&0xFFFF; pbr=templ>>16; }
void op4C()   { abs();   pc=templ&0xFFFF; }
void op6C()   { ind();   pc=templ&0xFFFF; }
void op7C()   { indx();  pc=templ&0xFFFF; }
void op5C()   { along(); pc=templ&0xFFFF; pbr=templ>>16; }

void op22()   { along(); JSR(templ>>16,templ,1); }
void op20()   { abs();   JSR(pbr,templ,0); }
void opFC()   { indx();  JSR(pbr,templ,0); }

void op60()   { pc=pull(); pc|=(pull()<<8); pc++; }
void op6B()   { pc=pull(); pc|=(pull()<<8); pbr=pull(); pc++; }
void op40()   { PLP(); pc=pull(); pc|=(pull()<<8); if (!p.e) pbr=pull(); }

void opEB()   { a.w=(a.w>>8)|(a.w<<8); setzn8(a.b.l); }
void opCB()   { cycles=0; }

void opC2() /*REP*/
{
  temp=imm();
  if (temp&1)    p.c=0;
  if (temp&2)    p.z=0;
  if (temp&4)    p.i=0;
  if (temp&8)    p.d=0;
  if (!p.e && temp&0x10) p.x=0;
  if (!p.e && temp&0x20) p.m=0;
  if (temp&0x40) p.v=0;
  if (temp&0x80) p.n=0;
  updatemode();
}

void opE2() /*SEP*/
{
  temp=imm();
  if (temp&1)    p.c=1;
  if (temp&2)    p.z=1;
  if (temp&4)    p.i=1;
  if (temp&8)    p.d=1;
  if (!p.e && temp&0x10) p.x=1;
  if (!p.e && temp&0x20) p.m=1;
  if (temp&0x40) p.v=1;
  if (temp&0x80) p.n=1;
  updatemode();
}

/*Bxx*/
void op90() { offset=(s8)readfast(pbr,pc); pc++; if (!p.c) pc+=offset; }
void opB0() { offset=(s8)readfast(pbr,pc); pc++; if (p.c)  pc+=offset; }
void op10() { offset=(s8)readfast(pbr,pc); pc++; if (!p.n) pc+=offset; }
void op30() { offset=(s8)readfast(pbr,pc); pc++; if (p.n)  pc+=offset; }
void op50() { offset=(s8)readfast(pbr,pc); pc++; if (!p.v) pc+=offset; }
void op70() { offset=(s8)readfast(pbr,pc); pc++; if (p.v)  pc+=offset; }
void opD0() { offset=(s8)readfast(pbr,pc); pc++; if (!p.z) pc+=offset; }
void opF0() { offset=(s8)readfast(pbr,pc); pc++; if (p.z)  pc+=offset; }
void op80() { offset=(s8)readfast(pbr,pc); pc++; pc+=offset; }
void op82() { tempw=immw(); pc+=tempw; }

void opF4() { tempw=immw(); push(tempw>>8); push(tempw); }
void opD4() { ind();  push(templ>>8); push(templ); }

void op54()
{
  dbr=readfast(pbr,pc);
  templ=readfast(pbr,pc+1); pc+=2;
  temp=readmem(templ,x.w++);
  writemem(dbr,y.w++,temp);
  a.w--;
  if (a.w!=0xFFFF)
    pc-=3;
}

void maketables()
{
  int c,d;
  for (d=0;d<4;d++)
    for (c=0;c<256;c++)
      opcoderout[d][c]=badopcode;
  opcoderout[0][0x4C]=opcoderout[1][0x4C]=opcoderout[2][0x4C]=opcoderout[3][0x4C]=op4C;
  opcoderout[0][0x5C]=opcoderout[1][0x5C]=opcoderout[2][0x5C]=opcoderout[3][0x5C]=op5C;
  opcoderout[0][0x6C]=opcoderout[1][0x6C]=opcoderout[2][0x6C]=opcoderout[3][0x6C]=op6C;
  opcoderout[0][0x7C]=opcoderout[1][0x7C]=opcoderout[2][0x7C]=opcoderout[3][0x7C]=op7C;
  opcoderout[0][0xDC]=opcoderout[1][0xDC]=opcoderout[2][0xDC]=opcoderout[3][0xDC]=opDC;
  opcoderout[0][0x22]=opcoderout[1][0x22]=opcoderout[2][0x22]=opcoderout[3][0x22]=op22;
  opcoderout[0][0x20]=opcoderout[1][0x20]=opcoderout[2][0x20]=opcoderout[3][0x20]=op20;
  opcoderout[0][0xFC]=opcoderout[1][0xFC]=opcoderout[2][0xFC]=opcoderout[3][0xFC]=opFC;
  opcoderout[0][0x40]=opcoderout[1][0x40]=opcoderout[2][0x40]=opcoderout[3][0x40]=op40;
  opcoderout[0][0x4B]=opcoderout[1][0x4B]=opcoderout[2][0x4B]=opcoderout[3][0x4B]=op4B;
  opcoderout[0][0x60]=opcoderout[1][0x60]=opcoderout[2][0x60]=opcoderout[3][0x60]=op60;
  opcoderout[0][0x18]=opcoderout[1][0x18]=opcoderout[2][0x18]=opcoderout[3][0x18]=op18;
  opcoderout[0][0xD8]=opcoderout[1][0xD8]=opcoderout[2][0xD8]=opcoderout[3][0xD8]=opD8;
  opcoderout[0][0x58]=opcoderout[1][0x58]=opcoderout[2][0x58]=opcoderout[3][0x58]=op58;
  opcoderout[0][0xB8]=opcoderout[1][0xB8]=opcoderout[2][0xB8]=opcoderout[3][0xB8]=opB8;
  opcoderout[0][0x8B]=opcoderout[1][0x8B]=opcoderout[2][0x8B]=opcoderout[3][0x8B]=op8B;
  opcoderout[0][0xCB]=opcoderout[1][0xCB]=opcoderout[2][0xCB]=opcoderout[3][0xCB]=opCB;
  opcoderout[0][0xEB]=opcoderout[1][0xEB]=opcoderout[2][0xEB]=opcoderout[3][0xEB]=opEB;
  opcoderout[0][0xEA]=opcoderout[1][0xEA]=opcoderout[2][0xEA]=opcoderout[3][0xEA]=opEA;
  opcoderout[0][0x0B]=opcoderout[1][0x0B]=opcoderout[2][0x0B]=opcoderout[3][0x0B]=op0B;
  opcoderout[0][0x4B]=opcoderout[1][0x4B]=opcoderout[2][0x4B]=opcoderout[3][0x4B]=op4B;
  opcoderout[0][0x08]=opcoderout[1][0x08]=opcoderout[2][0x08]=opcoderout[3][0x08]=op08;
  opcoderout[0][0xAB]=opcoderout[1][0xAB]=opcoderout[2][0xAB]=opcoderout[3][0xAB]=opAB;
  opcoderout[0][0x2B]=opcoderout[1][0x2B]=opcoderout[2][0x2B]=opcoderout[3][0x2B]=op2B;
  opcoderout[0][0x6B]=opcoderout[1][0x6B]=opcoderout[2][0x6B]=opcoderout[3][0x6B]=op6B;
  opcoderout[0][0x28]=opcoderout[1][0x28]=opcoderout[2][0x28]=opcoderout[3][0x28]=op28;
  opcoderout[0][0x38]=opcoderout[1][0x38]=opcoderout[2][0x38]=opcoderout[3][0x38]=op38;
  opcoderout[0][0xF8]=opcoderout[1][0xF8]=opcoderout[2][0xF8]=opcoderout[3][0xF8]=opF8;
  opcoderout[0][0x78]=opcoderout[1][0x78]=opcoderout[2][0x78]=opcoderout[3][0x78]=op78;
  opcoderout[0][0x7B]=opcoderout[1][0x7B]=opcoderout[2][0x7B]=opcoderout[3][0x7B]=op7B;
  opcoderout[0][0x1B]=opcoderout[1][0x1B]=opcoderout[2][0x1B]=opcoderout[3][0x1B]=op1B;
  opcoderout[0][0x3B]=opcoderout[1][0x3B]=opcoderout[2][0x3B]=opcoderout[3][0x3B]=op3B;
  opcoderout[0][0x5B]=opcoderout[1][0x5B]=opcoderout[2][0x5B]=opcoderout[3][0x5B]=op5B;
  opcoderout[0][0x9A]=opcoderout[1][0X9A]=opcoderout[2][0x9A]=opcoderout[3][0x9A]=op9A;
  opcoderout[0][0xBA]=opcoderout[1][0xBA]=opcoderout[2][0xBA]=opcoderout[3][0xBA]=opBA;
  opcoderout[0][0xC2]=opcoderout[1][0xC2]=opcoderout[2][0xC2]=opcoderout[3][0xC2]=opC2;
  opcoderout[0][0xE2]=opcoderout[1][0xE2]=opcoderout[2][0xE2]=opcoderout[3][0xE2]=opE2;
  opcoderout[0][0xFB]=opcoderout[1][0xFB]=opcoderout[2][0xFB]=opcoderout[3][0xFB]=opFB;
  opcoderout[0][0x82]=opcoderout[1][0x82]=opcoderout[2][0x82]=opcoderout[3][0x82]=op82;
  opcoderout[0][0xF4]=opcoderout[1][0xF4]=opcoderout[2][0xF4]=opcoderout[3][0xF4]=opF4;
  opcoderout[0][0xD4]=opcoderout[1][0xD4]=opcoderout[2][0xD4]=opcoderout[3][0xD4]=opD4;
  
  opcoderout[0][0x54]=opcoderout[1][0x54]=opcoderout[2][0x54]=opcoderout[3][0x54]=op54;
  
  opcoderout[0][0x10]=opcoderout[1][0x10]=opcoderout[2][0x10]=opcoderout[3][0x10]=op10;
  opcoderout[0][0x30]=opcoderout[1][0x30]=opcoderout[2][0x30]=opcoderout[3][0x30]=op30;
  opcoderout[0][0x50]=opcoderout[1][0x50]=opcoderout[2][0x50]=opcoderout[3][0x50]=op50;
  opcoderout[0][0x70]=opcoderout[1][0x70]=opcoderout[2][0x70]=opcoderout[3][0x70]=op70;
  opcoderout[0][0x80]=opcoderout[1][0x80]=opcoderout[2][0x80]=opcoderout[3][0x80]=op80;
  opcoderout[0][0x90]=opcoderout[1][0x90]=opcoderout[2][0x90]=opcoderout[3][0x90]=op90;
  opcoderout[0][0xB0]=opcoderout[1][0xB0]=opcoderout[2][0xB0]=opcoderout[3][0xB0]=opB0;
  opcoderout[0][0xD0]=opcoderout[1][0xD0]=opcoderout[2][0xD0]=opcoderout[3][0xD0]=opD0;
  opcoderout[0][0xF0]=opcoderout[1][0xF0]=opcoderout[2][0xF0]=opcoderout[3][0xF0]=opF0;

  opcoderout[0][0x69]=opcoderout[1][0x69]=op69m0;
  opcoderout[2][0x69]=opcoderout[3][0x69]=op69m1;
  opcoderout[0][0x65]=opcoderout[1][0x65]=op65m0;
  opcoderout[2][0x65]=opcoderout[3][0x65]=op65m1;
  opcoderout[0][0x75]=opcoderout[1][0x75]=op75m0;
  opcoderout[2][0x75]=opcoderout[3][0x75]=op75m1;
  opcoderout[0][0x72]=opcoderout[1][0x72]=op72m0;
  opcoderout[2][0x72]=opcoderout[3][0x72]=op72m1;
  opcoderout[0][0x61]=opcoderout[1][0x61]=op61m0;
  opcoderout[2][0x61]=opcoderout[3][0x61]=op61m1;
  opcoderout[0][0x71]=opcoderout[1][0x71]=op71m0;
  opcoderout[2][0x71]=opcoderout[3][0x71]=op71m1;
  opcoderout[0][0x67]=opcoderout[1][0x67]=op67m0;
  opcoderout[2][0x67]=opcoderout[3][0x67]=op67m1;
  opcoderout[0][0x77]=opcoderout[1][0x77]=op77m0;
  opcoderout[2][0x77]=opcoderout[3][0x77]=op77m1;
  opcoderout[0][0x6D]=opcoderout[1][0x6D]=op6Dm0;
  opcoderout[2][0x6D]=opcoderout[3][0x6D]=op6Dm1;
  opcoderout[0][0x7D]=opcoderout[1][0x7D]=op7Dm0;
  opcoderout[2][0x7D]=opcoderout[3][0x7D]=op7Dm1;
  opcoderout[0][0x79]=opcoderout[1][0x79]=op79m0;
  opcoderout[2][0x79]=opcoderout[3][0x79]=op79m1;
  opcoderout[0][0x6F]=opcoderout[1][0x6F]=op6Fm0;
  opcoderout[2][0x6F]=opcoderout[3][0x6F]=op6Fm1;
  opcoderout[0][0x7F]=opcoderout[1][0x7F]=op7Fm0;
  opcoderout[2][0x7F]=opcoderout[3][0x7F]=op7Fm1;
  opcoderout[0][0x63]=opcoderout[1][0x63]=op63m0;
  opcoderout[2][0x63]=opcoderout[3][0x63]=op63m1;
  opcoderout[0][0x73]=opcoderout[1][0x73]=op73m0;
  opcoderout[2][0x73]=opcoderout[3][0x73]=op73m1;

  opcoderout[0][0x29]=opcoderout[1][0x29]=op29m0;
  opcoderout[2][0x29]=opcoderout[3][0x29]=op29m1;
  opcoderout[0][0x25]=opcoderout[1][0x25]=op25m0;
  opcoderout[2][0x25]=opcoderout[3][0x25]=op25m1;
  opcoderout[0][0x35]=opcoderout[1][0x35]=op35m0;
  opcoderout[2][0x35]=opcoderout[3][0x35]=op35m1;
  opcoderout[0][0x32]=opcoderout[1][0x32]=op32m0;
  opcoderout[2][0x32]=opcoderout[3][0x32]=op32m1;
  opcoderout[0][0x21]=opcoderout[1][0x21]=op21m0;
  opcoderout[2][0x21]=opcoderout[3][0x21]=op21m1;
  opcoderout[0][0x31]=opcoderout[1][0x31]=op31m0;
  opcoderout[2][0x31]=opcoderout[3][0x31]=op31m1;
  opcoderout[0][0x27]=opcoderout[1][0x27]=op27m0;
  opcoderout[2][0x27]=opcoderout[3][0x27]=op27m1;
  opcoderout[0][0x37]=opcoderout[1][0x37]=op37m0;
  opcoderout[2][0x37]=opcoderout[3][0x37]=op37m1;
  opcoderout[0][0x2D]=opcoderout[1][0x2D]=op2Dm0;
  opcoderout[2][0x2D]=opcoderout[3][0x2D]=op2Dm1;
  opcoderout[0][0x3D]=opcoderout[1][0x3D]=op3Dm0;
  opcoderout[2][0x3D]=opcoderout[3][0x3D]=op3Dm1;
  opcoderout[0][0x39]=opcoderout[1][0x39]=op39m0;
  opcoderout[2][0x39]=opcoderout[3][0x39]=op39m1;
  opcoderout[0][0x2F]=opcoderout[1][0x2F]=op2Fm0;
  opcoderout[2][0x2F]=opcoderout[3][0x2F]=op2Fm1;
  opcoderout[0][0x3F]=opcoderout[1][0x3F]=op3Fm0;
  opcoderout[2][0x3F]=opcoderout[3][0x3F]=op3Fm1;
  opcoderout[0][0x23]=opcoderout[1][0x23]=op23m0;
  opcoderout[2][0x23]=opcoderout[3][0x23]=op23m1;
  opcoderout[0][0x33]=opcoderout[1][0x33]=op33m0;
  opcoderout[2][0x33]=opcoderout[3][0x33]=op33m1;
  
  opcoderout[0][0x0A]=opcoderout[1][0x0A]=op0Am0;
  opcoderout[2][0x0A]=opcoderout[3][0x0A]=op0Am1;
  opcoderout[0][0x06]=opcoderout[1][0x06]=op06m0;
  opcoderout[2][0x06]=opcoderout[3][0x06]=op06m1;
  opcoderout[0][0x16]=opcoderout[1][0x16]=op16m0;
  opcoderout[2][0x16]=opcoderout[3][0x16]=op16m1;
  opcoderout[0][0x0E]=opcoderout[1][0x0E]=op0Em0;
  opcoderout[2][0x0E]=opcoderout[3][0x0E]=op0Em1;
  opcoderout[0][0x1E]=opcoderout[1][0x1E]=op1Em0;
  opcoderout[2][0x1E]=opcoderout[3][0x1E]=op1Em1;
  
  opcoderout[0][0x89]=opcoderout[1][0x89]=op89m0;
  opcoderout[2][0x89]=opcoderout[3][0x89]=op89m1;
  opcoderout[0][0x24]=opcoderout[1][0x24]=op24m0;
  opcoderout[2][0x24]=opcoderout[3][0x24]=op24m1;
  opcoderout[0][0x34]=opcoderout[1][0x34]=op34m0;
  opcoderout[2][0x34]=opcoderout[3][0x34]=op34m1;
  opcoderout[0][0x2C]=opcoderout[1][0x2C]=op2Cm0;
  opcoderout[2][0x2C]=opcoderout[3][0x2C]=op2Cm1;
  opcoderout[0][0x3C]=opcoderout[1][0x3C]=op3Cm0;
  opcoderout[2][0x3C]=opcoderout[3][0x3C]=op3Cm1;
  
  opcoderout[0][0xC9]=opcoderout[1][0xC9]=opC9m0;
  opcoderout[2][0xC9]=opcoderout[3][0xC9]=opC9m1;
  opcoderout[0][0xC5]=opcoderout[1][0xC5]=opC5m0;
  opcoderout[2][0xc5]=opcoderout[3][0xC5]=opC5m1;
  opcoderout[0][0xD5]=opcoderout[1][0xD5]=opD5m0;
  opcoderout[2][0xD5]=opcoderout[3][0xD5]=opD5m1;
  opcoderout[0][0xD2]=opcoderout[1][0xD2]=opD2m0;
  opcoderout[2][0xD2]=opcoderout[3][0xD2]=opD2m1;
  opcoderout[0][0xC1]=opcoderout[1][0xC1]=opC1m0;
  opcoderout[2][0xC1]=opcoderout[3][0xC1]=opC1m1;
  opcoderout[0][0xD1]=opcoderout[1][0xD1]=opD1m0;
  opcoderout[2][0xD1]=opcoderout[3][0xD1]=opD1m1;
  opcoderout[0][0xC7]=opcoderout[1][0xC7]=opC7m0;
  opcoderout[2][0xC7]=opcoderout[3][0xC7]=opC7m1;
  opcoderout[0][0xD7]=opcoderout[1][0xD7]=opD7m0;
  opcoderout[2][0xD7]=opcoderout[3][0xD7]=opD7m1;
  opcoderout[0][0xCD]=opcoderout[1][0xCD]=opCDm0;
  opcoderout[2][0xCD]=opcoderout[3][0xCD]=opCDm1;
  opcoderout[0][0xDD]=opcoderout[1][0xDD]=opDDm0;
  opcoderout[2][0xDD]=opcoderout[3][0xDD]=opDDm1;
  opcoderout[0][0xD9]=opcoderout[1][0xD9]=opD9m0;
  opcoderout[2][0xD9]=opcoderout[3][0xD9]=opD9m1;
  opcoderout[0][0xCF]=opcoderout[1][0xCF]=opCFm0;
  opcoderout[2][0xCF]=opcoderout[3][0xCF]=opCFm1;
  opcoderout[0][0xDF]=opcoderout[1][0xDF]=opDFm0;
  opcoderout[2][0xDF]=opcoderout[3][0xDF]=opDFm1;
  opcoderout[0][0xC3]=opcoderout[1][0xC3]=opC3m0;
  opcoderout[2][0xC3]=opcoderout[3][0xC3]=opC3m1;
  opcoderout[0][0xD3]=opcoderout[1][0xD3]=opD3m0;
  opcoderout[2][0xD3]=opcoderout[3][0xD3]=opD3m1;

  opcoderout[0][0xE0]=opcoderout[2][0xE0]=opE0x0;
  opcoderout[1][0xE0]=opcoderout[3][0xE0]=opE0x1;
  opcoderout[0][0xE4]=opcoderout[2][0xE4]=opE4x0;
  opcoderout[1][0xE4]=opcoderout[3][0xE4]=opE4x1;
  opcoderout[0][0xEC]=opcoderout[2][0xEC]=opECx0;
  opcoderout[1][0xEC]=opcoderout[3][0xEC]=opECx1;

  opcoderout[0][0xC0]=opcoderout[2][0xC0]=opC0x0;
  opcoderout[1][0xC0]=opcoderout[3][0xC0]=opC0x1;
  opcoderout[0][0xC4]=opcoderout[2][0xC4]=opC4x0;
  opcoderout[1][0xC4]=opcoderout[3][0xC4]=opC4x1;
  opcoderout[0][0xCC]=opcoderout[2][0xCC]=opCCx0;
  opcoderout[1][0xCC]=opcoderout[3][0xCC]=opCCx1;

  opcoderout[0][0x3A]=opcoderout[1][0x3A]=op3Am0;
  opcoderout[2][0x3A]=opcoderout[3][0x3A]=op3Am1;
  opcoderout[0][0xC6]=opcoderout[1][0xC6]=opC6m0;
  opcoderout[2][0xC6]=opcoderout[3][0xC6]=opC6m1;
  opcoderout[0][0xD6]=opcoderout[1][0xD6]=opD6m0;
  opcoderout[2][0xD6]=opcoderout[3][0xD6]=opD6m1;
  opcoderout[0][0xCE]=opcoderout[1][0xCE]=opCEm0;
  opcoderout[2][0xCE]=opcoderout[3][0xCE]=opCEm1;
  opcoderout[0][0xDE]=opcoderout[1][0xDE]=opDEm0;
  opcoderout[2][0xDE]=opcoderout[3][0xDE]=opDEm1;

  opcoderout[0][0x49]=opcoderout[1][0x49]=op49m0;
  opcoderout[2][0x49]=opcoderout[3][0x49]=op49m1;
  opcoderout[0][0x45]=opcoderout[1][0x45]=op45m0;
  opcoderout[2][0x45]=opcoderout[3][0x45]=op45m1;
  opcoderout[0][0x55]=opcoderout[1][0x55]=op55m0;
  opcoderout[2][0x55]=opcoderout[3][0x55]=op55m1;
  opcoderout[0][0x52]=opcoderout[1][0x52]=op52m0;
  opcoderout[2][0x52]=opcoderout[3][0x52]=op52m1;
  opcoderout[0][0x41]=opcoderout[1][0x41]=op41m0;
  opcoderout[2][0x41]=opcoderout[3][0x41]=op41m1;
  opcoderout[0][0x51]=opcoderout[1][0x51]=op51m0;
  opcoderout[2][0x51]=opcoderout[3][0x51]=op51m1;
  opcoderout[0][0x47]=opcoderout[1][0x47]=op47m0;
  opcoderout[2][0x47]=opcoderout[3][0x47]=op47m1;
  opcoderout[0][0x57]=opcoderout[1][0x57]=op57m0;
  opcoderout[2][0x57]=opcoderout[3][0x57]=op57m1;
  opcoderout[0][0x4D]=opcoderout[1][0x4D]=op4Dm0;
  opcoderout[2][0x4D]=opcoderout[3][0x4D]=op4Dm1;
  opcoderout[0][0x5D]=opcoderout[1][0x5D]=op5Dm0;
  opcoderout[2][0x5D]=opcoderout[3][0x5D]=op5Dm1;
  opcoderout[0][0x59]=opcoderout[1][0x59]=op59m0;
  opcoderout[2][0x59]=opcoderout[3][0x59]=op59m1;
  opcoderout[0][0x4F]=opcoderout[1][0x4F]=op4Fm0;
  opcoderout[2][0x4F]=opcoderout[3][0x4F]=op4Fm1;
  opcoderout[0][0x5F]=opcoderout[1][0x5F]=op5Fm0;
  opcoderout[2][0x5F]=opcoderout[3][0x5F]=op5Fm1;
  opcoderout[0][0x43]=opcoderout[1][0x43]=op43m0;
  opcoderout[2][0x43]=opcoderout[3][0x43]=op43m1;
  opcoderout[0][0x53]=opcoderout[1][0x53]=op53m0;
  opcoderout[2][0x53]=opcoderout[3][0x53]=op53m1;
  
  opcoderout[0][0x1A]=opcoderout[1][0x1A]=op1Am0;
  opcoderout[2][0x1A]=opcoderout[3][0x1A]=op1Am1;
  opcoderout[0][0xE6]=opcoderout[1][0xE6]=opE6m0;
  opcoderout[2][0xE6]=opcoderout[3][0xE6]=opE6m1;
  opcoderout[0][0xF6]=opcoderout[1][0xF6]=opF6m0;
  opcoderout[2][0xF6]=opcoderout[3][0xF6]=opF6m1;
  opcoderout[0][0xEE]=opcoderout[1][0xEE]=opEEm0;
  opcoderout[2][0xEE]=opcoderout[3][0xEE]=opEEm1;
  opcoderout[0][0xFE]=opcoderout[1][0xFE]=opFEm0;
  opcoderout[2][0xFE]=opcoderout[3][0xFE]=opFEm1;
  
  opcoderout[0][0xA9]=opcoderout[1][0xA9]=opA9m0;
  opcoderout[2][0xA9]=opcoderout[3][0xA9]=opA9m1;
  opcoderout[0][0xA5]=opcoderout[1][0xA5]=opA5m0;
  opcoderout[2][0xA5]=opcoderout[3][0xA5]=opA5m1;
  opcoderout[0][0xB5]=opcoderout[1][0xB5]=opB5m0;
  opcoderout[2][0xB5]=opcoderout[3][0xB5]=opB5m1;
  opcoderout[0][0xB2]=opcoderout[1][0xB2]=opB2m0;
  opcoderout[2][0xB2]=opcoderout[3][0xB2]=opB2m1;
  opcoderout[0][0xA1]=opcoderout[1][0xA1]=opA1m0;
  opcoderout[2][0xA1]=opcoderout[3][0xA1]=opA1m1;
  opcoderout[0][0xB1]=opcoderout[1][0xB1]=opB1m0;
  opcoderout[2][0xB1]=opcoderout[3][0xB1]=opB1m1;
  opcoderout[0][0xA7]=opcoderout[1][0xA7]=opA7m0;
  opcoderout[2][0xA7]=opcoderout[3][0xA7]=opA7m1;
  opcoderout[0][0xB7]=opcoderout[1][0xB7]=opB7m0;
  opcoderout[2][0xB7]=opcoderout[3][0xB7]=opB7m1;
  opcoderout[0][0xAD]=opcoderout[1][0xAD]=opADm0;
  opcoderout[2][0xAD]=opcoderout[3][0xAD]=opADm1;
  opcoderout[0][0xBD]=opcoderout[1][0xBD]=opBDm0;
  opcoderout[2][0xBD]=opcoderout[3][0xBD]=opBDm1;
  opcoderout[0][0xB9]=opcoderout[1][0xB9]=opB9m0;
  opcoderout[2][0xB9]=opcoderout[3][0xB9]=opB9m1;
  opcoderout[0][0xAF]=opcoderout[1][0xAF]=opAFm0;
  opcoderout[2][0xAF]=opcoderout[3][0xAF]=opAFm1;
  opcoderout[0][0xBF]=opcoderout[1][0xBF]=opBFm0;
  opcoderout[2][0xBF]=opcoderout[3][0xBF]=opBFm1;
  opcoderout[0][0xA3]=opcoderout[1][0xA3]=opA3m0;
  opcoderout[2][0xA3]=opcoderout[3][0xA3]=opA3m1;
  opcoderout[0][0xB3]=opcoderout[1][0xB3]=opB3m0;
  opcoderout[2][0xB3]=opcoderout[3][0xB3]=opB3m1;
  
  opcoderout[0][0xA2]=opcoderout[2][0xA2]=opA2x0;
  opcoderout[1][0xA2]=opcoderout[3][0xA2]=opA2x1;
  opcoderout[0][0xA6]=opcoderout[2][0xA6]=opA6x0;
  opcoderout[1][0xA6]=opcoderout[3][0xA6]=opA6x1;
  opcoderout[0][0xB6]=opcoderout[2][0xB6]=opB6x0;
  opcoderout[1][0xB6]=opcoderout[3][0xB6]=opB6x1;
  opcoderout[0][0xAE]=opcoderout[2][0xAE]=opAEx0;
  opcoderout[1][0xAE]=opcoderout[3][0xAE]=opAEx1;
  opcoderout[0][0xBE]=opcoderout[2][0xBE]=opBEx0;
  opcoderout[1][0xBE]=opcoderout[3][0xBE]=opBEx1;
  
  opcoderout[0][0xA0]=opcoderout[2][0xA0]=opA0x0;
  opcoderout[1][0xA0]=opcoderout[3][0xA0]=opA0x1;
  opcoderout[0][0xA4]=opcoderout[2][0xA4]=opA4x0;
  opcoderout[1][0xA4]=opcoderout[3][0xA4]=opA4x1;
  opcoderout[0][0xB4]=opcoderout[2][0xB4]=opB4x0;
  opcoderout[1][0xB4]=opcoderout[3][0xB4]=opB4x1;
  opcoderout[0][0xAC]=opcoderout[2][0xAC]=opACx0;
  opcoderout[1][0xAC]=opcoderout[3][0xAC]=opACx1;
  opcoderout[0][0xBC]=opcoderout[2][0xBC]=opBCx0;
  opcoderout[1][0xBC]=opcoderout[3][0xBC]=opBCx1;
  
  opcoderout[0][0x4A]=opcoderout[1][0x4A]=op4Am0;
  opcoderout[2][0x4A]=opcoderout[3][0x4A]=op4Am1;
  opcoderout[0][0x46]=opcoderout[1][0x46]=op46m0;
  opcoderout[2][0x46]=opcoderout[3][0x46]=op46m1;
  opcoderout[0][0x56]=opcoderout[1][0x56]=op56m0;
  opcoderout[2][0x56]=opcoderout[3][0x56]=op56m1;
  opcoderout[0][0x4E]=opcoderout[1][0x4E]=op4Em0;
  opcoderout[2][0x4E]=opcoderout[3][0x4E]=op4Em1;
  opcoderout[0][0x5E]=opcoderout[1][0x5E]=op5Em0;
  opcoderout[2][0x5E]=opcoderout[3][0x5E]=op5Em1;
  
  opcoderout[0][0x09]=opcoderout[1][0x09]=op09m0;
  opcoderout[2][0x09]=opcoderout[3][0x09]=op09m1;
  opcoderout[0][0x05]=opcoderout[1][0x05]=op05m0;
  opcoderout[2][0x05]=opcoderout[3][0x05]=op05m1;
  opcoderout[0][0x15]=opcoderout[1][0x15]=op15m0;
  opcoderout[2][0x15]=opcoderout[3][0x15]=op15m1;
  opcoderout[0][0x12]=opcoderout[1][0x12]=op12m0;
  opcoderout[2][0x12]=opcoderout[3][0x12]=op12m1;
  opcoderout[0][0x01]=opcoderout[1][0x01]=op01m0;
  opcoderout[2][0x01]=opcoderout[3][0x01]=op01m1;
  opcoderout[0][0x11]=opcoderout[1][0x11]=op11m0;
  opcoderout[2][0x11]=opcoderout[3][0x11]=op11m1;
  opcoderout[0][0x07]=opcoderout[1][0x07]=op07m0;
  opcoderout[2][0x07]=opcoderout[3][0x07]=op07m1;
  opcoderout[0][0x17]=opcoderout[1][0x17]=op17m0;
  opcoderout[2][0x17]=opcoderout[3][0x17]=op17m1;
  opcoderout[0][0x0D]=opcoderout[1][0x0D]=op0Dm0;
  opcoderout[2][0x0D]=opcoderout[3][0x0D]=op0Dm1;
  opcoderout[0][0x1D]=opcoderout[1][0x1D]=op1Dm0;
  opcoderout[2][0x1D]=opcoderout[3][0x1D]=op1Dm1;
  opcoderout[0][0x19]=opcoderout[1][0x19]=op19m0;
  opcoderout[2][0x19]=opcoderout[3][0x19]=op19m1;
  opcoderout[0][0x0F]=opcoderout[1][0x0F]=op0Fm0;
  opcoderout[2][0x0F]=opcoderout[3][0x0F]=op0Fm1;
  opcoderout[0][0x1F]=opcoderout[1][0x1F]=op1Fm0;
  opcoderout[2][0x1F]=opcoderout[3][0x1F]=op1Fm1;
  opcoderout[0][0x03]=opcoderout[1][0x03]=op03m0;
  opcoderout[2][0x03]=opcoderout[3][0x03]=op03m1;
  opcoderout[0][0x13]=opcoderout[1][0x13]=op13m0;
  opcoderout[2][0x13]=opcoderout[3][0x13]=op13m1;
  
  opcoderout[0][0x2A]=opcoderout[1][0x2A]=op2Am0;
  opcoderout[2][0x2A]=opcoderout[3][0x2A]=op2Am1;
  opcoderout[0][0x26]=opcoderout[1][0x26]=op26m0;
  opcoderout[2][0x26]=opcoderout[3][0x26]=op26m1;
  opcoderout[0][0x36]=opcoderout[1][0x36]=op36m0;
  opcoderout[2][0x36]=opcoderout[3][0x36]=op36m1;
  opcoderout[0][0x2E]=opcoderout[1][0x2E]=op2Em0;
  opcoderout[2][0x2E]=opcoderout[3][0x2E]=op2Em1;
  opcoderout[0][0x3E]=opcoderout[1][0x3E]=op3Em0;
  opcoderout[2][0x3E]=opcoderout[3][0x3E]=op3Em1;
  
  opcoderout[0][0x6A]=opcoderout[1][0x6A]=op6Am0;
  opcoderout[2][0x6A]=opcoderout[3][0x6A]=op6Am1;
  opcoderout[0][0x66]=opcoderout[1][0x66]=op66m0;
  opcoderout[2][0x66]=opcoderout[3][0x66]=op66m1;
  opcoderout[0][0x76]=opcoderout[1][0x76]=op76m0;
  opcoderout[2][0x76]=opcoderout[3][0x76]=op76m1;
  opcoderout[0][0x6E]=opcoderout[1][0x6E]=op6Em0;
  opcoderout[2][0x6E]=opcoderout[3][0x6E]=op6Em1;
  opcoderout[0][0x7E]=opcoderout[1][0x7E]=op7Em0;
  opcoderout[2][0x7E]=opcoderout[3][0x7E]=op7Em1;
  
  opcoderout[0][0xE9]=opcoderout[1][0xE9]=opE9m0;
  opcoderout[2][0xE9]=opcoderout[3][0xE9]=opE9m1;
  opcoderout[0][0xE5]=opcoderout[1][0xE5]=opE5m0;
  opcoderout[2][0xE5]=opcoderout[3][0xE5]=opE5m1;
  opcoderout[0][0xF5]=opcoderout[1][0xF5]=opF5m0;
  opcoderout[2][0xF5]=opcoderout[3][0xF5]=opF5m1;
  opcoderout[0][0xF2]=opcoderout[1][0xF2]=opF2m0;
  opcoderout[2][0xF2]=opcoderout[3][0xF2]=opF2m1;
  opcoderout[0][0xE1]=opcoderout[1][0xE1]=opE1m0;
  opcoderout[2][0xE1]=opcoderout[3][0xE1]=opE1m1;
  opcoderout[0][0xF1]=opcoderout[1][0xF1]=opF1m0;
  opcoderout[2][0xF1]=opcoderout[3][0xF1]=opF1m1;
  opcoderout[0][0xE7]=opcoderout[1][0xE7]=opE7m0;
  opcoderout[2][0xE7]=opcoderout[3][0xE7]=opE7m1;
  opcoderout[0][0xF7]=opcoderout[1][0xF7]=opF7m0;
  opcoderout[2][0xF7]=opcoderout[3][0xF7]=opF7m1;
  opcoderout[0][0xED]=opcoderout[1][0xED]=opEDm0;
  opcoderout[2][0xED]=opcoderout[3][0xED]=opEDm1;
  opcoderout[0][0xFD]=opcoderout[1][0xFD]=opFDm0;
  opcoderout[2][0xFD]=opcoderout[3][0xFD]=opFDm1;
  opcoderout[0][0xF9]=opcoderout[1][0xF9]=opF9m0;
  opcoderout[2][0xF9]=opcoderout[3][0xF9]=opF9m1;
  opcoderout[0][0xEF]=opcoderout[1][0xEF]=opEFm0;
  opcoderout[2][0xEF]=opcoderout[3][0xEF]=opEFm1;
  opcoderout[0][0xFF]=opcoderout[1][0xFF]=opFFm0;
  opcoderout[2][0xFF]=opcoderout[3][0xFF]=opFFm1;
  opcoderout[0][0xE3]=opcoderout[1][0xE3]=opE3m0;
  opcoderout[2][0xE3]=opcoderout[3][0xE3]=opE3m1;
  opcoderout[0][0xF3]=opcoderout[1][0xF3]=opF3m0;
  opcoderout[2][0xF3]=opcoderout[3][0xF3]=opF3m1;
  
  opcoderout[0][0x89]=opcoderout[1][0x89]=op89m0;
  opcoderout[2][0x89]=opcoderout[3][0x89]=op89m1;
  opcoderout[0][0x85]=opcoderout[1][0x85]=op85m0;
  opcoderout[2][0x85]=opcoderout[3][0x85]=op85m1;
  opcoderout[0][0x95]=opcoderout[1][0x95]=op95m0;
  opcoderout[2][0x95]=opcoderout[3][0x95]=op95m1;
  opcoderout[0][0x92]=opcoderout[1][0x92]=op92m0;
  opcoderout[2][0x92]=opcoderout[3][0x92]=op92m1;
  opcoderout[0][0x81]=opcoderout[1][0x81]=op81m0;
  opcoderout[2][0x81]=opcoderout[3][0x81]=op81m1;
  opcoderout[0][0x91]=opcoderout[1][0x91]=op91m0;
  opcoderout[2][0x91]=opcoderout[3][0x91]=op91m1;
  opcoderout[0][0x87]=opcoderout[1][0x87]=op87m0;
  opcoderout[2][0x87]=opcoderout[3][0x87]=op87m1;
  opcoderout[0][0x97]=opcoderout[1][0x97]=op97m0;
  opcoderout[2][0x97]=opcoderout[3][0x97]=op97m1;
  opcoderout[0][0x8D]=opcoderout[1][0x8D]=op8Dm0;
  opcoderout[2][0x8D]=opcoderout[3][0x8D]=op8Dm1;
  opcoderout[0][0x9D]=opcoderout[1][0x9D]=op9Dm0;
  opcoderout[2][0x9D]=opcoderout[3][0x9D]=op9Dm1;
  opcoderout[0][0x99]=opcoderout[1][0x99]=op99m0;
  opcoderout[2][0x99]=opcoderout[3][0x99]=op99m1;
  opcoderout[0][0x8F]=opcoderout[1][0x8F]=op8Fm0;
  opcoderout[2][0x8F]=opcoderout[3][0x8F]=op8Fm1;
  opcoderout[0][0x9F]=opcoderout[1][0x9F]=op9Fm0;
  opcoderout[2][0x9F]=opcoderout[3][0x9F]=op9Fm1;
  opcoderout[0][0x83]=opcoderout[1][0x83]=op83m0;
  opcoderout[2][0x83]=opcoderout[3][0x83]=op83m1;
  opcoderout[0][0x93]=opcoderout[1][0x93]=op93m0;
  opcoderout[2][0x93]=opcoderout[3][0x93]=op93m1;
  
  opcoderout[0][0x86]=opcoderout[2][0x86]=op86x0;
  opcoderout[1][0x86]=opcoderout[3][0x86]=op86x1;
  opcoderout[0][0x96]=opcoderout[2][0x96]=op96x0;
  opcoderout[1][0x96]=opcoderout[3][0x96]=op96x1;
  opcoderout[0][0x8E]=opcoderout[2][0x8E]=op8Ex0;
  opcoderout[1][0x8E]=opcoderout[3][0x8E]=op8Ex1;
  
  opcoderout[0][0x84]=opcoderout[2][0x84]=op84x0;
  opcoderout[1][0x84]=opcoderout[3][0x84]=op84x1;
  opcoderout[0][0x94]=opcoderout[2][0x94]=op94x0;
  opcoderout[1][0x94]=opcoderout[3][0x94]=op94x1;
  opcoderout[0][0x8C]=opcoderout[2][0x8C]=op8Cx0;
  opcoderout[1][0x8C]=opcoderout[3][0x8C]=op8Cx1;

  opcoderout[0][0x64]=opcoderout[1][0x64]=op64m0;
  opcoderout[2][0x64]=opcoderout[3][0x64]=op64m1;
  opcoderout[0][0x74]=opcoderout[1][0x74]=op74m0;
  opcoderout[2][0x74]=opcoderout[3][0x74]=op74m1;
  opcoderout[0][0x9C]=opcoderout[1][0x9C]=op9Cm0;
  opcoderout[2][0x9C]=opcoderout[3][0x9C]=op9Cm1;
  opcoderout[0][0x9E]=opcoderout[1][0x9E]=op9Em0;
  opcoderout[2][0x9E]=opcoderout[3][0x9E]=op9Em1;
  
  opcoderout[0][0x14]=opcoderout[1][0x14]=op14m0;
  opcoderout[2][0x14]=opcoderout[3][0x14]=op14m1;
  opcoderout[0][0x1C]=opcoderout[1][0x1C]=op1Cm0;
  opcoderout[2][0x1C]=opcoderout[3][0x1C]=op1Cm1;
  opcoderout[0][0x04]=opcoderout[1][0x04]=op04m0;
  opcoderout[2][0x04]=opcoderout[3][0x04]=op04m1;
  opcoderout[0][0x0C]=opcoderout[1][0x0C]=op0Cm0;
  opcoderout[2][0x0C]=opcoderout[3][0x0C]=op0Cm1;

  opcoderout[0][0x48]=opcoderout[1][0x48]=op48m0;
  opcoderout[2][0x48]=opcoderout[3][0x48]=op48m1;
  opcoderout[0][0x68]=opcoderout[1][0x68]=op68m0;
  opcoderout[2][0x68]=opcoderout[3][0x68]=op68m1;

  opcoderout[0][0xDA]=opcoderout[2][0xDA]=opDAx0;
  opcoderout[1][0xDA]=opcoderout[3][0xDA]=opDAx1;
  opcoderout[0][0x5A]=opcoderout[2][0x5A]=op5Ax0;
  opcoderout[1][0x5A]=opcoderout[3][0x5A]=op5Ax1;
  opcoderout[0][0xFA]=opcoderout[2][0xFA]=opFAx0;
  opcoderout[1][0xFA]=opcoderout[3][0xFA]=opFAx1;
  opcoderout[0][0x7A]=opcoderout[2][0x7A]=op7Ax0;
  opcoderout[1][0x7A]=opcoderout[3][0x7A]=op7Ax1;

  opcoderout[0][0xAA]=opcoderout[2][0xAA]=opAAx0;
  opcoderout[1][0xAA]=opcoderout[3][0xAA]=opAAx1;
  opcoderout[0][0xA8]=opcoderout[2][0xA8]=opA8x0;
  opcoderout[1][0xA8]=opcoderout[3][0xA8]=opA8x1;
  opcoderout[0][0x9B]=opcoderout[2][0x9B]=op9Bx0;
  opcoderout[1][0x9B]=opcoderout[3][0x9B]=op9Bx1;
  opcoderout[0][0xBB]=opcoderout[2][0xBB]=opBBx0;
  opcoderout[1][0xBB]=opcoderout[3][0xBB]=opBBx1;
  
  opcoderout[0][0x8A]=opcoderout[1][0x8A]=op8Am0;
  opcoderout[2][0x8A]=opcoderout[3][0x8A]=op8Am1;
  opcoderout[0][0x98]=opcoderout[1][0x98]=op98m0;
  opcoderout[2][0x98]=opcoderout[3][0x98]=op98m1;
  
  opcoderout[0][0xCA]=opcoderout[2][0xCA]=opCAx0;
  opcoderout[1][0xCA]=opcoderout[3][0xCA]=opCAx1;
  opcoderout[0][0x88]=opcoderout[2][0x88]=op88x0;
  opcoderout[1][0x88]=opcoderout[3][0x88]=op88x1;
  opcoderout[0][0xE8]=opcoderout[2][0xE8]=opE8x0;
  opcoderout[1][0xE8]=opcoderout[3][0xE8]=opE8x1;
  opcoderout[0][0xC8]=opcoderout[2][0xC8]=opC8x0;
  opcoderout[1][0xC8]=opcoderout[3][0xC8]=opC8x1;

  /*Make memory pointers. I hope this works...*/
  for (c=0;c<0x40;c++)
  {
    mempointers[c][1]=rom[c];
    mempointers[c][0]=ram;
  }
  for (c=0;c<0x40;c++) /*I'm not sure about this at all. The way I've*/
  {                    /*emulated it allows 6mb ROMs, which is obviously wrong*/
    mempointers[c+0x40][0]=rom[0x40+(c<<1)];
    mempointers[c+0x40][1]=rom[0x41+(c<<1)];
  }
  mempointers[0x7E][0]=ram;
  mempointers[0x7E][1]=ram+0x8000;
  mempointers[0x7F][0]=ram+0x10000;
  mempointers[0x7F][1]=ram+0x18000;
  for (c=0;c<0x80;c++)
  {
    mempointers[c+0x80][0]=mempointers[c][0];
    mempointers[c+0x80][1]=mempointers[c][1];
  }
}

void reset65816()
{
  maketables();
  pc=readfast(0,0xFFFC)|(readfast(0,0xFFFD)<<8);
  p.i=p.e=p.x=p.m=1;
  smode=3;
  pbr=dbr=0;
//        dumpregs();
  s.w=0x1FF;
  atexit(dumpram);
}

void int65816()
{
  if (p.i)
    return;
  inint=1;
  if (p.e)
  {
    writemem(0,s.w--,pc>>8);
    writemem(0,s.w--,pc);
    s.b.h=1;
    PHP();
    pbr=0;
    pc=readfast(0,0xFFFE)|(readfast(0,0xFFFF)<<8);
    p.i=1;
    cycles-=4;
  }
  else
  {
    writemem(0,s.w--,pbr);
    writemem(0,s.w--,pc>>8);
    writemem(0,s.w--,pc);
    PHP();
    pbr=0;
    pc=readfast(0,0xFFEE)|(readfast(0,0xFFEF)<<8);
    p.i=1;
    cycles-=5;
  }
}

#if 0
int nmiena=1;
#endif

void nmi65816()
{
#if 0
  if (!nmiena)
#endif
  if (!nmienable)
    return;
  if (p.e)
  {
    writemem(0,s.w--,pc>>8);
    writemem(0,s.w--,pc);
    s.b.h=1;
    PHP();
    pbr=0;
    pc=readfast(0,0xFFFA)|(readfast(0,0xFFFB)<<8);
    p.i=1;
    cycles-=4;
  }
  else
  {
    writemem(0,s.w--,pbr);
    writemem(0,s.w--,pc>>8);
    writemem(0,s.w--,pc);
    PHP();
    pbr=0;
    pc=readfast(0,0xFFEA)|(readfast(0,0xFFEB)<<8);
    p.i=1;
    cycles-=5;
//                output=1;
  }
}

void exec65816(int cycs)
{
  cycles+=cycs;
  while (cycles>0)
  {
    oldoldpc=oldpc;
    oldpc=pc|(pbr<<16);
    if (output) printf("%i : A=%04X X=%04X Y=%04X S=%04X PC=%02X:%04X DBR=%02X D=%04X P=%02X %02X %i %02X %02X %02X\n",instructions,a.w,x.w,y.w,s.w,pbr,pc,dbr,d,compp(),opcode,smode,ram[0],ram[1],ram[0x1590]);
    opcode=readfast(pbr,pc);
    pc++;
    opcoderout[smode][opcode]();
    instructions++;
    
    if (output) loggedins++;
    cycles-=4;
  }
}
