/*
 * InfoSNES - Super Nintendo (TM) emulator
 * Copyrigh(c) 2003 Jay's Factory (jays_factory@excite.co.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*-------------------------------------------------------------------*/
/*  Include Files                                                    */
/*-------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>

#include "type.h"

/*-------------------------------------------------------------------*/
/*  Other Resources                                                  */
/*-------------------------------------------------------------------*/

u8 ram[0x10005];

typedef struct
{
  u8 op;
  int params;
  int mode; /*0=norm, 1=absolute, 2=relative, 3=dp+relative, 4=mem.bit*/
  char *format;
  int last;
} spuop;

spuop ops[]=
{
  {0xE8,1,0,"MOV   A,#%02X",0},
  {0xE6,0,0,"MOV   A,(X)",0},
  {0xBF,0,0,"MOV   A,(X)+",0},
  {0xE4,1,0,"MOV   A,%02X",0},
  {0xF4,1,0,"MOV   A,%02X+X",0},
  {0xE5,2,1,"MOV   A,%04X",0},
  {0xF5,2,1,"MOV   A,%04X+X",0},
  {0xF6,2,1,"MOV   A,%04X+Y",0},
  {0xE7,1,0,"MOV   A,(%02X+X)",0},
  {0xF7,1,0,"MOV   A,(%02X)+Y",0},
  
  {0xCD,1,0,"MOV   X,#%02X",0},
  {0xF8,1,0,"MOV   X,%02X",0},
  {0xF9,1,0,"MOV   X,%02X+Y",0},
  {0xE9,2,1,"MOV   X,%04X",0},
  
  {0x8D,1,0,"MOV   Y,#%02X",0},
  {0xEB,1,0,"MOV   Y,%02X",0},
  {0xFB,1,0,"MOV   Y,%02X+X",0},
  {0xEC,2,1,"MOV   Y,%04X",0},
  
  {0xC6,0,0,"MOV   (X),A",0},
  {0xAF,0,0,"MOV   (X)+,A",0},
  {0xC4,1,0,"MOV   %02X,A",0},
  {0xD4,1,0,"MOV   %02X+X,A",0},
  {0xC5,2,1,"MOV   %04X,A",0},
  {0xD5,2,1,"MOV   %04X+X,A",0},
  {0xD6,2,1,"MOV   %04X+Y,A",0},
  {0xC7,1,0,"MOV   (%02X+X),A",0},
  {0xD7,1,0,"MOV   (%02X)+Y,A",0},
  
  {0xD8,1,0,"MOV   %02X,X",0},
  {0xD9,1,0,"MOV   %02X+Y,X",0},
  {0xC9,2,1,"MOV   %04X,X",0},
  
  {0xCB,1,0,"MOV   %02X,Y",0},
  {0xDB,1,0,"MOV   %02X+X,Y",0},
  {0x0C,2,1,"MOV   %04X,Y",0},
  
  {0x7D,0,0,"MOV   A,X",0},
  {0xDD,0,0,"MOV   A,Y",0},
  {0x5D,0,0,"MOV   X,A",0},
  {0xFD,0,0,"MOV   Y,A",0},
  {0x9D,0,0,"MOV   X,SP",0},
  {0xBD,0,0,"MOV   SP,X",0},
  {0xFA,2,0,"MOV   %02X,%02X",0},
  {0x8F,2,0,"MOV   %02X,#%02X",0},
  
  {0x88,1,0,"ADC   A,#%02X",0},
  {0x86,0,0,"ADC   A,(X)",0},
  {0x84,1,0,"ADC   A,%02X",0},
  {0x94,1,0,"ADC   A,%02X+X",0},
  {0x85,2,1,"ADC   A,%04X",0},
  {0x95,2,1,"ADC   A,%04X+X",0},
  {0x96,2,1,"ADC   A,%04X+Y",0},
  {0x87,1,0,"ADC   A,(%02X+X)",0},
  {0x97,1,0,"ADC   A,(%02X)+Y",0},
  {0x99,0,0,"ADC   (X),(Y)",0},
  {0x89,2,0,"ADC   %02X,%02X",0},
  {0x98,2,0,"ADC   %02X,#%02X",0},
  
  {0xA8,1,0,"SBC   A,#%02X",0},
  {0xA6,0,0,"SBC   A,(X)",0},
  {0xA4,1,0,"SBC   A,%02X",0},
  {0xB4,1,0,"SBC   A,%02X+X",0},
  {0xA5,2,1,"SBC   A,%04X",0},
  {0xB5,2,1,"SBC   A,%04X+X",0},
  {0xB6,2,1,"SBC   A,%04X+Y",0},
  {0xA7,1,0,"SBC   A,(%02X+X)",0},
  {0xB7,1,0,"SBC   A,(%02X)+Y",0},
  {0xB9,0,0,"SBC   (X),(Y)",0},
  {0xA9,2,0,"SBC   %02X,%02X",0},
  {0xB8,2,0,"SBC   %02X,#%02X",0},
  
  {0x68,1,0,"CMP   A,#%02X",0},
  {0x66,0,0,"CMP   A,(X)",0},
  {0x64,1,0,"CMP   A,%02X",0},
  {0x74,1,0,"CMP   A,%02X+X",0},
  {0x65,2,1,"CMP   A,%04X",0},
  {0x75,2,1,"CMP   A,%04X+X",0},
  {0x76,2,1,"CMP   A,%04X+Y",0},
  {0x67,1,0,"CMP   A,(%02X+X)",0},
  {0x77,1,0,"CMP   A,(%02X)+Y",0},
  {0x79,0,0,"CMP   (X),(Y)",0},
  {0x69,2,0,"CMP   %02X,%02X",0},
  {0x78,2,0,"CMP   %02X,#%02X",0},
  
  {0xC8,1,0,"CMP   X,#%02X",0},
  {0x3E,1,0,"CMP   X,%02X",0},
  {0x1E,2,1,"CMP   X,%04X",0},
  
  {0xAD,1,0,"CMP   Y,#%02X",0},
  {0x7E,1,0,"CMP   Y,%02X",0},
  {0x5E,2,1,"CMP   Y,%04X",0},
  
  {0x28,1,0,"AND   A,#%02X",0},
  {0x26,0,0,"AND   A,(X)",0},
  {0x24,1,0,"AND   A,%02X",0},
  {0x34,1,0,"AND   A,%02X+X",0},
  {0x25,2,1,"AND   A,%04X",0},
  {0x35,2,1,"AND   A,%04X+X",0},
  {0x36,2,1,"AND   A,%04X+Y",0},
  {0x27,1,0,"AND   A,(%02X+X)",0},
  {0x37,1,0,"AND   A,(%02X)+Y",0},
  {0x39,0,0,"AND   (X),(Y)",0},
  {0x29,2,0,"AND   %02X,%02X",0},
  {0x38,2,0,"AND   %02X,#%02X",0},
  
  {0x08,1,0,"OR    A,#%02X",0},
  {0x06,0,0,"OR    A,(X)",0},
  {0x04,1,0,"OR    A,%02X",0},
  {0x14,1,0,"OR    A,%02X+X",0},
  {0x05,2,1,"OR    A,%04X",0},
  {0x15,2,1,"OR    A,%04X+X",0},
  {0x16,2,1,"OR    A,%04X+Y",0},
  {0x07,1,0,"OR    A,(%02X+X)",0},
  {0x17,1,0,"OR    A,(%02X)+Y",0},
  {0x19,0,0,"OR    (X),(Y)",0},
  {0x09,2,0,"OR    %02X,%02X",0},
  {0x18,2,0,"OR    %02X,#%02X",0},
  
  {0x48,1,0,"EOR   A,#%02X",0},
  {0x46,0,0,"EOR   A,(X)",0},
  {0x44,1,0,"EOR   A,%02X",0},
  {0x54,1,0,"EOR   A,%02X+X",0},
  {0x45,2,1,"EOR   A,%04X",0},
  {0x55,2,1,"EOR   A,%04X+X",0},
  {0x56,2,1,"EOR   A,%04X+Y",0},
  {0x47,1,0,"EOR   A,(%02X+X)",0},
  {0x57,1,0,"EOR   A,(%02X)+Y",0},
  {0x59,0,0,"EOR   (X),(Y)",0},
  {0x49,2,0,"EOR   %02X,%02X",0},
  {0x58,2,0,"EOR   %02X,#%02X",0},
  
  {0xBC,0,0,"INC   A",0},
  {0xAB,1,0,"INC   %02X",0},
  {0xBB,1,0,"INC   %02X+X",0},
  {0xAC,2,1,"INC   %04X",0},
  {0x3D,0,0,"INC   X",0},
  {0xFC,0,0,"INC   Y",0},
  
  {0x9C,0,0,"DEC   A",0},
  {0x8B,1,0,"DEC   %02X",0},
  {0x9B,1,0,"DEC   %02X+X",0},
  {0x8C,2,1,"DEC   %04X",0},
  {0x1D,0,0,"DEC   X",0},
  {0xDC,0,0,"DEC   Y",0},
  
  {0x1C,0,0,"ASL   A",0},
  {0x0B,1,0,"ASL   %02X",0},
  {0x1B,1,0,"ASL   %02X+X",0},
  {0xCC,2,1,"ASL   %04X",0},
  
  {0x5C,0,0,"LSR   A",0},
  {0x4B,1,0,"LSR   %02X",0},
  {0x5B,1,0,"LSR   %02X+X",0},
  {0x4C,2,1,"LSR   %04X",0},
  
  {0x3C,0,0,"ROL   A",0},
  {0x2B,1,0,"ROL   %02X",0},
  {0x3B,1,0,"ROL   %02X+X",0},
  {0x2C,2,1,"ROL   %04X",0},
  
  {0x7C,0,0,"ROR   A",0},
  {0x6B,1,0,"ROR   %02X",0},
  {0x7B,1,0,"ROR   %02X+X",0},
  {0x6C,2,1,"ROR   %04X",0},
  
  {0xBA,1,0,"MOVW  YA,%02X",0},
  {0xDA,1,0,"MOVW  %02X,YA",0},
  
  {0x3A,1,0,"INCW  %02X",0},
  {0x1A,1,0,"DECW  %02X",0},
  {0x7A,1,0,"ADDW  YA,%02X",0},
  {0x9A,1,0,"SUBW  YA,%02X",0},
  {0x5A,1,0,"CMPW  YA,%02X",0},
  
  {0xCF,0,0,"MUL   YA",0},
  {0x9E,0,0,"DIV   YA,X",0},
  
  {0xDF,0,0,"DAA   A",0},
  {0xBE,0,0,"DAS   A",0},
  
  {0x2F,1,2,"BRA   %04X",0},
  {0xF0,1,2,"BEQ   %04X",0},
  {0xD0,1,2,"BNE   %04X",0},
  {0xB0,1,2,"BCS   %04X",0},
  {0x90,1,2,"BCC   %04X",0},
  {0x70,1,2,"BVS   %04X",0},
  {0x50,1,2,"BVC   %04X",0},
  {0x30,1,2,"BMI   %04X",0},
  {0x10,1,2,"BPL   %04X",0},
  
  {0x03,2,3,"BBS   %02X.0,%04X",0},
  {0x23,2,3,"BBS   %02X.1,%04X",0},
  {0x43,2,3,"BBS   %02X.2,%04X",0},
  {0x63,2,3,"BBS   %02X.3,%04X",0},
  {0x83,2,3,"BBS   %02X.4,%04X",0},
  {0xA3,2,3,"BBS   %02X.5,%04X",0},
  {0xC3,2,3,"BBS   %02X.6,%04X",0},
  {0xE3,2,3,"BBS   %02X.7,%04X",0},
  
  {0x13,2,3,"BBC   %02X.0,%04X",0},
  {0x33,2,3,"BBC   %02X.1,%04X",0},
  {0x53,2,3,"BBC   %02X.2,%04X",0},
  {0x73,2,3,"BBC   %02X.3,%04X",0},
  {0x93,2,3,"BBC   %02X.4,%04X",0},
  {0xB3,2,3,"BBC   %02X.5,%04X",0},
  {0xD3,2,3,"BBC   %02X.6,%04X",0},
  {0xF3,2,3,"BBC   %02X.7,%04X",0},
  
  {0x2E,2,3,"CBNE  %02X,%04X",0},
  {0xDE,2,3,"CBNE  %02X+X,%04X",0},
  {0x6E,2,3,"DBNZ  %02X,%04X",0},
  {0xFE,1,2,"DBNZ  Y,%04X",0},
  
  {0x5F,2,1,"JMP   %04X",0},
  {0x1F,2,1,"JMP   (%04X+X)",0},
  
  {0x3F,2,1,"CALL  %04X",0},
  {0x4F,2,1,"PCALL %04X",0},
  
  {0x01,0,0,"TCALL 0",0},
  {0x11,0,0,"TCALL 1",0},
  {0x21,0,0,"TCALL 2",0},
  {0x31,0,0,"TCALL 3",0},
  {0x41,0,0,"TCALL 4",0},
  {0x51,0,0,"TCALL 5",0},
  {0x61,0,0,"TCALL 6",0},
  {0x71,0,0,"TCALL 7",0},
  {0x81,0,0,"TCALL 8",0},
  {0x91,0,0,"TCALL 9",0},
  {0xA1,0,0,"TCALL A",0},
  {0xB1,0,0,"TCALL B",0},
  {0xC1,0,0,"TCALL C",0},
  {0xD1,0,0,"TCALL D",0},
  {0xE1,0,0,"TCALL E",0},
  {0xF1,0,0,"TCALL F",0},
  
  {0x0F,0,0,"BRK",0},
  {0x6F,0,0,"RET",0},
  {0x7F,0,0,"RET1",0},
  
  {0x2D,0,0,"PUSH  A",0},
  {0x4D,0,0,"PUSH  X",0},
  {0x6D,0,0,"PUSH  Y",0},
  {0x0D,0,0,"PUSH  PSW",0},
  
  {0xAE,0,0,"POP   A",0},
  {0xCE,0,0,"POP   X",0},
  {0xEE,0,0,"POP   Y",0},
  {0x8E,0,0,"POP   PSW",0},
  
  {0x02,1,0,"SET1  %02X.0",0},
  {0x22,1,0,"SET1  %02X.1",0},
  {0x42,1,0,"SET1  %02X.2",0},
  {0x62,1,0,"SET1  %02X.3",0},
  {0x82,1,0,"SET1  %02X.4",0},
  {0xA2,1,0,"SET1  %02X.5",0},
  {0xC2,1,0,"SET1  %02X.6",0},
  {0xE2,1,0,"SET1  %02X.7",0},
  
  {0x12,1,0,"CLR1  %02X.0",0},
  {0x32,1,0,"CLR1  %02X.1",0},
  {0x52,1,0,"CLR1  %02X.2",0},
  {0x72,1,0,"CLR1  %02X.3",0},
  {0x92,1,0,"CLR1  %02X.4",0},
  {0xB2,1,0,"CLR1  %02X.5",0},
  {0xD2,1,0,"CLR1  %02X.6",0},
  {0xF2,1,0,"CLR1  %02X.7",0},
  
  {0x0E,2,1,"TSET1 %04X",0},
  {0x4E,2,1,"TCLR1 %04X",0},

  {0x4A,2,4,"AND1  C,%04X.%i",0},
  {0x6A,2,4,"AND1  C,/%04X.%i",0},
  {0x0A,2,4,"OR1   C,%04X.%i",0},
  {0x2A,2,4,"OR1   C,/%04X.%i",0},
  {0x8A,2,4,"EOR1  C,%04X.%i",0},
  {0xEA,2,4,"NOT1  %04X,%i",0},
  {0xAA,2,4,"MOV1  C,%04X.%i",0},
  {0xCA,2,4,"MOV1  %04X.%i,C",0},
  
  {0x60,0,0,"CLRC",0},
  {0x80,0,0,"SETC",0},
  {0xED,0,0,"NOTC",0},
  {0xE0,0,0,"CLRV",0},
  {0x20,0,0,"CLRP",0},
  {0x40,0,0,"SETP",0},
  {0xA0,0,0,"EI",0},
  {0xC0,0,0,"DI",0},
  
  {0x00,0,0,"NOP",0},
  {0xEF,0,0,"SLEEP",0},
  {0xFF,0,0,"STOP",1}
};

spuop bad={0,1,0, "DB    %02X",0};

/*-------------------------------------------------------------------*/
/*  Functions                                                        */
/*-------------------------------------------------------------------*/

spuop scantable(u8 opcode)
{
  int c=0;
  spuop temp;
  while (1)
  {
    temp=ops[c++];
    if (temp.last && opcode!=temp.op)
      return bad;
    if (opcode==temp.op)
      return temp;
  }
}

int disassemble(int pc)
{
  u8 opcode=ram[pc++];
  spuop temp=scantable(opcode);
  u8 op1,op2;
  u16 opw;
  printf("%04X : ",pc-1);
  switch (temp.params)
  {
  case 0:
    printf("%02X          : ",opcode);
    break;
  case 1:
    printf("%02X %02X       : ",opcode,ram[pc]);
    break;
  case 2:
    printf("%02X %02X %02X    : ",opcode,ram[pc],ram[pc+1]);
    break;
  case 3:
    printf("%02X %02X %02X %02X : ",opcode,ram[pc],ram[pc+1],ram[pc+2]);
    break;
  }
  switch (temp.mode)
  {
  case 0: /*normal*/
    op1=ram[pc];
    op2=ram[pc+1];
    pc+=temp.params;
    if (temp.params==1)
      printf(temp.format,op1);
    else
                   printf(temp.format,op2,op1);
    break;
  case 1: /*absolute*/
    op1=ram[pc];
    op2=ram[pc+1];
    pc+=temp.params;
    printf(temp.format,op1|(op2<<8));
    break;
  case 2: /*relative*/
    op1=ram[pc];
    pc+=temp.params;
    printf(temp.format,pc+(signed char)op1);
    break;
  case 3: /*relative+dp*/
    op1=ram[pc];
    op2=ram[pc+1];
    pc+=temp.params;
    printf(temp.format,op2,pc+(signed char)op1);
    break;
  case 4: /*mem.bit*/
    opw=ram[pc]|(ram[pc+1]<<8);
    pc+=temp.params;
    printf(temp.format,opw&0x1FFF,opw>>13);
    break;
  }
  printf("\n");
  return pc;
}

int main(int argc, char *argv[])
{
  int pc=0;
  FILE *f;
  if (argc<2)
  {
    printf("Syntax : disspc memdmp.bin\n");
    exit(-1);
  }
  f=fopen(argv[1],"rb");
  fread(ram,65536,1,f);
  fclose(f);
  while (pc<0x8000)
    pc=disassemble(pc);
  return 0;
}
