/*
 * InfoSNES - Super Nintendo (TM) emulator
 * Copyrigh(c) 2003 Jay's Factory (jays_factory@excite.co.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*-------------------------------------------------------------------*/
/*  Include Files                                                    */
/*-------------------------------------------------------------------*/

#include <allegro.h>

#include <stdio.h>
#include <stdlib.h>

#include "65816.h"
#include "io.h"
#include "type.h"

/*-------------------------------------------------------------------*/
/*  IO Resources                                                     */
/*-------------------------------------------------------------------*/

u8 hdmaon;
int padready=0;

u8 multa,multb,divb;
u16 multc,diva,divc;

int vertcomp;
int vblank=1;
int nmienable=0;
int intenable=0;
u8 dmactrl[8],dmadest[8];
unsigned long dmasrc[8],dmasize[8];

int dmaout=0;

/*-------------------------------------------------------------------*/
/*  Functions                                                        */
/*-------------------------------------------------------------------*/

u8 readio(u16 addr)
{
  u8 temp;

  if (addr==0x4016 || addr==0x4017)
    return 0xFF;
  switch (addr&0x1FF)
  {
  case 0x0B:
    return 0;
    
  case 0x10:
    if (vblank)
    {
      vblank=0;
      return 0x80;
    }
    return 0;
  case 0x11:
    if (inint)
    {
      inint=0;
      return 0x80;
    }
    return 0;
  case 0x12:
    temp=padready;
    padready^=0x41;
    if (vblank)
      temp|=0x80;
    return temp;
    
  case 0x14:
    return diva&0xFF;
  case 0x15:
    return diva>>8;
  case 0x16:
    return multc&0xFF;
  case 0x17:
    return multc>>8;
    
  case 0x18:
    temp=0;
    if (key[KEY_S]) temp|=0x80;
    if (key[KEY_A]) temp|=0x40;
    if (key[KEY_Q]) temp|=0x20;
    if (key[KEY_W]) temp|=0x10;
    return temp;
  case 0x19:
    temp=0;
    if (key[KEY_X]) temp|=0x80;
    if (key[KEY_Z]) temp|=0x40;
    if (key[KEY_D]) temp|=0x20;
    if (key[KEY_C]) temp|=0x10;
    if (key[KEY_UP]) temp|=0x8;
    if (key[KEY_DOWN]) temp|=0x4;
    if (key[KEY_LEFT]) temp|=0x2;
    if (key[KEY_RIGHT]) temp|=0x1;
    //                if (temp&0x10) output=1;
    return temp;

  case 0x1A:
  case 0x1B:
  case 0x1C:
  case 0x1D:
  case 0x1E:
  case 0x1F:
    return 0;
    
  case 0x100: case 0x110: case 0x120: case 0x130:
  case 0x140: case 0x150: case 0x160: case 0x170:
    return dmactrl[(addr>>4)&7];
    
  case 0x101: case 0x111: case 0x121: case 0x131:
  case 0x141: case 0x151: case 0x161: case 0x171:
    return dmadest[(addr>>4)&7];
    
  case 0x102: case 0x112: case 0x122: case 0x132:
  case 0x142: case 0x152: case 0x162: case 0x172:
    return dmasrc[(addr>>4)&7];
    
  case 0x103: case 0x113: case 0x123: case 0x133:
  case 0x143: case 0x153: case 0x163: case 0x173:
    return dmasrc[(addr>>4)&7]>>8;
    
  case 0x104: case 0x114: case 0x124: case 0x134:
  case 0x144: case 0x154: case 0x164: case 0x174:
    return dmasize[(addr>>4)&7]>>16;
    
  case 0x105: case 0x115: case 0x125: case 0x135:
  case 0x145: case 0x155: case 0x165: case 0x175:
    return dmasize[(addr>>4)&7];
    
  case 0x106: case 0x116: case 0x126: case 0x136:
  case 0x146: case 0x156: case 0x166: case 0x176:
    return dmasize[(addr>>4)&7]>>8;
    
  default:
    break;
  }
  return 0;
}

void writeio(u16 addr, u8 val)
{
  int c,bank;
  char st[80];

  switch (addr&0x1FF)
  {
  case 0:
    nmienable=val&0x80;
    intenable=val&0x30;
    break;
    
  case 0x02:
    multa=val;
    break;
  case 0x03:
    multb=val;
    multc=multa*multb;
    break;
    
  case 0x04:
    divc&=0xFF00;
    divc|=val;
    break;
  case 0x05:
    divc&=0xFF;
    divc|=val<<8;
    break;
  case 0x06:
    divb=val;
    if (!divc || !divb)
    {
      diva=0xFFFF;
      multc=divc;
    }
    else
    {
      diva=divc/divb;
      multc=divc%divb;
    }
    break;
    
  case 0x09:
    vertcomp=(vertcomp&0xFF00)|val;
    break;
  case 0xA:
    vertcomp=(vertcomp&0xFF)|(val<<8);
    break;

  case 0xB:
    for (c=0;c<8;c++)
    {
      if (val&(1<<c))
      {
	if (!dmasize[c])
	  dmasize[c]=0x10000;
	bank=dmasrc[c]>>16;

	while (dmasize[c])
	{

	  if (dmactrl[c]&0x80)
	    writemem(bank,dmasrc[c]&0xFFFF,readmem(0,0x2100|dmadest[c]));
	  else
	    writemem(0,0x2100|dmadest[c],readmem(bank,dmasrc[c]&0xFFFF));

	  if (!(dmactrl[c]&8))
	  {
	    if (dmactrl[c]&16)
	      dmasrc[c]--;
	    else
	      dmasrc[c]++;
	  }
	  switch (dmactrl[c]&7)
	  {
	  case 0:
	  case 2:
	    break;
	    
	  case 1:
	    dmadest[c]^=1;
	    break;
	    
	  default:
	    printf("DMA mode %i\n",dmactrl[c]&7);
	    exit(-1);
	  }
	  dmasize[c]--;
	}
	dmaout=0;
      }
    }
    break;
    
  case 0xC:
    hdmaon=val;
    break;

  case 0x10:
    vblank=0;
    break;

  case 0x11:
    inint=0;
    break;

  case 0x100: case 0x110: case 0x120: case 0x130:
  case 0x140: case 0x150: case 0x160: case 0x170:
    dmactrl[(addr>>4)&7]=val;
    break;

  case 0x101: case 0x111: case 0x121: case 0x131:
  case 0x141: case 0x151: case 0x161: case 0x171:
    dmadest[(addr>>4)&7]=val;
    break;

  case 0x102: case 0x112: case 0x122: case 0x132:
  case 0x142: case 0x152: case 0x162: case 0x172:
    dmasrc[(addr>>4)&7]&=0xFFFF00;
    dmasrc[(addr>>4)&7]|=val;
    break;
    
  case 0x103: case 0x113: case 0x123: case 0x133:
  case 0x143: case 0x153: case 0x163: case 0x173:
    dmasrc[(addr>>4)&7]&=0xFF00FF;
    dmasrc[(addr>>4)&7]|=val<<8;
    break;

  case 0x104: case 0x114: case 0x124: case 0x134:
  case 0x144: case 0x154: case 0x164: case 0x174:
    dmasrc[(addr>>4)&7]&=0x00FFFF;
    dmasrc[(addr>>4)&7]|=val<<16;
    break;

  case 0x105: case 0x115: case 0x125: case 0x135:
  case 0x145: case 0x155: case 0x165: case 0x175:
    dmasize[(addr>>4)&7]&=0xFF00;
    dmasize[(addr>>4)&7]|=val;
    break;
    
  case 0x106: case 0x116: case 0x126: case 0x136:
  case 0x146: case 0x156: case 0x166: case 0x176:
    dmasize[(addr>>4)&7]&=0xFF;
    dmasize[(addr>>4)&7]|=val<<8;
    break;
  }
}
