/*
 * InfoSNES - Super Nintendo (TM) emulator
 * Copyrigh(c) 2003 Jay's Factory (jays_factory@excite.co.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*-------------------------------------------------------------------*/
/*  Include Files                                                    */
/*-------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <allegro.h>

#include "snes.h"
#include "65816.h"
#include "ppu.h"
#include "spu.h"
#include "io.h"
#include "dsp.h"
#include "type.h"

/*-------------------------------------------------------------------*/
/*  PPU Resources                                                    */
/*-------------------------------------------------------------------*/

int ppumode;

int vraminc;
int vramrate[4]={1,32,128,128};
int firstread=1;
u16 bgaddr[4],chraddr[4];
u16 sprchraddr;
int xscroll[4],yscroll[4];
int bgsize[4];

u8 cgram[512];
u16 cgramaddr;

int firstobjwrite=1;

int bg3pri;
u8 bgenable,mainbg,subbg;

u16 matrixa,matrixb;
u32 matrixr;

u8 vramctrl;
u32 vramaddr;

u8 objram[1024];
u16 objaddr;

/*-------------------------------------------------------------------*/
/*  Other Resources                                                  */
/*-------------------------------------------------------------------*/

int spuskippos=0;
int spuskip;

int checkf000;
int screenon;
int rotozoom;
int hdmaena=0;
u8 extbgenable=0;

int first3d=1;
int hdmacount[8],hdmarepeat[8],hdmafirst[8];
u16 hdmaaddr[8];
u8 hdmabank[8];
u8 hdmadata[8][4];

u16 m7a,m7b,m7c,m7d,m7x,m7y;

int sprswap;
int videoon=0;
int fps,frames;

u32 bittable[256],bittable2[256],bittableh[256],bittableh2[256],masktable[256],masktable2[256],masktableh[256],masktableh2[256],coltable[8];

u32 wramaddr;

int objsize[4];
int objsizes[16][2]={ {1,1}, {2,2}, {1,1}, {4,4}, {1,1}, {8,8}, {2,2}, {4,4}, {2,2}, {8,8}, {4,4}, {8,8} };

u8 *vram;

BITMAP *b;
PALETTE pal;

/*-------------------------------------------------------------------*/
/*  Functions                                                        */
/*-------------------------------------------------------------------*/

void highcol()
{
  pal[0].r=63;
  set_color(0,&pal[0]);
}


void takescreenshot()
{
  save_pcx("scrshot.pcx",b,pal);
}

void initvideo()
{
  int c;
  videoon=1;

  set_gfx_mode(GFX_AUTODETECT,320,240,0,0);
  b=create_bitmap(512,256);

  for (c=0;c<256;c++)
  {
    bittable[c]=0;
    if (c&0x80) bittable[c]|=0x00000001;
    if (c&0x40) bittable[c]|=0x00000100;
    if (c&0x20) bittable[c]|=0x00010000;
    if (c&0x10) bittable[c]|=0x01000000;
    bittable2[c]=0;
    if (c&8) bittable2[c]|=0x00000001;
    if (c&4) bittable2[c]|=0x00000100;
    if (c&2) bittable2[c]|=0x00010000;
    if (c&1) bittable2[c]|=0x01000000;
    masktable[c]=0;
    if (c&0x80) masktable[c]|=0x000000FF;
    if (c&0x40) masktable[c]|=0x0000FF00;
    if (c&0x20) masktable[c]|=0x00FF0000;
    if (c&0x10) masktable[c]|=0xFF000000;
    masktable2[c]=0;
    if (c&8) masktable2[c]|=0x000000FF;
    if (c&4) masktable2[c]|=0x0000FF00;
    if (c&2) masktable2[c]|=0x00FF0000;
    if (c&1) masktable2[c]|=0xFF000000;
    bittableh[c]=0;
    if (c&0x80) bittableh[c]|=0x01000000;
    if (c&0x40) bittableh[c]|=0x00010000;
    if (c&0x20) bittableh[c]|=0x00000100;
    if (c&0x10) bittableh[c]|=0x00000001;
    bittableh2[c]=0;
    if (c&8) bittableh2[c]|=0x01000000;
    if (c&4) bittableh2[c]|=0x00010000;
    if (c&2) bittableh2[c]|=0x00000100;
    if (c&1) bittableh2[c]|=0x00000001;
    masktableh[c]=0;
    if (c&0x80) masktableh[c]|=0xFF000000;
    if (c&0x40) masktableh[c]|=0x00FF0000;
    if (c&0x20) masktableh[c]|=0x0000FF00;
    if (c&0x10) masktableh[c]|=0x000000FF;
    masktableh2[c]=0;
    if (c&8) masktableh2[c]|=0xFF000000;
    if (c&4) masktableh2[c]|=0x00FF0000;
    if (c&2) masktableh2[c]|=0x0000FF00;
    if (c&1) masktableh2[c]|=0x000000FF;
  }
  coltable[0]=0;
  coltable[1]=0x01010101;
  coltable[2]=0x02020202;
  coltable[3]=0x03030303;
  coltable[4]=0x04040404;
  coltable[5]=0x05050505;
  coltable[6]=0x06060606;
  coltable[7]=0x07070707;
  vram=(u8 *)malloc(65536);
}

void swapsprite()
{
  sprswap^=1;
}

void closevideo()
{
  FILE *f=fopen("vram.dmp","wb");
  fwrite(vram,0x10000,1,f);
  fwrite(objram,0x400,1,f);
  fclose(f);
  destroy_bitmap(b);
  free(vram);
}

void dumpppuregs()
{
  printf("BGEN %01X %04X %04X %04X %04X\n",bgenable&0xF,bgaddr[0],bgaddr[1],bgaddr[2],bgaddr[3]);
  printf("INTENA %02X  vertcomp %i\n",intenable,vertcomp);
  printf("mode %i\n",ppumode);
  printf("OBJ size %i %i %i %i\n",objsize[0],objsize[1],objsize[2],objsize[3]);
  printf("HDMA enable %02X\n",hdmaon);
}

void writeppu(u16 addr, u8 val)
{
  u16 tempw;
  switch (addr&0xFF)
  {
  case 0:
    screenon=!(val&0x80);
    break;

  case 0x01:
    sprchraddr=(val&3)<<14;
    val>>=5;
    val<<=1;
    objsize[0]=objsizes[val][0];
    objsize[1]=objsizes[val][1];
    objsize[2]=objsizes[val+1][0];
    objsize[3]=objsizes[val+1][1];
    break;

  case 0x02:
    objaddr&=0x100;
    objaddr|=val;
    break;
  case 0x03:
    if (val&1)
      objaddr|=0x100;
    else
      objaddr&=0xFE;
    break;
  case 0x04:
    if (objaddr>0x10F) return;
    if (firstobjwrite)
      objram[objaddr<<1]=val;
    else
    {
      objram[(objaddr<<1)+1]=val;
      objaddr++;
    }
    firstobjwrite^=1;
    break;

  case 0x05:
    ppumode=val&7;
    bg3pri=val&8;
    break;

  case 0x07:
    bgaddr[0]=(val&0xFC)<<9;
    bgsize[0]=val&3;
    break;
  case 0x08:
    bgaddr[1]=(val&0xFC)<<9;
    bgsize[1]=val&3;
    break;
  case 0x09:
    bgaddr[2]=(val&0xFC)<<9;
    bgsize[2]=val&3;
    break;
  case 0x0B:
    chraddr[0]=(val&0xF)<<13;
    chraddr[1]=(val>>4)<<13;
    break;
  case 0x0C:
    chraddr[2]=(val&0xF)<<13;
    chraddr[3]=(val>>4)<<13;
    break;
  case 0x0D:
    xscroll[0]>>=8;
    xscroll[0]|=(val<<8);
    break;
  case 0x0E:
    yscroll[0]>>=8;
    yscroll[0]|=(val<<8);
    break;
  case 0x0F:
    xscroll[1]>>=8;
    xscroll[1]|=(val<<8);
    break;
  case 0x10:
    yscroll[1]>>=8;
    yscroll[1]|=(val<<8);
    break;
  case 0x11:
    xscroll[2]>>=8;
    xscroll[2]|=(val<<8);
    break;
  case 0x12:
    yscroll[2]>>=8;
    yscroll[2]|=(val<<8);
    break;
  case 0x15:
    vramctrl=val;
    vraminc=vramrate[val&3];
    break;
  case 0x16:
    firstread=1;
    vramaddr&=0xFF00;
    vramaddr|=val&0xFF;
    break;
  case 0x17:
    firstread=1;
    vramaddr&=0xFF;
    vramaddr|=val<<8;
    break;
  case 0x18:
    firstread=1;
    vram[(vramaddr<<1)&0xFFFF]=val;
    if (!(vramctrl&0x80))
      vramaddr+=vraminc;
    break;
  case 0x19:
    firstread=1;
    vram[((vramaddr<<1)+1)&0xFFFF]=val;
    if (vramctrl&0x80)
      vramaddr+=vraminc;
    break;
  case 0x1B:
    m7a=(m7a>>8)|(val<<8);
    matrixr=(short)m7a*((short)m7b>>8);
    break;
  case 0x1C:
    m7b=(m7b>>8)|(val<<8);
    matrixr=(short)m7a*((short)m7b>>8);
    break;
  case 0x1D:
    m7c=(m7c>>8)|(val<<8);
    break;
  case 0x1E:
    m7d=(m7d>>8)|(val<<8);
    break;
  case 0x1F:
    m7x=(m7x>>8)|(val<<8);
    break;
  case 0x20:
    m7y=(m7y>>8)|(val<<8);
    break;
  case 0x21:
    cgramaddr=val<<1;
    break;
  case 0x22:
    cgram[cgramaddr&0x1FF]=val;
    tempw=cgram[cgramaddr&0x1FE]|(cgram[(cgramaddr&0x1FE)+1]<<8);
    pal[cgramaddr>>1].r=(tempw&31)<<1;
    pal[cgramaddr>>1].g=((tempw>>5)&31)<<1;
    pal[cgramaddr>>1].b=((tempw>>10)&31)<<1;
    if (videoon)
      set_color((cgramaddr>>1)&0xFF,&pal[cgramaddr>>1]);
    cgramaddr++;
    break;
  case 0x2C:
    mainbg=val;
    bgenable=(mainbg^extbgenable);//|subbg;
    break;
  case 0x2D:
    subbg=val;
    bgenable=(mainbg^extbgenable);//|subbg;
    break;
    
  case 0x40: case 0x41: case 0x42: case 0x43: case 0x44: case 0x45: case 0x46: case 0x47:
  case 0x48: case 0x49: case 0x4A: case 0x4B: case 0x4C: case 0x4D: case 0x4E: case 0x4F:
  case 0x50: case 0x51: case 0x52: case 0x53: case 0x54: case 0x55: case 0x56: case 0x57:
  case 0x58: case 0x59: case 0x5A: case 0x5B: case 0x5C: case 0x5D: case 0x5E: case 0x5F:
  case 0x60: case 0x61: case 0x62: case 0x63: case 0x64: case 0x65: case 0x66: case 0x67:
  case 0x68: case 0x69: case 0x6A: case 0x6B: case 0x6C: case 0x6D: case 0x6E: case 0x6F:
  case 0x70: case 0x71: case 0x72: case 0x73: case 0x74: case 0x75: case 0x76: case 0x77:
  case 0x78: case 0x79: case 0x7A: case 0x7B: case 0x7C: case 0x7D: case 0x7E: case 0x7F:
    writespuport(addr,val);
    break;
    
  case 0x80:
    ram[wramaddr++]=val;
    wramaddr&=0x1FFFF;
    break;
  case 0x81:
    wramaddr=(wramaddr&0x1FF00)|val;
    break;
  case 0x82:
    wramaddr=(wramaddr&0x100FF)|(val<<8);
    break;
  case 0x83:
    wramaddr=(wramaddr&0xFFFF)|((val&1)<<16);
    break;
  }
}

u8 spuskipper(u16 addr)
{
  switch (addr&0xFF)
  {
  case 0x40: case 0x42:
    oldcycles=cycles;
    spuskippos++;
    if (spuskippos>=35) spuskippos=0;
    switch (spuskippos>>1)
    {
    case 0:
      return a.b.l;
    case 1:
      return x.b.l;
    case 2:
      return y.b.l;
    case 3:
      return 0x55;
    case 4:
      return 0xAA;
    case 5:
      return 7;
    case 6:
      return 1;
    case 7:
      return 0xFF;
    case 8:
      return 0xAB;
    case 9:
      return 0xBC;
    case 10:
      return 0x22;
    case 11:
      return 0;
    case 12:
      return 0;
    case 13:
      return ram[0];
    case 14:
      return a.b.h;
    case 15:
      return x.b.h;
    case 16:
      return y.b.h;
    case 17:
      return 0xAA;
    }
    return 0;
    
  case 0x41: case 0x43:
    if (cycles!=oldcycles)
    {
      spuskippos++;
      if (spuskippos>=35) spuskippos=0;
    }
    switch (spuskippos>>1)
    {
    case 0:
      return a.b.h;
    case 1:
      return x.b.h;
    case 2:
      return y.b.h;
    case 3:
      return 0xAA;
    case 4:
      return 0x55;
    case 5:
      return 0;
    case 6:
      return 0;
    case 7:
      return 0xFF;
    case 8:
      return 0xCD;
    case 9:
      return 0xBC;
    case 10:
      return 0x22;
    case 11:
      return 4;
    case 12:
      return 4;
    case 13:
      return ram[0];
    case 14:
      return a.b.l;
    case 15:
      return x.b.l;
    case 16:
      return y.b.l;
    case 17:
      return 0xBB;
    }
    return 0;
  }
}

u8 readppu(u16 addr)
{
  u8 temp;
  switch (addr&0xFF)
  {
  case 0x34:
    return matrixr;
  case 0x35:
    return matrixr>>8;
  case 0x36:
    return matrixr>>16;
    
  case 0x37:
    latchedvline=v_line;
    first3d=1;
    return 0;
    
  case 0x39:
    if (firstread)
    {
      firstread=0;
      return vram[(vramaddr<<1)&0xFFFF];
    }
    temp=vram[((vramaddr<<1)-2)&0xFFFF];
    if (!(vramctrl&0x80))
      vramaddr+=vraminc;
    return temp;
    
  case 0x3A:
    if (firstread)
    {
      firstread=0;
      return vram[((vramaddr<<1)+1)&0xFFFF];
    }
    temp=vram[((vramaddr<<1)-1)&0xFFFF];
    if (vramctrl&0x80)
      vramaddr+=vraminc;
    return temp;
    
  case 0x3D:
    if (first3d)
    {
      first3d=0;
      return v_line;
    }
    else
    {
      first3d=1;
      return v_line>>8;
    }
    break;

  case 0x3F:
    if (palf) return 0x10;
    return 0;
    
  case 0x40: case 0x41: case 0x42: case 0x43: case 0x44: case 0x45: case 0x46: case 0x47:
  case 0x48: case 0x49: case 0x4A: case 0x4B: case 0x4C: case 0x4D: case 0x4E: case 0x4F:
  case 0x50: case 0x51: case 0x52: case 0x53: case 0x54: case 0x55: case 0x56: case 0x57:
  case 0x58: case 0x59: case 0x5A: case 0x5B: case 0x5C: case 0x5D: case 0x5E: case 0x5F:
  case 0x60: case 0x61: case 0x62: case 0x63: case 0x64: case 0x65: case 0x66: case 0x67:
  case 0x68: case 0x69: case 0x6A: case 0x6B: case 0x6C: case 0x6D: case 0x6E: case 0x6F:
  case 0x70: case 0x71: case 0x72: case 0x73: case 0x74: case 0x75: case 0x76: case 0x77:
  case 0x78: case 0x79: case 0x7A: case 0x7B: case 0x7C: case 0x7D: case 0x7E: case 0x7F:
    if (spuskip)
      return spuskipper(addr);
    return readspuport(addr);
    
  case 0x80:
    return ram[wramaddr++];
  }
  return 0;
}

inline void dohdma()
{
  int c,d;
  for (d=0;d<262;d++)
  {
    for (c=0;c<8;c++)
    {
      if (hdmaon & (1<<c))
      {
	if (!(hdmacount[c]&0x7F))
	{
	  hdmacount[c]=readmem(hdmabank[c],hdmaaddr[c]++);
	  hdmarepeat[c]=hdmacount[c]&0x80;
	  hdmacount[c]&=0x7F;
	  switch (dmactrl[c]&7)
	  {
	  case 0:
	    hdmadata[c][0]=readmem(hdmabank[c],hdmaaddr[c]++);
	    break;
	  case 1:
	  case 2:
	    hdmadata[c][0]=readmem(hdmabank[c],hdmaaddr[c]++);
	    hdmadata[c][1]=readmem(hdmabank[c],hdmaaddr[c]++);
	    break;
	  case 3:
	  case 4:
	    hdmadata[c][0]=readmem(hdmabank[c],hdmaaddr[c]++);
	    hdmadata[c][1]=readmem(hdmabank[c],hdmaaddr[c]++);
	    hdmadata[c][2]=readmem(hdmabank[c],hdmaaddr[c]++);
	    hdmadata[c][3]=readmem(hdmabank[c],hdmaaddr[c]++);
	    break;
	  }
	  hdmafirst[c]=1;
	}
	if (hdmafirst[c] || hdmarepeat[c])
	{
	  hdmafirst[c]=0;
	  switch (dmactrl[c]&7)
	  {
	  case 0:
	    writemem(0,0x2100|dmadest[c],hdmadata[c][0]);
	    break;
	  case 1:
	    writemem(0,0x2100|dmadest[c],hdmadata[c][0]);
	    writemem(0,0x2100|(dmadest[c]+1),hdmadata[c][1]);
	    break;
	  case 2:
	    writemem(0,0x2100|dmadest[c],hdmadata[c][0]);
	    writemem(0,0x2100|dmadest[c],hdmadata[c][1]);
	    break;
	  case 3:
	    writemem(0,0x2100|dmadest[c],hdmadata[c][0]);
	    writemem(0,0x2100|(dmadest[c]+1),hdmadata[c][1]);
	    writemem(0,0x2100|dmadest[c],hdmadata[c][2]);
	    writemem(0,0x2100|(dmadest[c]+1),hdmadata[c][3]);
	    break;
	  case 4:
	    writemem(0,0x2100|dmadest[c],hdmadata[c][0]);
	    writemem(0,0x2100|(dmadest[c]+1),hdmadata[c][1]);
	    writemem(0,0x2100|(dmadest[c]+2),hdmadata[c][2]);
	    writemem(0,0x2100|(dmadest[c]+3),hdmadata[c][3]);
	    break;
	  }
	}
	hdmacount[c]--;
      }
    }
  }
}

void drawmode0layer(BITMAP *b, u16 bgaddr, u16 chraddr, int layer, int xmask, int xoff)
{
  u16 temp;
  u32 templ;
  u8 dat1,dat2;
  int x,xx,y,yy;
  int col;
  bgaddr>>=1;
  xmask>>=2;
  xoff>>=2;
  xoff+=xscroll[layer]>>2;
  for (y=0;y<256;y+=8)
  {
    for (x=0;x<64;x+=2)
    {
      temp=((short *)vram)[bgaddr++];
      col=coltable[(temp>>10)&7]<<2;
      temp=((temp&1023)<<4)+chraddr;
      for (yy=0;yy<8;yy++)
      {
	dat1=vram[temp];
	dat2=vram[temp+1];
	temp+=2;
	if (((x-xoff)&xmask)<256)
	  {
	    templ=bittable[dat1]|(bittable[dat2]<<1);
	    ((long *)b->line[(y+yy-yscroll[layer])&255])[(x-xoff)&xmask]=templ|col;
	    templ=bittable2[dat1]|(bittable2[dat2]<<1);
	    ((long *)b->line[(y+yy-yscroll[layer])&255])[((x-xoff)+1)&xmask]=templ|col;
	  }
      }
    }
  }
}

void drawmode0layermask(BITMAP *b, u16 bgaddr, u16 chraddr, int layer)
{
  u16 temp,temp2;
  u32 templ,templ2;
  u8 dat1,dat2;
  int x,xx,y,yy;
  int col;
  bgaddr>>=1;
  for (y=0;y<256;y+=8)
  {
    for (x=0;x<64;x+=2)
    {
      temp2=((short *)vram)[bgaddr++];
      col=coltable[(temp2>>10)&7]<<2;
      temp=((temp2&1023)<<4)+chraddr;
      for (yy=0;yy<8;yy++)
      {
	dat1=vram[temp];
	dat2=vram[temp+1];
	temp+=2;
	templ2=bittable[dat1]|(bittable[dat2]<<1);
	if (templ2)
	{
	  templ=((long *)b->line[(y+yy-yscroll[layer])&255])[(x-(xscroll[layer]>>2))&255] & ~masktable[dat1|dat2];
	  templ|=(templ2|col)&masktable[dat1|dat2];
	  ((long *)b->line[(y+yy-yscroll[layer])&255])[(x-(xscroll[layer]>>2))&255]=templ;
	}
	templ2=bittable2[dat1]|(bittable2[dat2]<<1);
	if (templ2)
	{
	  templ=((long *)b->line[(y+yy-yscroll[layer])&255])[((x-(xscroll[layer]>>2))+1)&255] & ~masktable2[dat1|dat2];
	  templ|=(templ2|col)&masktable2[dat1|dat2];
	  ((long *)b->line[(y+yy-yscroll[layer])&255])[((x-(xscroll[layer]>>2))+1)&255]=templ;
	}
      }
    }
  }
}

void drawmode1layer(BITMAP *b, u16 bgaddr, u16 chraddr, int layer, int xmask, int xoff)
{
  u16 temp;
  u32 templ;
  u8 dat1,dat2,dat3,dat4;
  int x,xx,y,yy;
  int col;
  bgaddr>>=1;
  xmask>>=2;
  xoff>>=2;
  xoff+=xscroll[layer]>>2;
  for (y=0;y<256;y+=8)
  {
    for (x=0;x<64;x+=2)
    {
      temp=((short *)vram)[bgaddr++];
      col=coltable[(temp>>10)&7]<<4;
      temp=((temp&1023)<<5)+chraddr;
      for (yy=0;yy<8;yy++)
      {
	dat1=vram[temp];
	dat2=vram[temp+1];
	dat3=vram[temp+16];
	dat4=vram[temp+17];
	temp+=2;
	if (((x-xoff)&xmask)<256)
	{
	  templ=bittable[dat1]|(bittable[dat2]<<1)|(bittable[dat3]<<2)|(bittable[dat4]<<3);
	  ((long *)b->line[(y+yy-yscroll[layer])&255])[(x-xoff)&xmask]=templ|col;
	  templ=bittable2[dat1]|(bittable2[dat2]<<1)|(bittable2[dat3]<<2)|(bittable2[dat4]<<3);
	  ((long *)b->line[(y+yy-yscroll[layer])&255])[((x-xoff)+1)&xmask]=templ|col;
	}
      }
    }
  }
}

void drawmode1layermask(BITMAP *b, u16 bgaddr, u16 chraddr, int layer, int xmask, int xoff)
{
  u16 temp;
  u32 templ,templ2;
  u8 dat1,dat2,dat3,dat4;
  int x,xx,y,yy;
  int col;
  bgaddr>>=1;
  xmask>>=2;
  xoff>>=2;
  xoff+=xscroll[layer]>>2;
  for (y=0;y<256;y+=8)
  {
    for (x=0;x<64;x+=2)
    {
      temp=((short *)vram)[bgaddr++];
      col=coltable[(temp>>10)&7]<<4;
      temp=((temp&1023)<<5)+chraddr;
      for (yy=0;yy<8;yy++)
      {
	dat1=vram[temp];
	dat2=vram[temp+1];
	dat3=vram[temp+16];
	dat4=vram[temp+17];
	if (((x-xoff)&xmask)<256)
	{
	  temp+=2;
	  templ2=bittable[dat1]|(bittable[dat2]<<1)|(bittable[dat3]<<2)|(bittable[dat4]<<3);
	  if (templ2)
	  {
	    templ=((long *)b->line[(y+yy-yscroll[layer])&255])[(x-xoff)&xmask] & ~masktable[dat1|dat2|dat3|dat4];
	    templ|=(templ2|col)&masktable[dat1|dat2|dat3|dat4];
	    ((long *)b->line[(y+yy-yscroll[layer])&255])[(x-xoff)&xmask]=templ;
	  }
	  templ2=bittable2[dat1]|(bittable2[dat2]<<1)|(bittable2[dat3]<<2)|(bittable2[dat4]<<3);
	  if (templ2)
	  {
	    templ=((long *)b->line[(y+yy-yscroll[layer])&255])[((x-xoff)+1)&xmask] & ~masktable2[dat1|dat2|dat3|dat4];
	    templ|=(templ2|col)&masktable2[dat1|dat2|dat3|dat4];
	    ((long *)b->line[(y+yy-yscroll[layer])&255])[((x-xoff)+1)&xmask]=templ;
	  }
	}
      }
    }
  }
}

void drawmode3layer(BITMAP *b, u16 bgaddr, u16 chraddr, int layer)
{
  u16 temp;
  u32 templ;
  u8 dat1,dat2,dat3,dat4,dat5,dat6,dat7,dat8;
  int x,xx,y,yy;
  bgaddr>>=1;
  for (y=0;y<256;y+=8)
  {
    for (x=0;x<64;x+=2)
    {
      temp=((short *)vram)[bgaddr++];
      temp=((temp&1023)<<6)+chraddr;
      for (yy=0;yy<8;yy++)
      {
	dat1=vram[temp];
	dat2=vram[temp+1];
	dat3=vram[temp+16];
	dat4=vram[temp+17];
	dat5=vram[temp+32];
	dat6=vram[temp+33];
	dat7=vram[temp+48];
	dat8=vram[temp+49];
	temp+=2;
	templ=bittable[dat1]|(bittable[dat2]<<1)|(bittable[dat3]<<2)|(bittable[dat4]<<3)|(bittable[dat4]<<4)|(bittable[dat5]<<5)|(bittable[dat6]<<6)|(bittable[dat7]<<7);
	((long *)b->line[(y+yy-yscroll[layer])&255])[(x-(xscroll[layer]>>2))&255]=templ;
	//                                ((long *)b->line[y+yy])[x]=templ;
	templ=bittable2[dat1]|(bittable2[dat2]<<1)|(bittable2[dat3]<<2)|(bittable2[dat4]<<3)|(bittable2[dat4]<<4)|(bittable2[dat5]<<5)|(bittable2[dat6]<<6)|(bittable2[dat7]<<7);
	((long *)b->line[(y+yy-yscroll[layer])&255])[(x-(xscroll[layer]>>2)+1)&255]=templ;
	//                                ((long *)b->line[y+yy])[x+1]=templ;
      }
    }
  }
}

void drawmode3layermask(BITMAP *b, u16 bgaddr, u16 chraddr, int layer)
{
  u16 temp;
  u32 templ,templ2;
  u8 dat1,dat2,dat3,dat4,dat5,dat6,dat7,dat8;
  int x,xx,y,yy;
  bgaddr>>=1;
  for (y=0;y<256;y+=8)
  {
    for (x=0;x<64;x+=2)
    {
      temp=((short *)vram)[bgaddr++];
      temp=((temp&1023)<<6)+chraddr;
      for (yy=0;yy<8;yy++)
      {
	dat1=vram[temp];
	dat2=vram[temp+1];
	dat3=vram[temp+16];
	dat4=vram[temp+17];
	dat5=vram[temp+32];
	dat6=vram[temp+33];
	dat7=vram[temp+48];
	dat8=vram[temp+49];
	temp+=2;
	templ2=bittable[dat1]|(bittable[dat2]<<1)|(bittable[dat3]<<2)|(bittable[dat4]<<3)|(bittable[dat4]<<4)|(bittable[dat5]<<5)|(bittable[dat6]<<6)|(bittable[dat7]<<7);
	if (templ2)
	{
	  templ=((long *)b->line[(y+yy-yscroll[layer])&255])[(x-(xscroll[layer]>>2))&255] & ~masktable[dat1|dat2|dat3|dat4|dat5|dat6|dat7|dat8];
	  templ|=templ2;
	  ((long *)b->line[(y+yy-yscroll[layer])&255])[(x-(xscroll[layer]>>2))&255]=templ;
	}
	templ2=bittable2[dat1]|(bittable2[dat2]<<1)|(bittable2[dat3]<<2)|(bittable2[dat4]<<3)|(bittable2[dat4]<<4)|(bittable2[dat5]<<5)|(bittable2[dat6]<<6)|(bittable2[dat7]<<7);
	if (templ2)
	{
	  templ=((long *)b->line[(y+yy-yscroll[layer])&255])[(x-(xscroll[layer]>>2)+1)&255] & ~masktable2[dat1|dat2|dat3|dat4|dat5|dat6|dat7|dat8];
	  templ|=templ2;
	  ((long *)b->line[(y+yy-yscroll[layer])&255])[(x-(xscroll[layer]>>2)+1)&255]=templ;
	}
      }
    }
  }
}

void drawmode7()
{
  int x,y;
  int xx,yy;
  unsigned int aa=0,bb,cc=0,dd;
  unsigned int m7xx=(m7x<<19)>>19,m7yy=(m7y<<19)>>19;
  u8 temp,col;
  for (y=0;y<224;y++)
  {
    aa=(m7a*(-m7xx));
    bb=(m7b*(y-m7yy))+(m7xx<<8);
    cc=(m7c*(-m7xx));
    dd=(m7d*(y-m7yy))+(m7yy<<8);
    for (x=0;x<256;x++)
    {
      getm7pixel(((aa+bb)>>9)&1023,((cc+dd)>>9)&1023);
      b->line[y][x]=col;
      aa+=m7a;
      cc+=m7c;
    }
  }
}

#define drawtile()                                                              \
                        for (yy=0;yy<8;yy++)                                    \
                        {                                                       \
                                dat1=vram[temp];                                \
                                dat2=vram[temp+1];                              \
                                dat3=vram[temp+16];                             \
                                dat4=vram[temp+17];                             \
                                temp+=2;                                        \
                                templ2=bittable[dat1]|(bittable[dat2]<<1)|(bittable[dat3]<<2)|(bittable[dat4]<<3); \
                                if (templ2)                                                                        \
                                {                                                                                  \
                                        templ=((long *)(b->line[(y+yy)&255]+(holdx&3)))[x] & ~masktable[dat1|dat2|dat3|dat4]; \
                                        templ|=templ2|(col&masktable[dat1|dat2|dat3|dat4]);                                                               \
                                        ((long *)(b->line[(y+yy)&255]+(holdx&3)))[x]=templ;                                  \
                                }                                                                                    \
                                templ2=bittable2[dat1]|(bittable2[dat2]<<1)|(bittable2[dat3]<<2)|(bittable2[dat4]<<3); \
                                if (templ2)                                                                           \
                                {                                                                                     \
                                        templ=((long *)(b->line[(y+yy)&255]+(holdx&3)))[x+1] & ~masktable2[dat1|dat2|dat3|dat4]; \
                                        templ|=templ2|(col&masktable2[dat1|dat2|dat3|dat4]);                                                               \
                                        ((long *)(b->line[(y+yy)&255]+(holdx&3)))[x+1]=templ;                                \
                                } \
                        }

#define drawtileh()                                                              \
                        for (yy=0;yy<8;yy++)                                    \
                        {                                                       \
                                dat1=vram[temp];                                \
                                dat2=vram[temp+1];                              \
                                dat3=vram[temp+16];                             \
                                dat4=vram[temp+17];                             \
                                temp+=2;                                        \
                                templ2=bittableh[dat1]|(bittableh[dat2]<<1)|(bittableh[dat3]<<2)|(bittableh[dat4]<<3); \
                                if (templ2)                                                                        \
                                {                                                                                  \
                                        templ=((long *)(b->line[(y+yy)&255]+(holdx&3)))[x+1] & ~masktableh[dat1|dat2|dat3|dat4]; \
                                        templ|=templ2|(col&masktableh[dat1|dat2|dat3|dat4]);                                                               \
                                        ((long *)(b->line[(y+yy)&255]+(holdx&3)))[x+1]=templ;                                  \
                                }                                                                                    \
                                templ2=bittableh2[dat1]|(bittableh2[dat2]<<1)|(bittableh2[dat3]<<2)|(bittableh2[dat4]<<3); \
                                if (templ2)                                                                           \
                                {                                                                                     \
                                        templ=((long *)(b->line[(y+yy)&255]+(holdx&3)))[x] & ~masktableh2[dat1|dat2|dat3|dat4]; \
                                        templ|=templ2|(col&masktableh2[dat1|dat2|dat3|dat4]);                                                               \
                                        ((long *)(b->line[(y+yy)&255]+(holdx&3)))[x]=templ;                                \
                                } \
                        }

void dosprites()
{
  int c;
  int addr=512-4;
  int x,xx,y,yy,yyy,tile,col,hflip;
  int holdx,holdtile;
  u32 temp,templ,templ2;
  u8 dat1,dat2,dat3,dat4;
  int sprsize;
  int sprx=2,spry=2;
  int addrtemp=543;
  int tempval,temppos=0;
  tempval=objram[addrtemp];
  for (c=127;c>=0;c--)
  {
    x=objram[c<<2];
    if (tempval&0x40) x|=256;
    y=objram[(c<<2)+1];
    if (x && y>-1)
    {
      tile=objram[(c<<2)+2]|((objram[(c<<2)+3]&1)<<8);
      col=(coltable[(objram[(c<<2)+3]>>1)&7]|0x08080808)<<4;
      hflip=objram[(c<<2)+3]&0x40;
      sprsize=((tempval&0x80)>>1)^sprswap;
//                        printf("%i %04X %i %02X %i\n",c,addrtemp,temppos,tempval,sprsize);
      if (sprsize)
      {
	sprx=objsize[2];
	spry=objsize[3];
      }
      else
      {
	sprx=objsize[0];
	spry=objsize[1];
      }
      //                        x+=2;
      y++;
      if (x<256 && y<224)
      {
	holdx=x;
	holdtile=tile;
	for (yyy=0;yyy<spry;yyy++)
	  {
	    x=holdx;
	    tile=holdtile;
	    holdtile+=16;
	    x>>=2;
	    if (hflip)
	    {
	      x+=(sprx);
	      for (xx=0;xx<sprx;xx++)
	      {
		temp=((tile&511)<<5)+sprchraddr;
		drawtileh()
		  x-=2;
		tile++;
	      }
	    }
	    else
	    {
	      for (xx=0;xx<sprx;xx++)
	      {
		temp=((tile&511)<<5)+sprchraddr;
		drawtile()
		  x+=2;
		tile++;
	      }
	    }
	    y+=8;
	  }
      }
    }
    tempval<<=2;
    temppos++;
    if (temppos==4)
    {
      addrtemp--;
      tempval=objram[addrtemp];
    }
    temppos&=3;
  }
}

void drawscreen()
{
  int x,y,yy;
  u16 addr=0,tile;
  int drawn=0;
  char s[80];
  if (hdmaena)
    dohdma();

  if (ppumode==1)
  {
    if (bgenable&4 && !bg3pri)
    {
      drawmode0layer(b,bgaddr[2],chraddr[2],2,255,0);
      drawn=1;
    }
    if (bgenable&2)
    {
      if (drawn)
      {
	if (bgsize[1]&1)
	{
	  drawmode1layermask(b,bgaddr[1],chraddr[1],1,511,0);
	  drawmode1layermask(b,bgaddr[1]+0x800,chraddr[1],1,511,256);
	}
	else
	  drawmode1layermask(b,bgaddr[1],chraddr[1],1,255,0);
      }
      else
      {
	if (bgsize[1]&1)
	{
	  drawmode1layer(b,bgaddr[1],chraddr[1],1,511,0);
	  drawmode1layer(b,bgaddr[1]+0x800,chraddr[1],1,511,256);
	}
	else
	  drawmode1layer(b,bgaddr[1],chraddr[1],1,255,0);
      }
      drawn=1;
    }
    if (bgenable&1)
    {
      if (drawn)
      {
	if (bgsize[0]&1)
	{
	  drawmode1layermask(b,bgaddr[0],chraddr[0],0,511,0);
	  drawmode1layermask(b,bgaddr[0]+0x800,chraddr[0],0,511,256);
	}
	else
	  drawmode1layermask(b,bgaddr[0],chraddr[0],0,255,0);
      }
      else
      {
	if (bgsize[0]&1)
	{
	  drawmode1layer(b,bgaddr[0],chraddr[0],0,511,0);
	  drawmode1layer(b,bgaddr[0]+0x800,chraddr[0],0,511,256);
	}
	else
	  drawmode1layer(b,bgaddr[0],chraddr[0],0,255,0);
      }
      drawn=1;
    }
    if (bgenable&0x10)
    {
      dosprites();
      drawn=1;
    }
    if (bgenable&4 && bg3pri)
    {
      if (drawn)
	drawmode0layermask(b,bgaddr[2],chraddr[2],2);
      else
	drawmode0layer(b,bgaddr[2],chraddr[2],2,255,0);
    }
  }
  else if (ppumode==0)
  {
    if (bgenable&4)
    {
      drawmode0layer(b,bgaddr[2],chraddr[2],2,255,0);
      drawn=1;
    }
    if (bgenable&2)
    {
      if (drawn)
	drawmode0layermask(b,bgaddr[1],chraddr[1],1);
      else
	drawmode0layer(b,bgaddr[1],chraddr[1],1,255,0);
      drawn=1;
    }
    if (bgenable&1)
    {
      if (drawn)
	drawmode0layermask(b,bgaddr[0],chraddr[0],0);
      else
	drawmode0layer(b,bgaddr[0],chraddr[0],0,255,0);
      drawn=1;
    }
  }
  else if (ppumode==2)
  {
    if (bgenable&2)
    {
      drawmode1layer(b,bgaddr[1],chraddr[1],1,255,0);
      drawn=1;
    }
    if (bgenable&1)
    {
      if (drawn)
	drawmode1layermask(b,bgaddr[0],chraddr[0],0,255,0);
      else
	drawmode1layer(b,bgaddr[0],chraddr[0],0,255,0);
      drawn=1;
    }
  }
  else if (ppumode==3)
  {
    if (bgenable&2)
    {
      drawmode1layer(b,bgaddr[1],chraddr[1],1,255,0);
      drawn=1;
    }
    if (bgenable&1)
    {
      if (drawn)
	drawmode3layermask(b,bgaddr[0],chraddr[0],0);
      else
	drawmode3layer(b,bgaddr[0],chraddr[0],0);
      drawn=1;
    }
  }
  else if (ppumode==4)
  {
    if (bgenable&2)
    {
      drawmode0layer(b,bgaddr[1],chraddr[1],1,255,0);
      drawn=1;
    }
    if (bgenable&1)
    {
      if (drawn)
	drawmode3layermask(b,bgaddr[0],chraddr[0],0);
      else
	drawmode3layer(b,bgaddr[0],chraddr[0],0);
      drawn=1;
    }
  }
  else if (ppumode==7)
  {
    if (!rotozoom)
    {
      for (y=0;y<224;y+=8)
      {
	for (x=0;x<256;x+=8)
	{
	  tile=(vram[addr]<<7)+1;
	  addr+=2;
	  for (yy=0;yy<8;yy++)
	  {
	    b->line[y+yy][x]=vram[tile];
	    b->line[y+yy][x+1]=vram[tile+2];
	    b->line[y+yy][x+2]=vram[tile+4];
	    b->line[y+yy][x+3]=vram[tile+6];
	    b->line[y+yy][x+4]=vram[tile+8];
	    b->line[y+yy][x+5]=vram[tile+10];
	    b->line[y+yy][x+6]=vram[tile+12];
	    b->line[y+yy][x+7]=vram[tile+14];
	    tile+=16;
	  }
	}
	addr+=192;
      }
    }
    else
    drawmode7();
  }

  if (!(bgenable&0xF))
    clear(b);
  if (bgenable&0x10 && ppumode!=1)
    dosprites();
  if (activechan&1)   b->line[0][32]=makecol(255,0,0);
  if (activechan&2)   b->line[0][34]=makecol(255,0,0);
  if (activechan&4)   b->line[0][36]=makecol(255,0,0);
  if (activechan&8)   b->line[0][38]=makecol(255,0,0);
  if (activechan&16)  b->line[0][40]=makecol(255,0,0);
  if (activechan&32)  b->line[0][42]=makecol(255,0,0);
  if (activechan&64)  b->line[0][44]=makecol(255,0,0);
  if (activechan&128) b->line[0][46]=makecol(255,0,0);

  blit(b,screen,0,0,32,8,256,224);
  sprintf(s,"%i  ",fps);
  textout(screen,font,s,0,0,makecol(255,255,255));
  frames++;
}
