/*
 * InfoSNES - Super Nintendo (TM) emulator
 * Copyrigh(c) 2003 Jay's Factory (jays_factory@excite.co.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*-------------------------------------------------------------------*/
/*  Include Files                                                    */
/*-------------------------------------------------------------------*/

#include <allegro.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "snes.h"
#include "65816.h"
#include "spu.h"
#include "ppu.h"
#include "io.h"
#include "type.h"

/*-------------------------------------------------------------------*/
/*  SNES Resources                                                   */
/*-------------------------------------------------------------------*/

u8 ram[0x20000];
u8 rom[0xC0][0x8000]; /* 6 megs gone already!!!! */
u8 sram[0x10000];

/*-------------------------------------------------------------------*/
/*  Other Resources                                                  */
/*-------------------------------------------------------------------*/

int v_line, latchedvline;
int vbline;

/*-------------------------------------------------------------------*/
/*  Display and Others resouces                                      */
/*-------------------------------------------------------------------*/

int seconds=0;
int inton=1;
int palf;
int frame=0;

/*-------------------------------------------------------------------*/
/*  Functions                                                        */
/*-------------------------------------------------------------------*/

void dumpram()
{
  FILE *f=fopen("ram.dmp","wb");
  int c;
  fwrite(ram,0x20000,1,f);
  fclose(f);
}

void loadsmc(char *fn)
{
  FILE *f=fopen(fn,"rb");
  int c,len;
  u8 title[21];

  fseek(f,0,SEEK_END);
  len=ftell(f);
  printf("len %i\n",len);
  len&=512;
  fseek(f,len,SEEK_SET);
  c=0;
  fread(rom[0],32768,1,f);
  fseek(f,len,SEEK_SET);
  memcpy(title,rom[0]+0x7FC0,20);
  title[20]=0;
  printf("Game title : %s\n",title);
  printf("ROM type %01X\n",rom[0][0x7FD6]&0xF);
  printf("Country code %02X\n",rom[0][0x7FD9]);
  if ((rom[0][0x7FD9]&0xF)>=2) palf=1;
  printf("%s\n",(palf)?"PAL":"NTSC");
  while (!feof(f))
    fread(rom[c++],32768,1,f);
  fclose(f);
}

void countsecs()
{
  seconds++;
  fps=frames;
  frames=0;
}

void execsnes(int lines)
{
  for (v_line=0;v_line<lines;v_line++)
  {
    if (v_line==0)
      vblank=0;
    if (v_line==vbline)
    {
      vblank=1;
      if (nmienable)
	nmi65816();
    }
    if (v_line==vertcomp && intenable&0x20 && inton)
    {
      int65816();
      //                        printf("Interrupt line %i %02X:%04X\n",v_line,pbr,pc);
    }
    exec65816(192);
    if (!spuskip) execspu(128);
  }
}

int main(int argc, char *argv[])
{
  int cycle;
  FILE *f;
  if (argc<2)
  {
    printf("Syntax : InfoSNES smcfile.smc\n");
    exit(-1);
  }
  loadsmc(argv[1]);
  if (argc>2)
  {
    if (!stricmp(argv[2],"-spc"))
    {
      spuskip=0;
      printf("SPC700 enabled\n");
    }
  }
#if 0
        getch();
#endif
  if (palf) vbline=288;
  else     vbline=224;
  allegro_init();
#if 0
  get_palette(dospal);
#endif
  install_timer();
  initvideo();
  install_keyboard();
  reset65816();
  resetspu();
  install_int_ex(countsecs,MSEC_TO_TIMER(1000));

  // main loop
  while (!key[KEY_ESC])
  {
    if (palf)
      execsnes(312);
    else
      execsnes(262);
    //                rest(100);
    drawscreen();
    frame++;
    if (key[KEY_F1])
    {
      while (key[KEY_F1]);
      reset65816();
    }
    if (key[KEY_F2])
    {
      while (key[KEY_F2]);
      swapsprite();
    }
    if (key[KEY_F3])
    {
      while (key[KEY_F3]);
      hdmaena^=1;
    }
    if (key[KEY_F4])
    {
      while (key[KEY_F4]);
      inton^=1;
    }
    if (key[KEY_F5])
    {
      while (key[KEY_F5]);
      rotozoom^=1;
    }
    if (key[KEY_F6])
    {
      while (key[KEY_F6]);
      checkf000^=1;
    }
    if (key[KEY_F12])
    {
      while (key[KEY_F12]);
      takescreenshot();
    }
    if (key[KEY_1])
    {
      while (key[KEY_1]);
      extbgenable^=1;
    }
    if (key[KEY_2])
    {
      while (key[KEY_2]);
      extbgenable^=2;
    }
    if (key[KEY_3])
    {
      while (key[KEY_3]);
      extbgenable^=4;
    }
    if (key[KEY_4])
    {
      while (key[KEY_4]);
      extbgenable^=8;
    }
    cycle+=262*192;
  }
#if 0
  set_palette(dospal);
#endif
  closevideo();
  allegro_exit();
  printf("65816 regs :\n");
  dumpregs();
  printf("SPC700 regs :\n");
  dumpspuregs();
  printf("PPU regs :\n");
  dumpppuregs();
  printf("%i seconds %i cycles %i cycles/second %f snes speed\n",seconds,cycle,cycle/seconds,((float)cycle/(float)seconds)/2680000.0);
  f=fopen("ram.dmp","wb");
  fwrite(ram,0x20000,1,f);
  fclose(f);
  printf("NMI %i int %i\n",nmienable,intenable);
  printf("vertcomp %i\n",vertcomp);
  return 0;
}

END_OF_MAIN();
