/*
 * InfoSNES - Super Nintendo (TM) emulator
 * Copyrigh(c) 2003 Jay's Factory (jays_factory@excite.co.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*-------------------------------------------------------------------*/
/*  Include Files                                                    */
/*-------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "spu.h"
#include "dsp.h"
#include "65816.h"
#include "type.h"

/*-------------------------------------------------------------------*/
/*  SPU Resources                                                    */
/*-------------------------------------------------------------------*/

/* CPU Registers */
struct
{
  u16 pc;     /* PC */
  u8 x,s;     /* X, SP */
  struct
  {
    int n,v,p,h,z,c;
  } p;        /* PSW */
  union
  {
    struct
    {
      u8 a,y;
    } b;
    u16 w;
  } ya;       /* Y, A */
} spu;

u8 spuram[65536];
u8 spuctrl;
u8 spuoutp[4];

struct
{
  u8 latch[3],count[3];
  int lowclock,highclock;
} timers;

u8 spurom[64]=
{
  0xCD,0xEF,0xBD,0xE8,0x00,0xC6,0x1D,0xD0,
  0xFC,0x8F,0xAA,0xF4,0x8F,0xBB,0xF5,0x78,
  0xCC,0xF4,0xD0,0xFB,0x2F,0x19,0xEB,0xF4,
  0xD0,0xFC,0x7E,0xF4,0xD0,0x0B,0xE4,0xF5,
  0xCB,0xF4,0xD7,0x00,0xFC,0xD0,0xF3,0xAB,
  0x01,0x10,0xEF,0x7E,0xF4,0x10,0xEB,0xBA,
  0xF6,0xDA,0x00,0xBA,0xF4,0xC4,0xF4,0xDD,
  0x5D,0xD0,0xDB,0x1F,0x00,0x00,0xC0,0xFF
};

int spuout=0;
u8 dspregs[256],dspcurreg;

unsigned long lastread;
int spuins=0;

/*-------------------------------------------------------------------*/
/*  Functions                                                        */
/*-------------------------------------------------------------------*/

void dumpspuram()
{
  FILE *f=fopen("spu.dmp","wb");
  fwrite(spuram,65536,1,f);
  fclose(f);
}

u16 getspupc()
{
  return spu.pc;
}

void resetspu()
{
  memcpy(spuram+0xFFC0,spurom,0x40);
  spu.pc=spuram[0xFFFE]|(spuram[0xFFFF]<<8);
  spu.p.c=spu.p.z=spu.p.h=spu.p.p=spu.p.v=spu.p.n=0;
  atexit(dumpspuram);
  timers.count[0]=timers.count[1]=timers.count[2]=0;
}

void writespuport(int addr, u8 val)
{
  spuram[0xF4+(addr&3)]=val;
}

u8 readspuport(int addr)
{
  return spuoutp[addr&3];
}

void dumpspuregs()
{
  printf("YA=%04X X=%02X S=%02X PC=%04X\n",spu.ya.w,spu.x,spu.s,spu.pc);
  printf("%c%c%c%c%c%c\n",(spu.p.c)?'C':' ',(spu.p.z)?'Z':' ',(spu.p.h)?'H':' ',(spu.p.p)?'P':' ',(spu.p.v)?'V':' ',(spu.p.n)?'N':' ');
  printf("Last read %02X from %04X\n",lastread>>16,lastread&0xFFFF);
  printf("%i instructions\n",spuins);
  printf("Timer 0 : %02X %02X %01X\n",timers.count[0],timers.latch[0],spuram[0xFD]);
  printf("SPU ctrl : %02X\n",spuctrl);
  printf("SPU port in  : %02X %02X %02X %02X\n",spuram[0xF4],spuram[0xF5],spuram[0xF6],spuram[0xF7]);
  printf("SPU port out : %02X %02X %02X %02X\n\n",spuoutp[0],spuoutp[1],spuoutp[2],spuoutp[3]);
}

void writespumem(u16 addr, u8 val)
{
  switch (addr)
  {
  case 0xF1: /*SPU control*/
    if ((val&1) && !(spuctrl&1)) /*Reset Counter 0*/
    {
      spuram[0xFD]=0;
      timers.count[0]=0;
    }
    if ((val&1) && !(spuctrl&2)) /*Reset Counter 1*/
    {
      spuram[0xFE]=0;
      timers.count[0]=0;
    }
    if ((val&1) && !(spuctrl&4)) /*Reset Counter 2*/
    {
      spuram[0xFF]=0;
      timers.count[0]=0;
    }
    spuctrl=val;
    if (val&0x10)
      spuram[0xF4]=spuram[0xF5]=0;
    if (val&0x20)
      spuram[0xF6]=spuram[0xF7]=0;
    spuram[addr]=val;
    break;

  case 0xF2: /*DSP reg select*/
    dspcurreg=val;
    spuram[addr]=val;
    break;

  case 0xF3: /*DSP reg set*/
    dspregs[dspcurreg]=val;
    writedsp(dspcurreg,val);
    spuram[addr]=val;
    break;

  case 0xF4: case 0xF5: case 0xF6: case 0xF7: /*I/O ports*/
    spuoutp[addr&3]=val;
    break;

  case 0xFA: /*Timer 0*/
    timers.latch[0]=val;
    spuram[addr]=val;
    break;

  case 0xFB: /*Timer 1*/
    timers.latch[1]=val;
    spuram[addr]=val;
    break;

  case 0xFC: /*Timer 2*/
    timers.latch[2]=val;
    spuram[addr]=val;
    break;

  default:
    spuram[addr]=val;
    break;
  }
}

#define setspuzn(v)   spu.p.z=!(v); spu.p.n=(v)&0x80
#define setspuzn16(v) spu.p.z=!(v); spu.p.n=(v)&0x8000
#define setcmp(a,b)   spu.p.z=!(a-b); spu.p.n=(a-b)&0x80; spu.p.c=(a>=b)
#define setcmp16(a,b) spu.p.z=!(a-b); spu.p.n=(a-b)&0x8000; spu.p.c=(a>=b)

#define writefast(a,v) if ((a&0xFFF0)==0xF0) writespumem(a,v); else spuram[a]=v
#define spupull()     spuram[0x100|++spu.s]
#define spupush(v)    spuram[0x100|spu.s--]=v

#define SPUADC(r)       tempi=temp+r+((spu.p.c)?1:0);                          \
                        spu.p.c=tempi&0x100;                                   \
                        spu.p.h=((tempi&0x40) && !(tempi&0x40));               \
                        spu.p.v=(!((tempi^temp)&0x80)&&((r^tempi)&0x80));      \
                        r=tempi&0xFF;                                          \
                        setspuzn(r)

#define SPUSBC(r)       tempi=temp-(r-((spu.p.c)?0:1));                        \
                        spu.p.c=r>=temp;                                       \
                        spu.p.h=(!(tempi&0x40) && (tempi&0x40));               \
                        spu.p.v=(!((r^temp)&0x80)&&((r^tempi)&0x80));          \
                        r=tempi&0xFF;                                          \
                        setspuzn(r)

void execspu(int cycles)
{
  u8 opcode;
  int cycs;
  while (cycles>0)
  {
    opcode=spuram[spu.pc++];
    switch (opcode)
    { 
    case 0x02: /*SET1 0*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      writefast(tempa,spuram[tempa]|1);
      cycs=4;
      break;
      
    case 0x04: /*OR A,(DP)*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      spu.ya.b.a|=spuram[tempa];
      setspuzn(spu.ya.b.a);
      cycs=3;
      break;
      
    case 0x05: /*OR A,abs*/
      tempa=spuram[spu.pc]|(spuram[spu.pc+1]<<8);
      spu.pc+=2;
      spu.ya.b.a|=spuram[tempa];
      setspuzn(spu.ya.b.a);
      cycs=4;
      break;

    case 0x06: /*OR A,(X)*/
      tempa=spu.x;
      if (spu.p.p) tempa|=0x100;
      spu.ya.b.a|=spuram[tempa];
      setspuzn(spu.ya.b.a);
      cycs=3;
      break;
      
    case 0x08: /*OR A,#*/
      spu.ya.b.a|=spuram[spu.pc++];
      setspuzn(spu.ya.b.a);
      cycs=2;
      break;
      
    case 0x09: /*OR (DP),(DP)*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      temp=spuram[tempa];
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      temp2=spuram[tempa];
      temp2|=temp;
      setspuzn(temp2);
      writefast(tempa,temp2);
      cycs=6;
      break;
      
    case 0x0B: /*ASL DP*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      temp=spuram[tempa];
      spu.p.c=temp&0x80;
      temp<<=1;
      setspuzn(temp);
      writefast(tempa,temp);
      cycs=4;
      break;
			
    case 0x0C: /*MOV abs,Y*/
      tempa=spuram[spu.pc]|(spuram[spu.pc+1]<<8);
      spu.pc+=2;
      writefast(tempa,spu.ya.b.y);
      cycs=5;
      break;
      
    case 0x0E: /*TSET1*/
      tempa=spuram[spu.pc]|(spuram[spu.pc+1]<<8);
      spu.pc+=2;
      temp=spuram[tempa];
      setspuzn(temp&spu.ya.b.a);
      writefast(tempa,temp|spu.ya.b.a);
      cycs=6;
      break;
      
    case 0x10: /*BPL*/
      offset=(s8)spuram[spu.pc++];
      if (!spu.p.n)
      {
	spu.pc+=offset;
	cycs=4;
      }
      else
	cycs=2;
      break;
      
    case 0x12: /*CLR1 0*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      writefast(tempa,spuram[tempa]&~1);
      cycs=4;
      break;

    case 0x13: /*BBC0*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      offset=(s8)spuram[spu.pc++];
      cycs=5;
      if (!(spuram[tempa]&1))
      {
	spu.pc+=offset;
	cycs+=2;
      }
      break;

    case 0x1C: /*ASL A*/
      spu.p.c=spu.ya.b.a&0x80;
      spu.ya.b.a<<=1;
      setspuzn(spu.ya.b.a);
      cycs=2;
      break;
      
    case 0x1D: /*DEC X*/
      spu.x--;
      setspuzn(spu.x);
      cycs=2;
      break;
      
    case 0x1E: /*CMP X,abs*/
      tempa=spuram[spu.pc]|(spuram[spu.pc+1]<<8);
      spu.pc+=2;
      setcmp(spu.x,spuram[tempa]);
      cycs=3;
      break;
      
    case 0x1F: /*JMP (,x)*/
      tempa=spuram[spu.pc++];
      tempa|=(spuram[spu.pc++]<<8);
      tempa+=spu.x;
      spu.pc=spuram[tempa]|(spuram[tempa+1]<<8);
      cycs=6;
      break;
      
    case 0x20: /*CLRP*/
      spu.p.p=0;
      cycs=2;
      break;
      
    case 0x24: /*AND A,(DP)*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      spu.ya.b.a&=spuram[tempa];
      setspuzn(spu.ya.b.a);
      cycs=3;
      break;
      
    case 0x28: /*AND A,#*/
      spu.ya.b.a&=spuram[spu.pc++];
      setspuzn(spu.ya.b.a);
      cycs=2;
      break;

    case 0x2B: /*ROL (DP)*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      temp=spuram[tempa];
      tempi=spu.p.c;
      spu.p.c=temp&0x80;
      temp<<=1;
      if (tempi) temp|=1;
      writefast(tempa,temp);
      cycs=4;
      break;

    case 0x2D: /*PUSH A*/
      spupush(spu.ya.b.a);
      cycs=4;
      break;

    case 0x2F: /*BRA*/
      offset=(s8)spuram[spu.pc++];
      spu.pc+=offset;
      cycs=3;
      break;

    case 0x30: /*BMI*/
      offset=(s8)spuram[spu.pc++];
      if (spu.p.n)
      {
	spu.pc+=offset;
	cycs=4;
      }
      else
	cycs=2;
      break;
      
    case 0x33: /*BBC1*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      offset=(s8)spuram[spu.pc++];
      cycs=5;
      if (!(spuram[tempa]&2))
      {
	spu.pc+=offset;
	cycs+=2;
      }
      break;
      
    case 0x36: /*AND A,absy*/
      tempa=(spuram[spu.pc]|(spuram[spu.pc+1]<<8))+spu.ya.b.y;
      spu.pc+=2;
      spu.ya.b.a&=spuram[tempa];
      setspuzn(spu.ya.b.a);
      cycs=4;
      break;
      
    case 0x3A: /*INCW (DP)*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      tempw=spuram[tempa]|(spuram[tempa+1]<<8);
      tempw++;
      writefast(tempa,tempw);
      writefast(tempa+1,tempw>>8);
      setspuzn16(tempw);
      cycs=6;
      break;
      
    case 0x3D: /*INC X*/
      spu.x++;
      setspuzn(spu.x);
      cycs=2;
      break;

    case 0x3E: /*CMP X,(DP)*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      setcmp(spu.x,spuram[tempa]);
      cycs=3;
      break;
      
    case 0x3F: /*CALL*/
      tempa=spuram[spu.pc]|(spuram[spu.pc+1]<<8);
      spu.pc+=2;
      spupush(spu.pc>>8);
      spupush(spu.pc&0xFF);
      spu.pc=tempa;
      cycs=8;
      break;
      
    case 0x44: /*EOR A,(DP)*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      spu.ya.b.a^=spuram[tempa];
      setspuzn(spu.ya.b.a);
      cycs=3;
      break;
      
    case 0x48: /*EOR A,#*/
      spu.ya.b.a^=spuram[spu.pc++];
      setspuzn(spu.ya.b.a);
      cycs=2;
      break;
      
    case 0x4B: /*LSR DP*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      temp=spuram[tempa];
      spu.p.c=temp&1;
      temp>>=1;
      setspuzn(temp);
      writefast(tempa,temp);
      cycs=4;
      break;
      
    case 0x4D: /*PUSH X*/
      spupush(spu.x);
      cycs=4;
      break;
      
    case 0x4E: /*TCLR1*/
      tempa=spuram[spu.pc]|(spuram[spu.pc+1]<<8);
      spu.pc+=2;
      temp=spuram[tempa];
      setspuzn(temp&spu.ya.b.a);
      writefast(tempa,temp&~spu.ya.b.a);
      cycs=6;
      break;
      
//                        case 0x51: /*TCALL*/
//                        push(spu.pc>>8);
//                        push(spu.pc&0xFF);

    case 0x58: /*EOR dp,#*/
      temp2=spuram[spu.pc++];
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      temp=spuram[tempa];
      temp^=temp2;
      writefast(tempa,temp);
      setspuzn(temp);
      cycs=6;
      break;
      
    case 0x5A: /*CMPW YA,(DP)*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      tempw=spuram[tempa]|(spuram[tempa+1]<<8);
      setcmp16(spu.ya.w,tempw);
      cycs=6;
      break;
      
    case 0x5C: /*LSR A*/
      spu.p.c=spu.ya.b.a&1;
      spu.ya.b.a>>=1;
      setspuzn(spu.ya.b.a);
      cycs=2;
      break;

    case 0x5D: /*MOV X,A*/
      spu.x=spu.ya.b.a;
      setspuzn(spu.x);
      cycs=2;
      break;
      
    case 0x5E: /*CMP Y,abs*/
      tempa=spuram[spu.pc]|(spuram[spu.pc+1]<<8);
      spu.pc+=2;
      setcmp(spu.ya.b.y,spuram[tempa]);
      cycs=3;
      break;

    case 0x5F: /*JMP abs*/
      spu.pc=spuram[spu.pc]|(spuram[spu.pc+1]<<8);
      cycs=3;
      break;
      
    case 0x60: /*CLRC*/
      spu.p.c=0;
      cycs=2;
      break;
      
    case 0x64: /*CMP A,(DP)*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      setcmp(spu.ya.b.a,spuram[tempa]);
      cycs=3;
      break;
      
    case 0x65: /*CMP A,abs*/
      tempa=spuram[spu.pc]|(spuram[spu.pc+1]<<8);
      spu.pc+=2;
      setcmp(spu.ya.b.a,spuram[tempa]);
      cycs=3;
      break;
      
    case 0x68: /*CMP A,#*/
      temp=spuram[spu.pc++];
      setcmp(spu.ya.b.a,temp);
      cycs=2;
      break;
      
    case 0x69: /*CMP (DP),(DP)*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      temp=spuram[tempa];
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      setcmp(spuram[tempa],temp);
      cycs=6;
      break;
      
    case 0x6B: /*ROR (DP)*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      temp=spuram[tempa];
      tempi=spu.p.c;
      spu.p.c=temp&1;
      temp>>=1;
      if (tempi) temp|=0x80;
      writefast(tempa,temp);
      cycs=4;
      break;
      
    case 0x6D: /*PUSH Y*/
      spupush(spu.ya.b.y);
      cycs=4;
      break;
      
    case 0x6E: /*DBNZ (DP)*/
      tempa=spuram[spu.pc++];
      offset=(s8)spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      spuram[tempa]--;
      cycs=5;
      if (spuram[tempa])
      {
	spu.pc+=offset;
	cycs+=2;
      }
      break;
      
    case 0x6F: /*RET*/
      spu.pc=spupull();
      spu.pc|=spupull()<<8;
      cycs=5;
      break;
      
    case 0x74: /*CMP A,(DP+X)*/
      tempa=(spuram[spu.pc++]+spu.x)&0xFF;
      if (spu.p.p) tempa|=0x100;
      setcmp(spu.ya.b.a,spuram[tempa]);
      cycs=4;
      break;
      
    case 0x75: /*CMP A,absx*/
      tempa=(spuram[spu.pc]|(spuram[spu.pc+1]<<8))+spu.x;
      spu.pc+=2;
      setcmp(spu.ya.b.a,spuram[tempa]);
      cycs=3;
      break;
      
    case 0x76: /*CMP A,absy*/
      tempa=(spuram[spu.pc]|(spuram[spu.pc+1]<<8))+spu.ya.b.y;
      spu.pc+=2;
      setcmp(spu.ya.b.a,spuram[tempa]);
      cycs=3;
      break;
      
    case 0x78: /*CMP (DP),#*/
      temp=spuram[spu.pc++];
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      setcmp(spuram[tempa],temp);
      cycs=4;
      break;
      
    case 0x7A: /*ADDW YA,(DP)*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      tempa2=spuram[tempa]|(spuram[tempa+1]<<8);
      tempi=spu.ya.w+tempa2;
      spu.p.c=tempi&0x10000;
      spu.p.v=(!((spu.ya.w^tempa2)&0x8000)&&((spu.ya.w^tempi)&0x8000));
      spu.ya.w+=tempa2;
      setspuzn16(spu.ya.w);
      cycs=5;
      break;
      
    case 0x7C: /*ROR A*/
      tempi=spu.p.c;
      spu.p.c=spu.ya.b.a&1;
      spu.ya.b.a>>=1;
      if (tempi) spu.ya.b.a|=0x80;
      setspuzn(spu.ya.b.a);
      cycs=2;
      break;
      
    case 0x7D: /*MOV A,X*/
      spu.ya.b.a=spu.x;
      setspuzn(spu.ya.b.a);
      cycs=2;
      break;
      
    case 0x7E: /*CMP Y,(DP)*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      setcmp(spu.ya.b.y,spuram[tempa]);
      cycs=3;
      //                        spuout=1;
      break;
      
    case 0x80: /*SETC*/
      spu.p.c=1;
      cycs=2;
      break;
      
    case 0x84: /*ADC A,(DP)*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      temp=spuram[tempa];
      SPUADC(spu.ya.b.a);
      cycs=3;
      break;
      
    case 0x85: /*ADC A,abs*/
      tempa=spuram[spu.pc]|(spuram[spu.pc+1]<<8);
      spu.pc+=2;
      temp=spuram[tempa];
      SPUADC(spu.ya.b.a);
      cycs=4;
      break;

    case 0x86: /*ADC A,(X)*/
      tempa=spu.x;
      if (spu.p.p) tempa|=0x100;
      temp=spuram[tempa];
      SPUADC(spu.ya.b.a);
      cycs=3;
      break;

    case 0x87: /*ADC A,(DP+X)*/
      tempa2=(spuram[spu.pc++]+spu.x)&0xFF;
      if (spu.p.p) tempa2|=0x100;
      tempa=spuram[tempa2]|(spuram[tempa2+1]<<8);
      temp=spuram[tempa];
      SPUADC(spu.ya.b.a);
      cycs=6;
      break;
      
    case 0x88: /*ADC A,#*/
      temp=spuram[spu.pc++];
      SPUADC(spu.ya.b.a);
      cycs=2;
      break;
      
    case 0x8B: /*DEC (DP)*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      temp=spuram[tempa]-1;
      setspuzn(temp);
      writefast(tempa,temp);
      cycs=4;
      break;
      
    case 0x8D: /*MOV Y,#*/
      spu.ya.b.y=spuram[spu.pc++];
      setspuzn(spu.ya.b.y);
      cycs=2;
      break;
      
    case 0x8F: /*MOV (DP),#*/
      temp=spuram[spu.pc++];
      tempa=spuram[spu.pc++];
      if (spu.p.p)
	writefast(tempa|0x100,temp);
      else
	writefast(tempa,temp);
      cycs=4;
      break;
      
    case 0x90: /*BCC*/
      offset=(s8)spuram[spu.pc++];
      if (!spu.p.c)
      {
	spu.pc+=offset;
	cycs=4;
      }
      else
	cycs=2;
      break;
      
    case 0x94: /*ADC A,(DP+X)*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      temp=spuram[tempa+spu.x];
      SPUADC(spu.ya.b.a);
      cycs=3;
      break;
      
    case 0x95: /*ADC A,absx*/
      tempa=(spuram[spu.pc]|(spuram[spu.pc+1]<<8))+spu.x;
      spu.pc+=2;
      temp=spuram[tempa];
      SPUADC(spu.ya.b.a);
      cycs=5;
      break;
      
    case 0x96: /*ADC A,absy*/
      tempa=(spuram[spu.pc]|(spuram[spu.pc+1]<<8))+spu.ya.b.y;
      spu.pc+=2;
      temp=spuram[tempa];
      SPUADC(spu.ya.b.a);
      cycs=5;
      break;
      
    case 0x97: /*ADC A,(DP)+Y*/
      tempa2=spuram[spu.pc++];
      if (spu.p.p) tempa2|=0x100;
      tempa=spuram[tempa2]|(spuram[tempa2+1]<<8);
      tempa+=spu.ya.b.y;
      temp=spuram[tempa];
      SPUADC(spu.ya.b.a);
      cycs=6;
      break;

    case 0x98: /*ADC (DP),#*/
      temp=spuram[spu.pc++];
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      temp2=spuram[tempa];
      SPUADC(temp2);
      writefast(tempa,temp2);
      cycs=5;
      break;
      
    case 0x9A: /*SUBW YA,(DP)*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      tempa2=spuram[tempa]|(spuram[tempa+1]<<8);
      tempa=spu.ya.w-tempa2;
      spu.p.c=(spu.ya.w>=tempa2);
      spu.p.v=(!((tempa^tempa2)&0x8000)&&((spu.ya.w^tempa)&0x8000)); \
      spu.ya.w-=tempa2;
      setspuzn16(spu.ya.w);
      cycs=5;
      break;
      
    case 0x9B: /*DEC (DP+X)*/
      tempa=(spuram[spu.pc++]+spu.x)&0xFF;
      if (spu.p.p) tempa|=0x100;
      temp=spuram[tempa]-1;
      setspuzn(temp);
      writefast(tempa,temp);
      cycs=4;
      break;
      
    case 0x9C: /*DEC A*/
      spu.ya.b.a--;
      setspuzn(spu.ya.b.a);
      cycs=2;
      break;

    case 0x9D: /*MOV X,SP*/
      spu.x=spu.s;
      setspuzn(spu.x);
      cycs=2;
      break;

    case 0x9E: /*DIV YA,X*/
      if (!spu.x)
      {
	spu.p.v=1;
	spu.ya.w=0xFFFF;
      }
      else
      {
	spu.p.v=0;
	temp=spu.ya.w/spu.x;
	spu.ya.b.y=spu.ya.w%spu.x;
	spu.ya.b.a=temp;
	setspuzn(spu.ya.b.a);
      }
      cycs=12;
      break;
      
    case 0x9F: /*XCN*/
      spu.ya.b.a=(spu.ya.b.a>>4)|(spu.ya.b.a<<4);
      setspuzn(spu.ya.b.a);
      cycs=5;
      break;
      
    case 0xA2: /*SET1 5*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      writefast(tempa,spuram[tempa]|0x20);
      cycs=4;
      break;
      
    case 0xA4: /*SBC (DP)*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      temp=spuram[tempa];
      SPUSBC(spu.ya.b.a);
      cycs=3;
      break;
      
    case 0xA6: /*SBC (X)*/
      tempa=spu.x;
      if (spu.p.p) tempa|=0x100;
      temp=spuram[tempa];
      SPUSBC(spu.ya.b.a);
      cycs=3;
      break;
      
    case 0xA7: /*SBC (DP+X)*/
      tempa=(spuram[spu.pc++]+spu.x)&0xFF;
      if (spu.p.p) tempa|=0x100;
      temp=spuram[tempa];
      SPUSBC(spu.ya.b.a);
      cycs=3;
      break;
      
    case 0xA8: /*SBC A,#*/
      temp=spuram[spu.pc++];
      SPUSBC(spu.ya.b.a);
      cycs=3;
      break;
      
    case 0xA9: /*SBC (DP),(DP)*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      temp=spuram[tempa];
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      temp2=spuram[tempa];
      SPUSBC(temp2);
      writefast(tempa,temp2);
      cycs=6;
      break;
      
    case 0xAB: /*INC (DP)*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      temp=spuram[tempa]+1;
      setspuzn(temp);
      writefast(tempa,temp);
      cycs=4;
      break;
      
    case 0xAC: /*INC abs*/
      tempa=spuram[spu.pc]|(spuram[spu.pc+1]<<8);
      spu.pc+=2;
      temp=spuram[tempa]+1;
      setspuzn(temp);
      writefast(tempa,temp);
      cycs=4;
      break;
      
    case 0xAD: /*CMP Y,#*/
      temp=spuram[spu.pc++];
      setcmp(spu.ya.b.y,temp);
      cycs=2;
      break;
      
    case 0xAE: /*PULL A*/
      spu.ya.b.a=spupull();
      setspuzn(spu.ya.b.a);
      cycs=4;
      break;
      
    case 0xAF: /*MOV (X)+,A*/
      if (spu.p.p) writefast(spu.x|0x100,spu.ya.b.a);
      else         writefast(spu.x,spu.ya.b.a);
      spu.x++;
      cycs=4;
      break;
      
    case 0xB0: /*BCS*/
      offset=(s8)spuram[spu.pc++];
      if (spu.p.c)
      {
	spu.pc+=offset;
	cycs=4;
      }
      else
	cycs=2;
      break;
      
    case 0xB2: /*CLR1 5*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      writefast(tempa,spuram[tempa]&~0x20);
      cycs=4;
      break;
      
    case 0xB5: /*SBC absx*/
      tempa=(spuram[spu.pc]|(spuram[spu.pc+1]<<8))+spu.x;
      spu.pc+=2;
      temp=spuram[tempa];
      SPUSBC(spu.ya.b.a);
      cycs=3;
      break;
      
    case 0xB6: /*SBC absy*/
      tempa=(spuram[spu.pc]|(spuram[spu.pc+1]<<8))+spu.ya.b.y;
      spu.pc+=2;
      temp=spuram[tempa];
      SPUSBC(spu.ya.b.a);
      cycs=3;
      break;
      
    case 0xBA: /*MOV YA,(DP)*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      spu.ya.b.a=spuram[tempa];
      spu.ya.b.y=spuram[tempa+1];
      setspuzn16(spu.ya.w);
      cycs=4;
      break;
      
    case 0xBB: /*INC (DP+X)*/
      tempa=(spuram[spu.pc++]+spu.x)&0xFF;
      if (spu.p.p) tempa|=0x100;
      temp=spuram[tempa]+1;
      setspuzn(temp);
      writefast(tempa,temp);
      cycs=4;
      break;
      
    case 0xBC: /*INC A*/
      spu.ya.b.a++;
      setspuzn(spu.ya.b.a);
      cycs=2;
      break;
      
    case 0xBD: /*MOV SP,X*/
      spu.s=spu.x;
      cycs=2;
      break;
      
    case 0xBF: /*MOV A,(X)+*/
      tempa=spu.x;
      if (spu.p.p) tempa|=0x100;
      spu.ya.b.a=spuram[tempa];
      setspuzn(spu.ya.b.a);
      spu.x++;
      cycs=4;
      break;
      
    case 0xC4: /*MOV (DP),A*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      writefast(tempa,spu.ya.b.a);
      cycs=3;
      break;
      
    case 0xC5: /*MOV abs,A*/
      tempa=spuram[spu.pc]|(spuram[spu.pc+1]<<8);
      spu.pc+=2;
      writefast(tempa,spu.ya.b.a);
      cycs=5;
      break;
      
    case 0xC6: /*MOV (X),A*/
      if (spu.p.p) writefast(spu.x|0x100,spu.ya.b.a);
      else         writefast(spu.x,spu.ya.b.a);
      cycs=2;
      break;

    case 0xC7: /*MOV (DP+X),A*/
      tempa2=(spuram[spu.pc++]+spu.x)&0xFF;
      if (spu.p.p) tempa2|=0x100;
      tempa=spuram[tempa2]|(spuram[tempa2+1]<<8);
      writefast(tempa,spu.ya.b.a);
      cycs=7;
      break;

    case 0xC8: /*CMP X,#*/
      temp=spuram[spu.pc++];
      setcmp(spu.x,temp);
      cycs=2;
      break;
      
    case 0xC9: /*MOV abs,X*/
      tempa=spuram[spu.pc]|(spuram[spu.pc+1]<<8);
      spu.pc+=2;
      writefast(tempa,spu.x);
      cycs=5;
      break;
      
    case 0xCB: /*MOV (DP),Y*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      writefast(tempa,spu.ya.b.y);
      cycs=3;
      break;
      
    case 0xCC: /*ASL abs*/
      tempa=spuram[spu.pc]|(spuram[spu.pc+1]<<8);
      spu.pc+=2;
      temp=spuram[tempa];
      spu.p.c=temp&0x80;
      temp<<=1;
      writefast(tempa,temp);
      setspuzn(temp);
      cycs=5;
      break;
      
    case 0xCD: /*MOV X,#*/
      spu.x=spuram[spu.pc++];
      setspuzn(spu.x);
      cycs=2;
      break;
      
    case 0xCE: /*PULL X*/
      spu.x=spupull();
      setspuzn(spu.x);
      cycs=4;
      break;
      
    case 0xCF: /*MUL YA*/
      spu.ya.w=spu.ya.b.a*spu.ya.b.y;
      setspuzn16(spu.ya.w);
      cycs=9;
      break;
      
    case 0xD0: /*BNE*/
      offset=(s8)spuram[spu.pc++];
      if (!spu.p.z)
      {
	spu.pc+=offset;
	cycs=4;
      }
      else
	cycs=2;
      break;
      
    case 0xD4: /*MOV (DP+X),A*/
      tempa=(spuram[spu.pc++]+spu.x)&0xFF;
      if (spu.p.p) tempa|=0x100;
      writefast(tempa,spu.ya.b.a);
      cycs=3;
      break;
      
    case 0xD5: /*MOV absx,A*/
      tempa=(spuram[spu.pc]|(spuram[spu.pc+1]<<8))+spu.x;
      spu.pc+=2;
      writefast(tempa,spu.ya.b.a);
      cycs=6;
      break;
      
    case 0xD6: /*MOV absy,A*/
      tempa=(spuram[spu.pc]|(spuram[spu.pc+1]<<8))+spu.ya.b.y;
      spu.pc+=2;
      writefast(tempa,spu.ya.b.a);
      cycs=6;
      break;

    case 0xD7: /*MOV (DP)+Y,A*/
      tempa2=spuram[spu.pc++];
      if (spu.p.p) tempa2|=0x100;
      tempa=spuram[tempa2]|(spuram[tempa2+1]<<8);
      tempa+=spu.ya.b.y;
      writefast(tempa,spu.ya.b.a);
      cycs=7;
      break;
      
    case 0xD8: /*MOV (DP),X*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      writefast(tempa,spu.x);
      cycs=4;
      break;

    case 0xD9: /*MOV (DP+Y),X*/
      tempa=(spuram[spu.pc++]+spu.ya.b.y)&0xFF;
      if (spu.p.p) tempa|=0x100;
      writefast(tempa,spu.x);
      cycs=5;
      break;
      
    case 0xDA: /*MOV (DP),YA*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      writefast(tempa,spu.ya.b.a);
      writefast(tempa+1,spu.ya.b.y);
      if (spuout) printf("%04X=%04X\n",tempa,spu.ya.w);
      cycs=4;
      break;
      
    case 0xDB: /*MOV (DP+X),Y*/
      tempa=(spuram[spu.pc++]+spu.x)&0xFF;
      if (spu.p.p) tempa|=0x100;
      writefast(tempa,spu.ya.b.y);
      cycs=5;
      break;

    case 0xDC: /*DEC Y*/
      spu.ya.b.y--;
      setspuzn(spu.ya.b.y);
      cycs=2;
      break;
      
    case 0xDD: /*MOV A,Y*/
      spu.ya.b.a=spu.ya.b.y;
      setspuzn(spu.ya.b.a);
      cycs=2;
      break;
      
    case 0xDE: /*CBNE (DP+X)*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      offset=(s8)spuram[spu.pc++];
      cycs=6;
      if (spuram[tempa]!=spu.ya.b.a)
      {
	spu.pc+=offset;
	cycs+=2;
      }
      break;

    case 0xE0: /*CLRV*/
      spu.p.v=spu.p.h=0;
      cycs=2;
      break;
      
    case 0xE2: /*SET1 7*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      writefast(tempa,spuram[tempa]|0x80);
      cycs=4;
      break;
      
    case 0xE3: /*BBS7*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      offset=(s8)spuram[spu.pc++];
      cycs=5;
      if (spuram[tempa]&0x80)
      {
	spu.pc+=offset;
	cycs+=2;
      }
      break;
      
    case 0xE4: /*MOV A,(DP)*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      spu.ya.b.a=spuram[tempa];
      setspuzn(spu.ya.b.a);
      cycs=3;
      break;
      
    case 0xE5: /*MOV A,abs*/
      tempa=spuram[spu.pc]|(spuram[spu.pc+1]<<8);
      spu.pc+=2;
      spu.ya.b.a=spuram[tempa];
      setspuzn(spu.ya.b.a);
      cycs=4;
      break;
      
    case 0xE6: /*MOV A,(X)*/
      tempa=spu.x;
      if (spu.p.p) tempa|=0x100;
      spu.ya.b.a=spuram[tempa];
      setspuzn(spu.ya.b.a);
      cycs=3;
      break;
      
    case 0xE7: /*MOV A,((DP+X))*/
      tempa=(spuram[spu.pc++]+spu.x)&0xFF;
      if (spu.p.p) tempa|=0x100;
      tempa2=spuram[tempa]|(spuram[tempa+1]<<8);
      spu.ya.b.a=spuram[tempa2];
      setspuzn(spu.ya.b.a);
      cycs=6;
      break;
      
    case 0xE8: /*MOV A,#*/
      spu.ya.b.a=spuram[spu.pc++];
      setspuzn(spu.ya.b.a);
      cycs=2;
      break;
      
    case 0xE9: /*MOV X,abs*/
      tempa=spuram[spu.pc]|(spuram[spu.pc+1]<<8);
      spu.pc+=2;
      spu.x=spuram[tempa];
      setspuzn(spu.x);
      cycs=4;
      break;
      
    case 0xEB: /*MOV Y,(DP)*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      spu.ya.b.y=spuram[tempa];
      setspuzn(spu.ya.b.y);
      cycs=3;
      break;
      
    case 0xEC: /*MOV Y,abs*/
      tempa=spuram[spu.pc]|(spuram[spu.pc+1]<<8);
      spu.pc+=2;
      spu.ya.b.y=spuram[tempa];
      setspuzn(spu.ya.b.y);
      cycs=4;
      break;
      
    case 0xED: /*NOTC*/
      if (spu.p.c) spu.p.c=0;
      else         spu.p.c=1;
      cycs=3;
      break;
      
    case 0xEE: /*PULL Y*/
      spu.ya.b.y=spupull();
      setspuzn(spu.ya.b.y);
      cycs=4;
      break;
      
    case 0xF0: /*BEQ*/
      offset=(s8)spuram[spu.pc++];
      if (spu.p.z)
      {
	spu.pc+=offset;
	cycs=4;
      }
      else
	cycs=2;
      break;
      
    case 0xF2: /*CLR1 7*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      writefast(tempa,spuram[tempa]&0x7F);
      cycs=4;
      break;
      
    case 0xF3: /*BBC7*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      offset=(s8)spuram[spu.pc++];
      cycs=5;
      if (!(spuram[tempa]&0x80))
      {
	spu.pc+=offset;
	cycs+=2;
      }
      break;

    case 0xF4: /*MOV A,(DP+X)*/
      tempa=(spuram[spu.pc++]+spu.x)&0xFF;
      if (spu.p.p) tempa|=0x100;
      spu.ya.b.a=spuram[tempa];
      setspuzn(spu.ya.b.a);
      cycs=3;
      break;

    case 0xF5: /*MOV A,absx*/
      tempa=(spuram[spu.pc]|(spuram[spu.pc+1]<<8))+spu.x;
      spu.pc+=2;
      spu.ya.b.a=spuram[tempa];
      setspuzn(spu.ya.b.a);
      cycs=4;
      break;

    case 0xF6: /*MOV A,absy*/
      tempa=(spuram[spu.pc]|(spuram[spu.pc+1]<<8))+spu.ya.b.y;
      spu.pc+=2;
      spu.ya.b.a=spuram[tempa];
      setspuzn(spu.ya.b.a);
      cycs=5;
      break;

    case 0xF7: /*MOV A,(DP)+Y*/
      tempa2=spuram[spu.pc++];
      if (spu.p.p) tempa2|=0x100;
      tempa=spuram[tempa2]|(spuram[tempa2+1]<<8);
      tempa+=spu.ya.b.y;
      spu.ya.b.a=spuram[tempa];
      setspuzn(spu.ya.b.a);
      cycs=5;
      break;

    case 0xF8: /*MOV X,(DP)*/
      tempa=spuram[spu.pc++];
      if (spu.p.p) tempa|=0x100;
      spu.x=spuram[tempa];
      setspuzn(spu.x);
      cycs=3;
      break;

    case 0xF9: /*MOV X,(DP+Y)*/
      tempa=(spuram[spu.pc++]+spu.ya.b.y)&0xFF;
      if (spu.p.p) tempa|=0x100;
      spu.x=spuram[tempa];
      setspuzn(spu.x);
      cycs=4;
      break;

    case 0xFB: /*MOV Y,(DP+X)*/
      tempa=(spuram[spu.pc++]+spu.x)&0xFF;
      if (spu.p.p) tempa|=0x100;
      spu.ya.b.y=spuram[tempa];
      setspuzn(spu.ya.b.y);
      cycs=4;
      break;
      
    case 0xFC: /*INC Y*/
      spu.ya.b.y++;
      setspuzn(spu.ya.b.y);
      cycs=2;
      break;
      
    case 0xFD: /*MOV Y,A*/
      spu.ya.b.y=spu.ya.b.a;
      setspuzn(spu.ya.b.y);
      cycs=2;
      break;
      
    case 0xFE: /*DBNZ Y*/
      offset=(s8)spuram[spu.pc++];
      spu.ya.b.y--;
      cycs=4;
      if (spu.ya.b.y)
	{
	  spu.pc+=offset;
	  cycs+=2;
	}
      break;
      
    default:
      printf("\nBad SPU opcode %02X\n",opcode);
      spu.pc--;
      dumpspuregs();
      exit(-1);
    }
    cycles-=cycs;
    spuins++;

    if (spuout) printf("%04X %02X %02X %04X %c%c%c%c%c%c %06X %02X\n",spu.ya.w,spu.x,spu.s,spu.pc,(spu.p.c)?'C':' ',(spu.p.z)?'Z':' ',(spu.p.h)?'H':' ',(spu.p.p)?'P':' ',(spu.p.v)?'V':' ',(spu.p.n)?'N':' ',lastread,opcode);

    timers.lowclock+=cycs;
    if (timers.lowclock>=256)
    {
      timers.lowclock-=256;
      if (spuctrl&1) /*Timer 1*/
      {
	timers.count[0]++;
	if (timers.count[0]==timers.latch[0])
	{
	  timers.count[0]=0;
	  spuram[0xFD]++;
	  spuram[0xFD]&=0xF;
	}
      }
      if (spuctrl&2) /*Timer 2*/
      {
	timers.count[1]++;
	if (timers.count[1]==timers.latch[1])
	{
	  timers.count[1]=0;
	  spuram[0xFE]++;
	  spuram[0xFE]&=0xF;
	}
      }
    }
    if (spuctrl&4) /*Timer 3*/
    {
      timers.highclock+=cycs;
      if (timers.highclock>=32)
      {
	timers.highclock-=32;
	timers.count[2]++;
	if (timers.count[2]==timers.latch[2])
	{
	  timers.count[2]=0;
	  spuram[0xFF]++;
	  spuram[0xFF]&=0xF;
	}
      }
    }
  }
}
